# -*- coding: latin-1; -*-
# $Id:  $
#################################################
#
#################################################
#-- Initialisation python
from wrap import *
import math
metafor = None

#=========================================================================
#
def getParameters(_parameters):
    parameters = {}
    # Paramtres serre-Flanc
    parameters['forceSF']  = 1500.0 # Force de serrage du serre-Flanc
    parameters['amortissement'] = 5.0e-7 # Paramtre d'amortissement
    #paramtres contact 
    parameters['peno']   = 1.0e4
    parameters['mu']     = 0.12     
    # Donnes matriau tle
    parameters['rho']      = 7.8e-9   # densit
    parameters['Young']    = 210000.0 # Module de Young
    parameters['poisson']  = 0.3      # Coef Poisson
    
    parameters['trac_Rm']  = 625      # Donnes matriau
    parameters['trac_R02'] = 369      # Donnes matriau
    parameters['trac_n']   = 0.194    # Donnes matriau
    
    # Gomtrie - maillage du Flanc
    parameters['nx'] = 135
    parameters['ny'] = 2
    parameters['Lx'] = 135.
    parameters['Ly'] = 1.
    parameters['nipX'] = 2
    parameters['nipY'] = 2

    # Gomtrie de la Matrice
    parameters['Y0M']      = 0.0 
    parameters['LxM1']     = 50.0
    #parameters['LxM2']    =  post trait    
    parameters['LxM3']     = 140.0 
    parameters['LyM']      = 61.1
    parameters['rM1']      = 10.0 
    parameters['rM2']      = 10.0 
    parameters['thetaM']   = 80.0     
    parameters['ddyM']     = 1.0
    
    # Gomtrie du Poincon
    parameters['Y0P']      = parameters['Ly']
    parameters['LxP1']      = 49.9 
    parameters['rP1']      =  8.9
    parameters['rP2']      = 11.1 
    parameters['rP3']      =  5.0    
    parameters['LyP']      = 61.1  #0.110000002E+01+0.590441971E+02
    parameters['thetaP']   = 80.0   
    parameters['ddyP']   = parameters['LyP']

    # Gomtrie du Serre Flanc
    parameters['X0SF']      = 85.8
    parameters['rSF']       = 2.0 
    parameters['Y0SF']      = parameters['Ly'] + parameters['rSF']
    parameters['LxSF']      = 50.0    
    parameters['ddySF']    = 0.01 # dplacement initial du serre-Flanc 
    
    # Paramtres du calcul 
    #parameters['ndyn'] = 0 
    parameters['ndyn'] = 2
    parameters['prec']   = 1.0e-4
	
    parameters['cmvim']   = VES_CMVIM_SRI
    parameters['PEAS']    = 1.0e-8
    


    # Mise  jour des parametres par defaut avec les parametres passs
    parameters.update(_parameters)
    #
    #Ne pas remonter ce qui est ci dessous au dessus  parameters.update(_parameters)
    parameters['ih_n']     = parameters['trac_n']     # Krupkowsky    
    parameters['ih_k']     = parameters['trac_Rm']*math.exp(parameters['trac_n'])/ math.pow(parameters['trac_n'], parameters['trac_n'])  # Krupkowsky
    parameters['ih_evpl0'] = math.pow((parameters['trac_R02']/parameters['ih_k']), (1.0/parameters['trac_n']) )-0.002  # Krupkowsky : evpl0
    
    print "ih_n = ", parameters['ih_n'] 
    print "ih_k = ", parameters['ih_k'] 
    print "ih_evpl0 = ", parameters['ih_evpl0'] 
    
    parameters['thetaMRad'] = parameters['thetaM'] * math.pi / 180.0
    parameters['thetaPRad'] = parameters['thetaP'] * math.pi / 180.0
    
    lxm2  = parameters['LxM1'] + parameters['rM1'] * math.sin(parameters['thetaMRad'])
    lxm2 += (parameters['LyM'] - (parameters['rM1'] + parameters['rM2']) * (1.0-math.cos(parameters['thetaMRad']))) / math.tan(parameters['thetaMRad'])
    lxm2 += parameters['rM2'] * math.sin(parameters['thetaMRad'])
    parameters['LxM2'] = lxm2
    
    lxp2  = parameters['LxP1'] + parameters['rP1'] * math.sin(parameters['thetaPRad'])
    lxp2 += (parameters['LyP'] - (parameters['rP1'] + parameters['rP2']) * (1.0-math.cos(parameters['thetaPRad']))) / math.tan(parameters['thetaPRad'])
    lxp2 += parameters['rP2'] * math.sin(parameters['thetaPRad'])
    parameters['LxP2'] = lxp2
    #parameters['LxM2'] = parameters['LxM1'] + (parameters['LyM'] - parameters['rM2'] * (1.0-math.cos(parameters['thetaMRad']))) / math.tan(parameters['thetaMRad']) + parameters['rM1']*math.sin(parameters['thetaMRad'])
    # Calcul des paramtres dpendant des paramtres de base    
    return parameters
#=========================================================================


def getMetafor(_parameters={}):
    global metafor
    if metafor: return metafor
    p = getParameters(_parameters)
    metafor    = Metafor()
    domain     = metafor.getDomain()
    geometry = domain.getGeometry()
    geometry.setDimPlaneStrain(1.0)

    #-- maillage --
    from toolbox.meshedGeometry2D import createRectangleSameId
    createRectangleSameId(domain, p['nx'], p['ny'], p['Lx'], p['Ly'], 0.0,   0.0,   0)

    # Gomtrie des outils : matrice
    pset = geometry.getPointSet()
    p1001 = pset.define(1001, -1.0,                                             p['Y0M']-p['LyM'])
    p1002 = pset.define(1002, p['LxM1'],                                        p['Y0M']-p['LyM'])
    p1003 = pset.define(1003, p['LxM1']+p['rM1']*math.sin(p['thetaMRad']/2.0),  p['Y0M']-p['LyM']+p['rM1']*(1.-math.cos(p['thetaMRad']/2.0)))
    p1004 = pset.define(1004, p['LxM1']+p['rM1']*math.sin(p['thetaMRad']),      p['Y0M']-p['LyM']+p['rM1']*(1.-math.cos(p['thetaMRad'])))
    p1005 = pset.define(1005, p['LxM2']-p['rM2']*math.sin(p['thetaMRad']),      p['Y0M']-p['rM2']*(1.-math.cos(p['thetaMRad'])))
    p1006 = pset.define(1006, p['LxM2']-p['rM2']*math.sin(p['thetaMRad']/2.0),  p['Y0M']-p['rM2']*(1.-math.cos(p['thetaMRad']/2.)))
    p1007 = pset.define(1007, p['LxM2'],                                        p['Y0M'])
    p1008 = pset.define(1008, p['LxM3'],                                        p['Y0M'])

    # curves
    cset = geometry.getCurveSet()
    c1001 = cset.add(Line(1001, p1001, p1002))
    c1002 = cset.add(Arc( 1002, p1002, p1003, p1004))
    c1003 = cset.add(Line(1003, p1004, p1005))
    c1004 = cset.add(Arc( 1004, p1005, p1006, p1007))
    c1005 = cset.add(Line(1005, p1007, p1008))

    wset    = geometry.getWireSet()
    #w1001   = wset.add(Wire(1001,[c1001, c1002, c1003, c1004]))
    w1001   = wset.add(Wire(1001,[c1005, c1004, c1003, c1002, c1001]))
    sidset  = geometry.getSideSet()
    sid1001 = sidset.add(Side(1001,[w1001, ]))

    # Gomtrie des outils : poincon
    p2001 = pset.define(2001, -1.0,                                                  p['Y0P'])
    p2002 = pset.define(2002, p['LxP1'],                                             p['Y0P'])
    p2003 = pset.define(2003, p['LxP1']+p['rP1']*math.sin(p['thetaPRad']/2.0),       p['Y0P']+p['rP1']*(1.-math.cos(p['thetaPRad']/2.0)))
    p2004 = pset.define(2004, p['LxP1']+p['rP1']*math.sin(p['thetaPRad']),           p['Y0P']+p['rP1']*(1.-math.cos(p['thetaPRad'])) )
    #p2005 = pset.define(2005, p['LxP1']+p['rP1']*math.sin(p['thetaPRad'])+p['LyP']*math.cos(p['thetaPRad']),       p['Y0P']+p['rP1']*(1.-math.cos(p['thetaPRad']))+p['LyP']*math.sin(p['thetaPRad']))
    p2005 = pset.define(2005, p['LxP2']-p['rP2']*math.sin(p['thetaPRad']),           p['Y0P']+p['LyP']-p['rP2']*(1.-math.cos(p['thetaPRad'])))
    p2006 = pset.define(2006, p['LxP2']-p['rP2']*math.sin(p['thetaPRad']/2.0),       p['Y0P']+p['LyP']-p['rP2']*(1.-math.cos(p['thetaPRad']/2.)))
    p2007 = pset.define(2007, p['LxP2'],                                             p['Y0P']+p['LyP'])
    p2008 = pset.define(2008, p['LxP2']+p['rP3']*math.sin(math.pi/4.0),              p['Y0P']+p['LyP']+p['rP3']*(1.-math.cos(math.pi/4.0)))
    p2009 = pset.define(2009, p['LxP2']+p['rP3']*math.sin(math.pi/2.0),              p['Y0P']+p['LyP']+p['rP3']*(1.-math.cos(math.pi/2.0)))


    # curves
    cset = geometry.getCurveSet()
    c2001 = cset.add(Line(2001, p2001, p2002))
    c2002 = cset.add(Arc( 2002, p2002, p2003, p2004))
    c2003 = cset.add(Line(2003, p2004, p2005))
    c2004 = cset.add(Arc( 2004, p2005, p2006, p2007))
    c2005 = cset.add(Arc( 2005, p2007, p2008, p2009))
    wset    = geometry.getWireSet()
    w2001   = wset.add(Wire(2001,[c2001, c2002, c2003, c2004, c2005]))
    sidset  = geometry.getSideSet()
    sid2001 = sidset.add(Side(2001,[w2001, ]))

    # Gomtrie des outils : Serre-Flanc

    p3001 = pset.define(3001, p['X0SF'],                                                 p['Y0SF'])
    p3002 = pset.define(3002, p['X0SF']+p['rSF']*(1.0-math.cos(math.pi/3.)),             p['Y0SF']-p['rSF']*math.sin(math.pi/3.))
    p3003 = pset.define(3003, p['X0SF']+p['rSF'],                                        p['Y0SF']-p['rSF'])
    p3004 = pset.define(3004, p['X0SF']+p['LxSF']-p['rSF'],                              p['Y0SF']-p['rSF'])
    p3005 = pset.define(3005, p['X0SF']+p['LxSF']-p['rSF']*(1.0-math.cos(math.pi/3.)),   p['Y0SF']-p['rSF']*math.sin(math.pi/3.))
    p3006 = pset.define(3006, p['X0SF']+p['LxSF'],                                       p['Y0SF'])

    # curves
    cset = geometry.getCurveSet()
    c3001 = cset.add(Arc(3001, p3001, p3002, p3003))
    c3002 = cset.add(Line(3002, p3003, p3004))
    c3003 = cset.add(Arc( 3003, p3004, p3005, p3006))
    wset    = geometry.getWireSet()
    w3001   = wset.add(Wire(3001,[c3001, c3002, c3003]))
    sidset  = geometry.getSideSet()
    sid3001 = sidset.add(Side(3001,[w3001, ]))

    # == SCHEMA D'INTEGRATION ======================================================================
    # Schma d'intgration  0 : QS / 1 : Expl / 2 : Implicite dynamique
    # Time Integration
    if p['ndyn'] == 0 :
        ti=QuasiStaticTimeIntegration(metafor)
        metafor.setTimeIntegration(ti)
    elif p['ndyn'] == 2 :
        ti = AlphaGeneralizedTimeIntegration(metafor)
        metafor.setTimeIntegration(ti)
    else:
        raise Exception("Unknown time integration")
    mim = metafor.getMechanicalIterationManager()
    mim.setResidualComputationMethod(Method4ResidualComputation(100.0))
    mim.setResidualTolerance(p['prec'] )
    mim.setMaxNbOfIterations(10)
    #mim.setVerbose()


    tsm = metafor.getTimeStepManager()
    stages = metafor.getStageManager()
    tsm.setInitialTime(0.0, 0.01)
    tsm.setNextTime(0.1, 1, 0.01)
    tsm.setNextTime(1., 1, 0.01)
    tsm.setNextTime(2., 1, 0.1)
    tsm.setNextTime(3., 1, 0.01)
    tsm.setNextTime(10., 1, 0.1)

    # == Matriaux ==
    materials = domain.getMaterialSet()
    mat1 = materials.define (1, EvpIsoHHypoMaterial)
    mat1.put(MASS_DENSITY,     p['rho']) # tonne/mm^3
    mat1.put(ELASTIC_MODULUS, p['Young']) # MPa
    mat1.put(POISSON_RATIO,   p['poisson'])
    mat1.put(YIELD_NUM,     1)
    matLawSet = domain.getMaterialLawSet()
    matlaw = matLawSet.define(1, KrupkowskyIsotropicHardening)
    matlaw.put(IH_EVPL0, p['ih_evpl0'])
    matlaw.put(IH_K,     p['ih_k'])
    matlaw.put(IH_N,     p['ih_n'])

    # Elements
    prp1 = ElementProperties(Volume2DElement)
    prp1.put(MATERIAL, 1)
    prp1.put(CAUCHYMECHVOLINTMETH , p['cmvim'])
    if p['cmvim'] == VES_CMVIM_EAS:
        prp1.put(EASS, 2)
        prp1.put(EASV, 2)
    app = FieldApplicator(1)
    app.push(sidset(1))
    app.addProperty(prp1)
    domain.getInteractionSet().add(app)
    '''
    # Elements
    mat2 = materials.define (2, DampingMaterial)
    mat2.put(SPRING_FC,     0.0) #.0001)

    prp2 = ElementProperties(Damping2DElement)
    prp2.put(MATERIAL, 2)
    prp2.put(STIFFMETHOD, STIFF_NUMERIC)
    app2 = FieldApplicator(2)
    app2.push(sidset(1))
    app2.addProperty(prp2)
    domain.getInteractionSet().add(app2)

    app2.deactivate(stages[0])
    app2.activate(stages[4])
    '''

    # == FIXATION ==================================================================================
    fctM = PieceWiseLinearFunction()
    fctM.setData(0.0, 0.0)
    fctM.setData(1.0, 0.0)
    fctM.setData(2.0, 0.0)
    fctM.setData(3.0, 0.0)

    fctP = PieceWiseLinearFunction()
    fctP.setData(0.0, -0.01)
    fctP.setData(0.1, -0.01)
    fctP.setData(1.0, 1.0)
    fctP.setData(2.0, 0.0)
    fctP.setData(3.0, 0.0)


    inicond = metafor.getInitialConditionSet()
    loads = domain.getLoadingSet()
    loads.define(cset(4),Field1D(TX,RE))
    fP1 = loads.define(pset(1),Field1D(TY,RE))
    fP1.deactivate(stages[0])
    fP1.activate(stages[3])


    loads.define(w1001,Field1D(TY,RE), -p['ddyM'], fctM , TOTAL_LOAD)
    loads.define(w2001,Field1D(TY,RE), -p['ddyP'], fctP , TOTAL_LOAD)

    p['pilotF'] = True
    if p['pilotF'] :

        mesher = RigidToolMesher(w3001)
        mesher.execute()
        loads.define(w3001,Field1D(TX,RE))

        fctSF = PieceWiseLinearFunction()
        fctSF.setData(0.0,  0.0)
        fctSF.setData(0.1,  1.0)
        fctSF.setData(1.0,  1.0)
        fctSF.setData(1.5,  1.0)
        fctSF.setData(2.0,  1.0)
        fctSF.setData(3.0,  1.0)

        TYSFLoad = loads.define(w3001,Field1D(TY,RE), -p['ddySF'],fctSF, INCREMENTAL_LOAD)
        TYSFLoad.deactivate(stages[0])
        TYSFLoad.activate(stages[2])

        fctForceSF = PieceWiseLinearFunction()
        fctForceSF.setData(0.0, 0.0)
        fctForceSF.setData(0.1, 1.0)
        fctForceSF.setData(1.0, 1.0)
        fctForceSF.setData(2.0, 1.0)
        fctForceSF.setData(3.0, 1.0)

        GYFSLoad = loads.define(w3001,Field1D(TY,GF2), p['forceSF'], fctForceSF, TOTAL_LOAD)# force driven

        #GYFSLoad.deactivate(stages[0])
        #GYFSLoad.activate(stages[1])
        GYFSLoad.deactivate(stages[2])

    else :
        fctSF = PieceWiseLinearFunction()
        fctSF.setData(0.0, 0.0)
        fctSF.setData(0.1, 1.0)
        fctSF.setData(1.0, 1.0)
        fctSF.setData(2.0, 1.0)
        fctSF.setData(3.0, 1.0)
        TYSFLoad = loads.define(w3001,Field1D(TY,RE), -p['ddySF'],fctSF, TOTAL_LOAD)

    # == Contact ==

    fctPeno = PieceWiseLinearFunction()
    fctPeno.setData(0.0, 0.5)
    fctPeno.setData(0.1, 1.0)
    fctPeno.setData(1.0, 1.0)
    fctPeno.setData(2.0, 1.0)
    fctPeno.setData(3.0, 0.0)
    fctPeno.setData(4.0, 0.0)

    mat1001 = materials.define(1001, CoulombContactMaterial)
    mat1001.put(PEN_NORMALE, p['peno'])
    mat1001.depend(PEN_NORMALE, fctPeno, Field(TM))
    mat1001.put(PEN_TANGENT, p['peno']*p['mu'])
    mat1001.depend(PEN_TANGENT, fctPeno, Field(TM))
    mat1001.put(COEF_FROT_STA, p['mu'])
    mat1001.put(COEF_FROT_DYN, p['mu'])
    mat1001.put(PROF_CONT, 0.95*p['rM2'] )

    prp_c1 = ElementProperties(Contact2DElement)
    prp_c1.put(MATERIAL, 1001)
    prp_c1.put(AREAINCONTACT,AIC_ONCE)


    fctPeno2 = PieceWiseLinearFunction()
    fctPeno2.setData(0.0, 0.5)
    fctPeno2.setData(0.1, 1.0)
    fctPeno2.setData(2.0, 1.0)
    fctPeno2.setData(2.5, 0.01)
    fctPeno2.setData(3.0, 0.0)
    fctPeno2.setData(4.0, 0.0)

    mat1002 = materials.define(1002, CoulombContactMaterial)
    mat1002.put(PEN_NORMALE, p['peno'])
    mat1002.depend(PEN_NORMALE, fctPeno2, Field(TM))
    mat1002.put(PEN_TANGENT, p['peno']*p['mu'])
    mat1002.depend(PEN_TANGENT, fctPeno, Field(TM))
    mat1002.put(COEF_FROT_STA, p['mu'])
    mat1002.put(COEF_FROT_DYN, p['mu'])
    mat1002.put(PROF_CONT, 0.95*p['rSF'] )
    prp_c2 = ElementProperties(Contact2DElement)
    prp_c2.put(MATERIAL, 1002)
    prp_c2.put(AREAINCONTACT,AIC_ONCE)

    fctPeno3 = PieceWiseLinearFunction()
    fctPeno3.setData(0.0, 0.5)
    fctPeno3.setData(0.1, 1.0)
    fctPeno3.setData(2.0, 1.0)
    fctPeno3.setData(2.5, 0.0)
    fctPeno3.setData(3.0, 0.0)
    fctPeno3.setData(4.0, 0.0)

    mat1003 = materials.define(1003, CoulombContactMaterial)
    mat1003.put(PEN_NORMALE, p['peno'])
    mat1003.depend(PEN_NORMALE, fctPeno3, Field(TM))
    mat1003.put(PEN_TANGENT, p['peno']*p['mu'])
    mat1003.depend(PEN_TANGENT, fctPeno, Field(TM))
    mat1003.put(COEF_FROT_STA, p['mu'])
    mat1003.put(COEF_FROT_DYN, p['mu'])
    mat1003.put(PROF_CONT, 0.95*p['rP3'] )
    prp_c3 = ElementProperties(Contact2DElement)
    prp_c3.put(MATERIAL, 1003)
    prp_c3.put(AREAINCONTACT,AIC_ONCE)


    #Outil 1 : Matrice

    ci1 = RdContactInteraction(1001)
    ci1.setTool(w1001)
    ci1.push(cset(1))
    ci1.addProperty(prp_c1)
    domain.getInteractionSet().add(ci1)
    #Outil 2 : Poincon
    ci2 = RdContactInteraction(2001)
    ci2.setTool(w2001)
    ci2.push(cset(3))
    ci2.addProperty(prp_c2)
    domain.getInteractionSet().add(ci2)
    #Outil 3 : Serre-Flanc
    if p['pilotF'] :
        ci3 = FdRdContactInteraction(3001)
    else :
        ci3 = RdContactInteraction(3001)
    ci3.setTool(w3001)
    ci3.push(cset(3))
    ci3.addProperty(prp_c3)
    domain.getInteractionSet().add(ci3)


    ci1.deactivate(stages[4])
    ci2.deactivate(stages[4])
    ci3.deactivate(stages[4])

    valuesmanager = metafor.getValuesManager()
    valuesmanager.add(1, MiscValueExtractor(metafor,EXT_T),'time')
    valuesmanager.add(2, IFNodalValueExtractor(pset(1), IF_EVMS), 'evms_p1')
    valuesmanager.add(11, DbNodalValueExtractor(pset(2), Field1D(TX,RE)), 'TX_p2')
    valuesmanager.add(12, DbNodalValueExtractor(pset(2), Field1D(TY,RE)), 'TY_p2')
    valuesmanager.add(13, DbNodalValueExtractor(pset(2), Field1D(TX,GV)), 'VX_p2')
    valuesmanager.add(14, DbNodalValueExtractor(pset(2), Field1D(TY,GV)), 'VY_p2')

    valuesmanager.add(16, DbNodalValueExtractor(pset(2001), Field1D(TY,RE)), 'TY_Poincon')
    valuesmanager.add(17, DbNodalValueExtractor(pset(3001), Field1D(TY,RE)), 'TY_SerreFlanc')


    objFSet = metafor.getObjectiveFunctionSet()
    objFSet.add(SpringBackObjF(1, metafor, 11, 12, 1, 2.0, 10.0) )
    objFSet.add(LastVelocity(2, metafor, 13, 14, 1, 10.0) )

    testSuite = metafor.getTestSuiteChecker()
    testSuite.checkObjectiveFunction(1)
    testSuite.checkObjectiveFunction(2)

#    valuesmanager.add(21, InteractionValueExtractor(ci1, TX, GEN_EXT_FORC), 'FX_Matrice')
#    valuesmanager.add(22, InteractionValueExtractor(ci1, TY, GEN_EXT_FORC), 'FY_Matrice')
#    valuesmanager.add(23, InteractionValueExtractor(ci2, TX, GEN_EXT_FORC), 'FX_Poincon')
#    valuesmanager.add(24, InteractionValueExtractor(ci2, TY, GEN_EXT_FORC), 'FY_Poincon')
#    valuesmanager.add(25, InteractionValueExtractor(ci3, TX, GEN_EXT_FORC), 'FX_SerreFlan')
#    valuesmanager.add(26, InteractionValueExtractor(ci3, TY, GEN_EXT_FORC), 'FY_SerreFlan')
    valuesmanager.add(21, ContactForceValueExtractor(ci1, TX, TOTAL_FORCE), SumOperator(), 'FX_Matrice')
    valuesmanager.add(22, ContactForceValueExtractor(ci1, TY, TOTAL_FORCE), SumOperator(), 'FY_Matrice')
    valuesmanager.add(23, ContactForceValueExtractor(ci2, TX, TOTAL_FORCE), SumOperator(), 'FX_Poincon')
    valuesmanager.add(24, ContactForceValueExtractor(ci2, TY, TOTAL_FORCE), SumOperator(), 'FY_Poincon')
    valuesmanager.add(25, ContactForceValueExtractor(ci3, TX, TOTAL_FORCE), SumOperator(), 'FX_SerreFlan')
    valuesmanager.add(26, ContactForceValueExtractor(ci3, TY, TOTAL_FORCE), SumOperator(), 'FY_SerreFlan')

    valuesmanager.add(31, NormalGapValueExtractor(ci1), MaxOperator(), 'GapMaxMatrice')
    valuesmanager.add(32, NormalGapValueExtractor(ci2), MaxOperator(), 'GapMaxPoincon')
    valuesmanager.add(33, NormalGapValueExtractor(ci3), MaxOperator(), 'GapMaxSerreFlanc')

    cur1 = VectorDataCurve(1, valuesmanager.getDataVector(1), valuesmanager.getDataVector(11),valuesmanager.getDataVector(11).getName())
    cur2 = VectorDataCurve(2, valuesmanager.getDataVector(1), valuesmanager.getDataVector(12),valuesmanager.getDataVector(12).getName())
    cur3 = VectorDataCurve(3, valuesmanager.getDataVector(1), valuesmanager.getDataVector(16),valuesmanager.getDataVector(16).getName())
    cur4 = VectorDataCurve(4, valuesmanager.getDataVector(1), valuesmanager.getDataVector(17),valuesmanager.getDataVector(17).getName())

    cur11 = VectorDataCurve(1, valuesmanager.getDataVector(1), valuesmanager.getDataVector(21), valuesmanager.getDataVector(21).getName())
    cur12 = VectorDataCurve(2, valuesmanager.getDataVector(1), valuesmanager.getDataVector(22), valuesmanager.getDataVector(22).getName())
    cur13 = VectorDataCurve(3, valuesmanager.getDataVector(1), valuesmanager.getDataVector(23), valuesmanager.getDataVector(23).getName())
    cur14 = VectorDataCurve(4, valuesmanager.getDataVector(1), valuesmanager.getDataVector(24), valuesmanager.getDataVector(24).getName())
    cur15 = VectorDataCurve(5, valuesmanager.getDataVector(1), valuesmanager.getDataVector(25), valuesmanager.getDataVector(25).getName())
    cur16 = VectorDataCurve(6, valuesmanager.getDataVector(1), valuesmanager.getDataVector(26), valuesmanager.getDataVector(26).getName())

    cur31 = VectorDataCurve(31, valuesmanager.getDataVector(1), valuesmanager.getDataVector(31),valuesmanager.getDataVector(31).getName())
    cur32 = VectorDataCurve(32, valuesmanager.getDataVector(1), valuesmanager.getDataVector(32),valuesmanager.getDataVector(32).getName())
    cur33 = VectorDataCurve(33, valuesmanager.getDataVector(1), valuesmanager.getDataVector(33),valuesmanager.getDataVector(33).getName())


    dataCurveSet1 = DataCurveSet()
    dataCurveSet1.add(cur1)
    dataCurveSet1.add(cur2)
    dataCurveSet1.add(cur3)
    dataCurveSet1.add(cur4)

    dataCurveSet2 = DataCurveSet()
    dataCurveSet2.add(cur11)
    dataCurveSet2.add(cur12)
    dataCurveSet2.add(cur13)
    dataCurveSet2.add(cur14)
    dataCurveSet2.add(cur15)
    dataCurveSet2.add(cur16)

    dataCurveSet3 = DataCurveSet()
    dataCurveSet3.add(cur31)
    dataCurveSet3.add(cur32)
    dataCurveSet3.add(cur33)

    try:
        winc1 = VizWin()
        winc1.add(dataCurveSet1)
        metafor.addObserver(winc1)

        winc2 = VizWin()
        winc2.add(dataCurveSet2)
        metafor.addObserver(winc2)

        winc3 = VizWin()
        winc3.add(dataCurveSet3)
        metafor.addObserver(winc3)
    except NameError:
        pass


    solman = metafor.getSolverManager();
    try:
        solman.setSolver(DSSolver());
    except NameError:
        pass

    return metafor
    
    
class SpringBackObjF(PythonObjectiveFunction):
    def __init__(self,_no, _meta, _noPtTXRe, _noPtTYRe, _noTime , _time1, _time2) :   
        print "SpringBackObjF : __init__"
        PythonObjectiveFunction.__init__(self,_no,_meta)  
        self.noPtTXRe  = _noPtTXRe   
        self.noPtTYRe  = _noPtTYRe   
        self.noTime    = _noTime  
        self.time1     = _time1
        self.time2     = _time2
               
        print "SpringBackObjF : __init__ finished"
        
    def __del__(self):
        print "SpringBackObjectiveFunction : __del__"
        print "callToDestructor of SpringBackObjectiveFunction not allowed. Add SpringBackObjectiveFunction.__disown__()"
        raw_input('')
        exit(1)
        
    def compute(self, meta):    
        print "entering SpringBackObjF compute"
        valuesmanager = meta.getValuesManager()
        curTX = VectorDataCurve(1, valuesmanager.getDataVector(self.noTime), valuesmanager.getDataVector(self.noPtTXRe))
        curTY = VectorDataCurve(2, valuesmanager.getDataVector(self.noTime), valuesmanager.getDataVector(self.noPtTYRe))
                
        DX  = curTX.eval(self.time2) - curTX.eval(self.time1)
        DY  = curTY.eval(self.time2) - curTY.eval(self.time1)
        print "TX(%f) = %f \t TX(%f) = %f \t DX = %f "% (self.time1, curTX.eval(self.time1), self.time2, curTX.eval(self.time2), DX)
        print "TY(%f) = %f \t TY(%f) = %f \t DY = %f "% (self.time1, curTY.eval(self.time1), self.time2, curTY.eval(self.time2), DY)
        springBack  = math.sqrt(DX*DX+DY*DY)        
        print "SpringBack = ", springBack


        SpringBackfile = open('SpringBack.ascii','w')
        SpringBackfile.write(str(springBack))
        SpringBackfile.close()

        return springBack    
         
class LastVelocity(PythonObjectiveFunction):
    def __init__(self,_no, _meta, _noPtTXGV, _noPtTYGV, _noTime, _time) :   
        print "LastVelocity : __init__"
        PythonObjectiveFunction.__init__(self,_no,_meta)  
        self.noPtTXGV  = _noPtTXGV   
        self.noPtTYGV  = _noPtTYGV  
        self.noTime    = _noTime  
        self.time      = _time
               
        print "LastVelocity : __init__ finished"
        
    def __del__(self):
        print "LastVelocity : __del__"
        print "callToDestructor of LastVelocity not allowed. Add LastVelocity.__disown__()"
        raw_input('')
        exit(1)
        
    def compute(self, meta):    
        print "entering LastVelocity compute"
        valuesmanager = meta.getValuesManager()
        curVX = VectorDataCurve(1, valuesmanager.getDataVector(self.noTime), valuesmanager.getDataVector(self.noPtTXGV))
        curVY = VectorDataCurve(2, valuesmanager.getDataVector(self.noTime), valuesmanager.getDataVector(self.noPtTYGV))
                
        VX  = curVX.eval(self.time)
        VY  = curVY.eval(self.time)
        print "VX(%f) = %f "% (self.time, VX)
        print "VY(%f) = %f "% (self.time, VY)
        Veloc  = math.sqrt(VX*VX+VY*VY)        
        print "Veloc = ", Veloc
        return Veloc    
        
