/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wFpeLSFunction.h"
using namespace tbox;
using namespace flow;

FpeLSFunction::FpeLSFunction(flow::Newton &_solver, std::vector<double> &_u,
                             std::vector<double> &_u0, std::vector<double> &_du,
                             std::vector<double> &_fRes) : LSFunction(),
                                                           solver(_solver), u(_u), u0(_u0), du(_du), fRes(_fRes)
{
}

/**
 * @brief Update initial and change in solution 
 * @authors Adrien Crovato
 */
void FpeLSFunction::update(std::vector<double> &_u0, std::vector<double> &_du)
{
    u0 = _u0;
    du = _du;
}

/**
 * @brief Evaluate residual
 * @authors Adrien Crovato
 */
double FpeLSFunction::eval(double alpha)
{
    double resNorm = 0.;
    gmm::add(u0, gmm::scaled(du, alpha), u);
    gmm::clear(fRes);
    solver.buildRes(fRes);
    resNorm = gmm::vect_norm2(fRes);
    return resNorm;
}

void FpeLSFunction::write(std::ostream &out) const
{
    out << "flow::FpeLSFunction " << std::endl;
}
