#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @brief Boundary Layer Equations viscous solver
# @authors Adrien Crovato, Amaury Bilocq

import numpy as np

class Solver():
    def __init__(self, _boundary):
        '''...
        '''
        # TODO generalize to handle several boundaries (use python array)
        self.boundary = _boundary # boundary on which BLE are sovled
        self.nN = len(self.boundary.nodes) # number of nodes
        self.nE = len(self.boundary.groups[0].tag.elems) # number of elements
        self.v = np.zeros((self.nN, 3)) # velocity at edge of BL
        self.u = np.zeros(self.nE) # blowing velocity

    def run(self):
        ''' Solve BLE
        '''
        print '--- I am a fake BL solver and I am setting dummy blowing velocities! ---'
        # 6th order fitting of delta_star for naca 0012 at alpha = 0, from XFoil
        a = 0.000054925976379640116
        b = 0.007407077078697043
        c = -0.043918316829113194
        d = 0.1487463222137225
        e = -0.20162146328539832
        f = 0.0880698637668166
        g = 0.004436812366681563
        for i in range(0, self.nE):
            x = 0.5*(self.boundary.groups[0].tag.elems[i].nodes[0].pos.x[0]+self.boundary.groups[0].tag.elems[i].nodes[1].pos.x[0])
            # CAUTION: a positive velocity will result in a suction!
            self.u[i] = -(b + 2*c*x + 3*d*x*x* + 4*e*x*x*x + 5*f*x*x*x*x + 6*g*x*x*x*x*x) # ue = d(delta_star) / dx
