#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 1er test multiscale (sans MPI)

import heat as h
import tbox
import tbox.gmsh as gmsh
import math

class MicroModel(tbox.Fct2UdU):
    def __init__(self):
        tbox.Fct2UdU.__init__(self)
        print "[m] creation of the micro model"

        self.msh = gmsh.MeshLoader("heat_ms_micro.geo", __file__).execute()
        self.pbl = h.Problem(self.msh)
    
        self.solver = h.Solver(self.pbl)
        self.solver.nthreads = 1
        self.solver.save = False
        self.solver.verbose = 0
        
        h.Medium(self.pbl, "Matrix",    tbox.Fct2C(1.,1.,0), 1.0)
        h.Medium(self.pbl, "Inclusion", tbox.Fct2C(100.,100.,0), 1.0)
        self.p = h.Periodic(self.pbl, 0.0, 0.0, 0.0)    
       
    def eval(self, elem, npg, u, gradu, k, fake):           
    #def eval(self, u, gradu, k, fake):
        verb=False
        
        TM   = u
        dTxM = gradu[0]
        dTyM = gradu[1]
        dTnorm = math.sqrt(dTxM*dTxM+dTyM*dTyM)
        if dTnorm==0.0:
            dTxM = 1.0
            dTyM = 1.0
            dTnorm=math.sqrt(dTxM*dTxM+dTyM*dTyM)
            
        if verb: 
            print "-"*10
            print "**solving micro problem..."
            print "[m] TM   =",TM
            print "[m] dTxM =",dTxM
            print "[m] dTyM =",dTyM
            
        #if dTxM==0: dTxM=1e-5
        #if dTyM==0: dTyM=1e-5
      
        self.p.reset(TM, dTxM, dTyM)
        self.solver.start() 
        
        qMx = self.p.qM.x[0]
        qMy = self.p.qM.x[1] 
        
        if verb:
            print "[m] qMx =", qMx 
            print "[m] qMy =", qMy 
        
        dTnorm = math.sqrt(dTxM*dTxM+dTyM*dTyM)
        qnorm  = math.sqrt(qMx*qMx+qMy*qMy)
        ampl   = qnorm/dTnorm
        
        ps = -(dTxM*qMx+dTyM*qMy)/dTnorm/qnorm
        if ps>1.0: ps=1.0
        if ps<-1.0: ps=-1.0
        
        #print ps
        ang = math.acos(ps)
        print "ang=", ang, "ampl=", ampl
        #print "qMx=", qMx, "ampl*dT=", dTxM*ampl, "res=", qMx-dTxM*ampl
        cos = math.cos(ang)
        sin = math.sin(ang)
        K11 = cos*ampl
        K21 = sin*ampl
        K22 = K11
        K12 = -K21
          
        k[0,0]=K11
        k[1,1]=K22 
        k[1,0]=K21
        k[0,1]=K12             

        if verb:
            print k

def main():

    msh = gmsh.MeshLoader("heat_ms_macro.geo", __file__).execute()
    pbl = h.Problem(msh)

    solver = h.Solver(pbl)
    solver.nthreads = 1
    
    myk = MicroModel()
    h.Medium(pbl, "S1", myk)
    
    h.Dirichlet(pbl,"Left", 0.)
    h.Dirichlet(pbl,"Right", 10.)

    solver.restol = 1e-6
    solver.start()

    from heat.viewer import GUI
    GUI().open('heat_ms_macro')
    
if __name__ == "__main__":
    main()
