#!/usr/bin/env python
# -*- coding: utf-8 -*-
# test multiscale - version "task pull"
# exc:
# mpiexec.openmpi --bycore --bind-to-core -n 6 ./run.py heat/tests/jalil2_taskpull.py

import heat as h
import tbox
import tbox.gmsh as gmsh
import heat.fe2 as fe2

class HoneyComb:
    def __init__(self):
        self.msh = gmsh.MeshLoader("jalil2_honey.geo", __file__).execute()
        self.pbl = h.Problem(self.msh)
    
        self.solver = h.Solver(self.pbl)
        self.solver.nthreads = 1
        self.solver.save = False
        self.solver.verbose = 0
        
        self.fa=tbox.PwLf()
        self.fa.add(0, 80)
        self.fa.add(300, 80)
        self.fa.add(350, 20)
        self.fa.add(450, 10)
        self.fa.add(550, 5)
        self.fa.add(750, 1)
        self.fa.add(1000, 1)
        
        self.fs=tbox.PwLf()
        self.fs.add(0, 140)
        self.fs.add(300, 140)
        self.fs.add(350, 180)
        self.fs.add(450, 200)
        self.fs.add(550, 210)
        self.fs.add(750, 220)
        self.fs.add(1000, 220)
        
        self.fz = tbox.PwLf()
        self.fz.add(0,0) 
           
        h.Medium(self.pbl, "Air", tbox.Fct2PwLf(self.fa,self.fa,self.fz), 1.)
        h.Medium(self.pbl, "Skeleton", tbox.Fct2PwLf(self.fs,self.fs,self.fz), 1.) 
           
        self.p = h.Periodic(self.pbl, 0.0, 0.0, 0.0) 


class Lshape:
    def __init__(self, law):
        self.msh = gmsh.MeshLoader("jalil1_corner.geo", __file__).execute()
        self.pbl = h.Problem(self.msh)

        self.solver = h.Solver(self.pbl)
        self.solver.nthreads = 1

        h.Medium(self.pbl, "Volume", law)
        
        h.Dirichlet(self.pbl,"Top BC", 400.)
        h.Dirichlet(self.pbl,"Right BC", 600.)

        self.solver.restol = 1e-6



if __name__ == "__main__":

    micro = HoneyComb()

    if fe2.rank==0:
        law = fe2.FE2(micro)
        macro = Lshape(law)
        master = fe2.Master(macro)
        master.start()
    else:
        worker = fe2.Worker(micro)
        worker.start()

    


