/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WCOMPILEDFCT_H
#define WCOMPILEDFCT_H

#include "heat.h"
#include "wFct0.h"
#include "wFct1.h"
#include "wFct2.h"
#include "wPt.h"
#include "gmm_gmm.h"

using namespace tbox;

namespace heat
{

/**
 * @brief cfr heat_vark.py (version scalaire)
 */

class HEAT_API CompiledFct1a : public Fct0XYZ
{
    double k1, k2;

public:
    CompiledFct1a(double _k1, double _k2) : k1(_k1), k2(_k2) {}

    virtual double eval(Pt const &pos) const override
    {
        double x = pos.x[0];
        double y = pos.x[1];
        double L = 1;
        return (k1 + k2) / 2 + fabs(k2 - k1) / 2 * sin(2 * M_PI / L * 4 * (x + y / 2));
    }
};

/**
 * @brief cfr heat_vark.py (version matricielle)
 */

class HEAT_API CompiledFct1b : public Fct2XYZ
{
    double k1, k2;

public:
    CompiledFct1b(double _k1, double _k2) : k1(_k1), k2(_k2) {}

    virtual void eval(Pt const &pos, gmm::dense_matrix<double> &out, bool fake) const override
    {
        double x = pos.x[0];
        double y = pos.x[1];
        double L = 1;
        out(1, 1) = out(0, 0) = (k1 + k2) / 2 + fabs(k2 - k1) / 2 * sin(2 * M_PI / L * 4 * (x + y / 2));
        out(0, 1) = out(1, 0) = 0.0;
    }
};

/**
 * @brief cfr heat_nl.py (version scalaire)
 */
class HEAT_API CompiledFct2a : public Fct0U
{
public:
    CompiledFct2a() {}

    virtual double eval(double u) const override
    {
        return 10 + u * u;
    }
};

/**
 * @brief cfr heat_nl.py (version matricielle)
 */
class HEAT_API CompiledFct2b : public Fct2U
{
public:
    CompiledFct2b() {}

    virtual void eval(double u, gmm::dense_matrix<double> &out, bool fake) const override
    {
        out(0, 0) = out(1, 1) = 10 + u * u;
        out(0, 1) = out(1, 0) = u * u;
    }
};

}; // namespace heat

#endif //WCOMPILEDFCT_H
