/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wProblem.h"
#include "wMedium.h"
#include "wBoundary.h"
#include "wSource.h"
#include "wPeriodic.h" // sinon:  warning C4150: deletion of pointer to incomplete type 'heat::Periodic'; no destructor called
#include "wElement.h"
#include "wMem.h"
#include "wTag.h"
using namespace heat;

Problem::Problem(std::shared_ptr<MshData> _msh) : msh(_msh), pdic(NULL)
{
}

void Problem::write(std::ostream &out) const
{
    out << "heat::Problem:\n";
}

Problem::~Problem()
{
    std::cout << "~Problem()\n";
}

/**
 * @brief Update the elements memory (Jacobian)
 * @authors Adrien Crovato
 */
void Problem::updateMem()
{
    // Update volume Jacobian
    for (auto vol : media)
        for (auto e : vol->tag->elems)
            e->getVMem().update(true);
    // Update volume Jacobian (volume source)
    for (auto vol : srcs)
        for (auto e : vol->tag->elems)
            e->getVMem().update(true);
    // Update surface Jacobian (Neumann B.C.)
    for (auto surf : bnds)
        for (auto e : surf->tag->elems)
            e->getVMem().update(false);
}