/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WUDIRICHLET_H
#define WUDIRICHLET_H

#include "mirrors.h"
#include "wGroup.h"
#include "wPt.h"
#include <vector>
#include <string>

using namespace tbox;

namespace mirrors
{

/**
 * @brief manage mechanical Dirichlet BC
 */

class MIRRORS_API uDirichlet : public Group
{
public:
    std::string BC_name;
    std::vector<Node *> nodes;
    Pt which_dof;
    Pt values;

    uDirichlet(Problem &pbl, int no, std::string _BC_name, Pt _which_dof, Pt _values);
    uDirichlet(Problem &pbl, std::string const &name, std::string _BC_name, Pt _which_dof, Pt _values);
    //virtual void apply(std::vector<double> &rhs);

#ifndef SWIG
    virtual void write(std::ostream &out) const override;
#endif
};

} // namespace mirrors

#endif //WUDIRICHLET_H
