#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os, os.path

class GUI:
    def __init__(self):
        pass
    def open(self, gmshfile):
        pyprefix = os.path.splitext(gmshfile)[0] # could be in python tree (.geo or .msh)
        wkprefix = os.path.basename(pyprefix)   # in current directory (workspace)
        
        # check .msh
        if os.path.splitext(gmshfile)[1]=='.msh':
            mshfile = gmshfile
        else:
            mshfile = wkprefix+'.msh'
        if not os.path.isfile(mshfile):
            raise Exception("file not found %s" % mshfile)
            
        # check results
        posfile = wkprefix+".pos"
        if not os.path.isfile(posfile):
            raise Exception("file not found %s" % posfile)
            
        # check options
        optfile1 = pyprefix+".opt"
        optfile2 = wkprefix+".opt"
        if os.path.isfile(optfile1):
            optfile=optfile1
        else:
            if not os.path.isfile(optfile2):
                self.defopt(optfile2)
            optfile=optfile2
            
        cmd="gmsh %s %s %s" % (mshfile, posfile, optfile)
        print "cmd=",cmd
        os.system(cmd)        

    def defopt(self,fname='default.opt'):
        f = open(fname,'w')
        f.write("General.Orthographic = 0;\n")
        f.write("General.RotationX = 25;\n")
        f.write("General.RotationY = 45;\n")
        f.write("General.RotationZ = 0;\n")
        f.write("Mesh.SurfaceEdges = 0;\n")
        f.write("Mesh.VolumeEdges = 0;\n")
        
        for i in range(0,60):
            f.write("View[%d].Visible = 0;\n" % i)
            f.write("View[%d].DrawQuadrangles = 0;\n" % i)
            f.write("View[%d].DrawTriangles = 0;\n" % i)
            f.write("View[%d].IntervalsType = 3;\n" % i)
        f.write("View[0].Visible = 1;\n")
        f.close()
    
