#!/usr/bin/env python
# -*- coding: latin-1; -*-

from mrstlnos.eigenvalues import *
import os
import fwk

import vtk
import tboxVtk
import tboxVtk.reader as vtkR
import tboxVtk.cutter as vtkC

from mrstlnos.CMAME_tests.model import Compute_QoIs

def cut_of_one_case(case_id,n_samples,v_min,v_max,file_dir,at_point=False):
  import matplotlib.pyplot as plt
  
  mshName_mean = file_dir + '/../../../workspace/mrstlnos_CMAME_tests_plate/plate_fused_ast1_s0_0'
  pts, tri, vals = cut_plate(mshName_mean,at_point)
  
  x = pts[:,0]
  y = pts[:,1]
  
  QoIs = np.zeros((4,n_samples))
  
  plot_one_sampel(x,y,tri,vals,at_point)
  plt.savefig('mean_case.png', dpi=400)

  file_MC_base_dir = '/Volumes/HD_SONY/CMAME/Plate/mrstlnos_CMAME_tests_plate_MC_direct_case_' + str(case_id) + '_NER/1/'
  file_MC_base_dir = '/Users/kimliegeois/Desktop/mrstlnos_CMAME_tests_plate_MC_GS_case_' + str(case_id) + '/1/'
  
  directory = 'case_' + str(case_id)
  if not os.path.exists(directory):
      os.makedirs(directory)
      os.makedirs(directory+'/sample')
      os.makedirs(directory+'/abs')
      os.makedirs(directory+'/rel')

  for i in range(0,n_samples):
    input_MC_file_dir = file_MC_base_dir + str(i)+'/'
    mshName_sample = input_MC_file_dir+'plate_fused_ast1_s0_0'

    QoIs[:,i]=Compute_QoIs(mshName_sample)
    
    pts_s, tri_s, vals_s = cut_plate(mshName_sample,at_point)
    plot_one_sampel(x,y,tri,vals_s,at_point)
    plt.savefig(directory+'/sample/'+ str(i) + '.png', dpi=400)
    plot_diff_rel(x,y,tri,vals,vals_s,v_min = v_min, v_max = v_max,at_point=at_point)
    plt.savefig(directory+'/rel/'+ str(i) + '.png', dpi=400)
    plot_diff_abs(x,y,tri,vals,vals_s,at_point)
    plt.savefig(directory+'/abs/'+ str(i) + '.png', dpi=400)
    plt.close('all')
  for i in range(0,4):
    plt.figure()
    plt.hist(QoIs[i,:], bins=80,density=True)
    plt.xlabel('[MPa]')
    plt.savefig('QoI_'+str(i)+'_case_'+str(case_id)+'.png', dpi=400)

    plt.close('all')

def cut_plate(mshName,at_point=False):
  reader = vtkR.Reader()
  reader.open(mshName)
  cutter = vtkC.Cutter(reader.reader.GetOutputPort())
  c_output = cutter.cut(1, [0., 0., 0.5], [0., 0., 1.], tag_name='volume_id')
  if at_point:
    pts, tri, vals = cutter.extract(c_output, 3, ['u','x','y','z','G','von Mises'], atPoint = True)
  else:
    pts, tri, vals = cutter.extract(c_output, 3, ['u','x','y','z'], atPoint = True)
    pts_2, tri_2, vals_2 = cutter.extract(c_output, 3, ['G','von Mises'], atPoint = False)
    vals.update(vals_2)
  
  return pts, tri, vals
    
def plot_2D_cut_values_at_node(x,y,tri,value,n_levels=16):
  import matplotlib.pyplot as plt
  value = value.reshape((len(x),))
  if max(value)-min(value) > 0:
    levels = np.linspace(min(value), max(value), n_levels)
    plt.set_cmap('coolwarm')
    plt.tricontour(x,y, tri, value, levels, linestyles='-',
              colors='black', linewidths=0.5)
    axtri = plt.tricontourf(x,y, tri, value, levels)
  else:
    axtri = plt.tricontourf(x,y, tri, value)
  plt.gca().set_aspect('equal')
  return axtri

def plot_2D_cut_values_at_elem(x,y,tri_1,value,v_min=0,v_max=0,n_levels=16,shading=False):
  import matplotlib.pyplot as plt
  
  value = value.reshape((tri_1.shape[0],))
  if v_min==v_max:
    v_min = min(value)
    v_max = max(value)
  
  levels = np.linspace(v_min, v_max, n_levels)
  plt.set_cmap('coolwarm')
  if shading:
    axtri = plt.tripcolor(x,y, tri_1, value, shading='gouraud')
  else:
    axtri = plt.tripcolor(x,y, tri_1, value, levels)
  plt.gca().set_aspect('equal')
  return axtri

def plot_one_sampel(x,y,tri,vals_s,at_point=False):
  import matplotlib.pyplot as plt
  
  fig = plt.figure(figsize=(15, 4))
  
  plt.subplot(131)
  if not at_point:
    tmp = plot_2D_cut_values_at_elem(x,y,tri,vals_s['G'])
  else:
    tmp = plot_2D_cut_values_at_node(x,y,tri,vals_s['G'])
  fig.colorbar(tmp, ax=plt.gca(), orientation='horizontal')
  plt.title('G')
  
  plt.subplot(132)
  tmp = plot_2D_cut_values_at_node(x,y,tri,vals_s['y'])
  fig.colorbar(tmp, ax=plt.gca(), orientation='horizontal', format='%.0e')
  plt.title('y')
  
  plt.subplot(133)
  if not at_point:
    tmp =plot_2D_cut_values_at_elem(x,y,tri,vals_s['von Mises'])
  else:
    tmp = plot_2D_cut_values_at_node(x,y,tri,vals_s['von Mises'])
  fig.colorbar(tmp, ax=plt.gca(), orientation='horizontal')
  plt.title('von Mises')

def plot_diff_abs(x,y,tri,vals,vals_s,at_point=False):
  import matplotlib.pyplot as plt
  
  fig = plt.figure(figsize=(15, 4))
  
  plt.subplot(131)
  if not at_point:
    tmp = plot_2D_cut_values_at_elem(x,y,tri,vals_s['G']-vals['G'])
  else:
    tmp = plot_2D_cut_values_at_node(x,y,tri,vals_s['G']-vals['G'])
  fig.colorbar(tmp, ax=plt.gca(), orientation='horizontal')
  plt.title('delta G')
  
  plt.subplot(132)
  tmp = plot_2D_cut_values_at_node(x,y,tri,vals_s['y']-vals['y'])
  fig.colorbar(tmp, ax=plt.gca(), orientation='horizontal', format='%.0e')
  plt.title('delta y')
  
  plt.subplot(133)
  if not at_point:
    tmp =plot_2D_cut_values_at_elem(x,y,tri,vals_s['von Mises']-vals['von Mises'])
  else:
    tmp = plot_2D_cut_values_at_node(x,y,tri,vals_s['von Mises']-vals['von Mises'])
  fig.colorbar(tmp, ax=plt.gca(), orientation='horizontal')
  plt.title('delta von Mises')

def plot_diff_rel(x,y,tri,vals,vals_s,v_min = -10,v_max = 10,at_point=False):
  import matplotlib.pyplot as plt

  fig = plt.figure(figsize=(15, 4))
  
  diff_G = vals_s['G']-vals['G']
  diff_G_rel = 100*diff_G/vals['G']
  
  plt.subplot(121)
  if not at_point:
    tmp = plot_2D_cut_values_at_elem(x,y,tri,diff_G_rel,v_min=v_min,v_max=v_max)
  else:
    tmp = plot_2D_cut_values_at_node(x,y,tri,diff_G_rel)
  fig.colorbar(tmp, ax=plt.gca(), orientation='horizontal')
  plt.clim(v_min,v_max)
  plt.title('delta G relative')
  
  diff_vM = vals_s['von Mises']-vals['von Mises']
  diff_vM_rel = 100*diff_vM/vals['von Mises']
  
  plt.subplot(122)
  if not at_point:
    tmp =plot_2D_cut_values_at_elem(x,y,tri,diff_vM_rel,v_min=v_min,v_max=v_max)
  else:
    tmp = plot_2D_cut_values_at_node(x,y,tri,diff_vM_rel)
  fig.colorbar(tmp, ax=plt.gca(), orientation='horizontal')
  plt.clim(v_min,v_max)
  plt.title('delta von Mises relative')

def main():
  import matplotlib.pyplot as plt
  
  file_dir = os.path.dirname(__file__)
  
  n_samples = 96
  cut_of_one_case(1,n_samples,-10,10,file_dir,True)
  #cut_of_one_case(2,n_samples,-40,40,file_dir)
  cut_of_one_case(3,n_samples,-10,10,file_dir,True)
  #cut_of_one_case(4,n_samples,-40,40,file_dir)


if __name__ == "__main__":
  main()
