#!/usr/bin/env python
# -*- coding: latin-1; -*-

from mrstlnos.eigenvalues import *
import os
import fwk

def plot_as_bar_plot(x,y,semilog=False):
    import matplotlib.pyplot as plt
    plt.figure()

    for i in range(0,len(x)):
      if i == 0:
        x_bar = np.array([x[0], 0.5*x[0]+0.5*x[1]])
        y_bar = np.array([y[0], y[0]])
      elif i!= len(x)-1:
        x_bar = np.append( x_bar , [x_bar[-1], 0.5*x[i]+0.5*x[i+1]])
        y_bar = np.append( y_bar , [y[i], y[i]])
      else:
        x_bar = np.append( x_bar , [x_bar[-1], x[i]])
        y_bar = np.append( y_bar , [y[i], y[i]])
        break
    if semilog:
      plt.semilogy(x_bar,y_bar)
    else:
      plt.plot(x_bar,y_bar)

def main():
    import matplotlib.pyplot as plt
    
    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()
    
    display = True
    
    current_n = 1
    n_samples = 32
    
    theta = np.zeros((current_n,n_samples))
    omegas = np.zeros((current_n,n_samples))
    deltas = np.zeros((current_n,n_samples), dtype=np.complex128)
    
    
    
    for i in range(0,n_samples):
      tmp_delta = np.loadtxt(file_dir + '/../../../workspace/mrstlnos_CMAME_tests_plate_deltas_omegas_GS/deltas_GS_GS_'+str(i)+'.txt')
      omegas_c = np.loadtxt(file_dir + '/../../../workspace/mrstlnos_CMAME_tests_plate_deltas_omegas_GS/omegas_GS_GS_'+str(i)+'.txt')
      
      n = tmp_delta.shape[0]
      
      deltas_c = np.zeros((n,), dtype=np.complex128)
    
      deltas_c.real = tmp_delta[:,0]
      deltas_c.imag = tmp_delta[:,1]
      
      index = np.argwhere(np.absolute(deltas_c)>=0.95)
      n = len(index)
      
      deltas_c = deltas_c[index].reshape((n,))
      omegas_c = omegas_c[index].reshape((n,))

      if current_n < n:

        theta_tmp = np.zeros((n,n_samples))
        omegas_tmp = np.zeros((n,n_samples))
        deltas_tmp = np.zeros((n,n_samples), dtype=np.complex128)
        
        theta_tmp[0:current_n,:] = theta
        omegas_tmp[0:current_n,:] = omegas
        deltas_tmp[0:current_n,:] = deltas

        theta = theta_tmp
        omegas = omegas_tmp
        deltas = deltas_tmp
    
        current_n = n
      
      deltas[0:n,i] = deltas_c
      omegas[0:n,i] = omegas_c
      
      theta_c = compute_theta(deltas_c)
      
      theta[0:n,i] = theta_c
      
      N = 720
      theta_discrete, cum_delta, deriv_delta, cum_delta_omega, deriv_delta_omega, percent_delta, percent_delta_omega = compute_cum_delta(theta_c,deltas_c,omegas_c,N)
      
      #plot_as_bar_plot(theta_discrete,deriv_delta_omega)
      
      indices = np.zeros((n,))
      sort_indices = np.argsort(-omegas_c)
      indices[sort_indices[0:30]] = 1
      #plot_eigenvalues_on_unit_circle_highlight(deltas_c,indices)
      #plot_polar_function(theta_discrete,deriv_delta,1.5,2.5)
      #plot_polar_function(theta_discrete,deriv_delta_omega,3,4)
      #plt.title('Sample ' + str(i))
      #plt.savefig('fig_'+str(i)+'.png')
      
      
      #sort_indices = np.argsort(theta_c)
      #plt.figure()
      #plt.bar(theta_c[sort_indices],omegas_c[sort_indices], width=0.1, alpha=0.6, align="center")
      #plt.plot(theta_c[sort_indices],omegas_c[sort_indices])
      
      #tmp = np.ones((len(theta[0:n,i]),2))
      #tmp[:,0] = theta[0:n,i]
      #np.savetxt('theta_'+str(i)+'.txt',tmp)
    
      #tmp = np.ones((len(theta_discrete),2))
      #tmp[:,0] = theta_discrete
      #np.savetxt('theta_discrete_'+str(i)+'.txt',tmp)
      #np.savetxt('deriv_delta_'+str(i)+'.tex',deriv_delta)
      #np.savetxt('deriv_delta_omega_'+str(i)+'.tex',deriv_delta_omega)

    #np.savetxt('omegas_all.txt',omegas)
    
    theta = theta.flatten('F')
    deltas = deltas.flatten('F')
    omegas = omegas.flatten('F')
    
    N = 720
    theta_discrete, cum_delta, deriv_delta, cum_delta_omega, deriv_delta_omega, percent_delta, percent_delta_omega = compute_cum_delta(theta,deltas,omegas,N)
    plot_as_bar_plot(theta_discrete,deriv_delta_omega)
    #sort_indices = np.argsort(theta)
    #plt.figure()
    #plt.bar(theta[sort_indices],omegas[sort_indices], width=0.1, alpha=0.6, align="center")
    
    '''
    ensemble_sizes = np.array([8,16,32])

    for i_s in range(0,len(ensemble_sizes)):
      s = ensemble_sizes[i_s]
      for i in range(0,n_samples,s):
        i_sample_min = i
        i_sample_max = i_sample_min + s
        
        omegas_s = omegas[:,i_sample_min:i_sample_max].flatten('F')
        deltas_s = deltas[:,i_sample_min:i_sample_max].flatten('F')
        theta_s = theta[:,i_sample_min:i_sample_max].flatten('F')
        index = np.argwhere(np.absolute(deltas_s)>=0.9)
        
        omegas_s = omegas_s[index]
        deltas_s = deltas_s[index]
        theta_s = theta_s[index]
        
        N = 1000
        theta_discrete, cum_delta, deriv_delta, cum_delta_omega, deriv_delta_omega, percent_delta, percent_delta_omega = compute_cum_delta(theta_s,deltas_s,omegas_s,N)

        n = len(omegas_s)
        indices = np.zeros((n,))
        sort_indices = np.argsort(-omegas_s)
        indices[sort_indices[0:30]] = 1
        plot_eigenvalues_on_unit_circle_highlight(deltas_s,indices)
        plot_polar_function(theta_discrete,deriv_delta,1.5,2.5)
        plot_polar_function(theta_discrete,deriv_delta_omega,3,4)
        plt.title('Ensemble ' + str(int(i/s)) + ' of ensemble size ' + str(s))
        
        plt.savefig('fig_s_'+ str(s) +'_' + str(int(i/s)) + '.png')
    '''
    if display: plt.show()

if __name__ == "__main__":
    main()
