#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"



    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos
    import numpy as np

    solverList = Teuchos.ParameterList()
    solverList['type']= "BlockGmres"
    solverList['Convergence Tolerance'] = 10**(-7)
    solverList['Maximum Iterations'] = 2000
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    solverList['Use preconditioner'] = True
    
    solverList['Maximum active set iteration'] = 10

    solverList['Print Teuchos timers'] = True
    solverList['Write txt files'] = True
    
    import shlex, subprocess
    import os

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()
    geo_dir = file_dir + '/../mesh'
    
    solverList["use xml file"] = True
    solverList["MueLu xml file name"] = file_dir + '/../smoothers/heat_no_contact_mtgs.xml'
    
    geo_name = 'fused_mirror_assembly.geo'
    mesh_name = 'fused_mirror_assembly.msh'

    if siz > 1:
      # We have to distribute the work among MPI process
      if rank == 0:
          command_line = 'gmsh -3 ' + geo_dir + '/' + geo_name + ' -o ' + work_dir + '/' +  mesh_name +' -part ' + str(siz)
          print command_line
          tmp = shlex.split(command_line)
          fileout = open('gmsh.log','w')
          p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
          retcode = p.wait()
          for i in range(1,siz):
              comm.send(1, dest=i, tag=11)
      else:
          data = comm.recv(source=0, tag=11)
      msh = gmsh.MeshLoader(mesh_name,work_dir).myrank_execute(rank)
    else:
      msh = gmsh.MeshLoader(geo_name,geo_dir).execute()

    
    
    pbl = m.Problem(msh,comm)
    
    # Distance in the mesh are represented in meters

    ensemble_size = 1

    #379 000 MPa = 379 GPa = 379E9 Pa
    E_rhodium = 379E9 * np.ones(ensemble_size)
    nu_rhodium = 0.26 * np.ones(ensemble_size)
    #0.150 kW/(m K) = 150 W/(m K)
    k_rhodium = 150
    d_rhodium = 0.781E-05

    E_WCu = 280E9 * np.ones(ensemble_size)
    nu_WCu = 0.298 * np.ones(ensemble_size)
    k_WCu = 180
    d_WCu = 0.88E-05
    
    E_SS = 200E9 * np.ones(ensemble_size)
    nu_SS = 0.3 * np.ones(ensemble_size)
    k_SS = 15.3
    d_SS = 1.57E-05

    E_CuCrZr = 118E9 * np.ones(ensemble_size)
    nu_CuCrZr = 0.33 * np.ones(ensemble_size)
    k_CuCrZr = 345
    d_CuCrZr = 1.8E-05

    E_I718 = 183E9 * np.ones(ensemble_size)
    nu_I718 = 0.31 * np.ones(ensemble_size)
    k_I718 = 15.8
    d_I718 = 1.38E-05
    
    T_ref = 22.
    

    m.Medium(pbl, "Mirror", "rhodium",E_rhodium,nu_rhodium,k_rhodium,d_rhodium)
    m.Medium(pbl, "Substrate", "WCu",E_WCu,nu_WCu,k_WCu,d_WCu)
    
    m.Medium(pbl, "Holder", "SS",E_SS,nu_SS,k_SS,d_SS)

    m.Medium(pbl, "Washer 1", "I718",E_I718,nu_I718,k_I718,d_I718)
    m.Medium(pbl, "Washer 2", "I718",E_I718,nu_I718,k_I718,d_I718)
    m.Medium(pbl, "Washer 3", "I718",E_I718,nu_I718,k_I718,d_I718)
    
    m.Medium(pbl, "Shaft 1", "I718",E_I718,nu_I718,k_I718,d_I718)
    m.Medium(pbl, "Shaft 2", "I718",E_I718,nu_I718,k_I718,d_I718)
    m.Medium(pbl, "Shaft 3", "I718",E_I718,nu_I718,k_I718,d_I718)
    
    m.Medium(pbl, "Nut 1", "I718",E_I718,nu_I718,k_I718,d_I718)
    m.Medium(pbl, "Nut 2", "I718",E_I718,nu_I718,k_I718,d_I718)
    m.Medium(pbl, "Nut 3", "I718",E_I718,nu_I718,k_I718,d_I718)
    
    m.Medium(pbl, "Spacer 1 part 1", "CuCrZr",E_CuCrZr,nu_CuCrZr,k_CuCrZr,d_CuCrZr)
    m.Medium(pbl, "Spacer 2 part 1", "CuCrZr",E_CuCrZr,nu_CuCrZr,k_CuCrZr,d_CuCrZr)
    m.Medium(pbl, "Spacer 3 part 1", "CuCrZr",E_CuCrZr,nu_CuCrZr,k_CuCrZr,d_CuCrZr)

    m.Medium(pbl, "Spacer 1 part 2", "WCu",E_WCu,nu_WCu,k_WCu,d_WCu)
    m.Medium(pbl, "Spacer 2 part 2", "WCu",E_WCu,nu_WCu,k_WCu,d_WCu)
    m.Medium(pbl, "Spacer 3 part 2", "WCu",E_WCu,nu_WCu,k_WCu,d_WCu)
    
    m.uDirichlet(pbl, "Holder z clamping", "Clamped",0,0.,0,0.,1,0.,0,0.,ensemble_size)
    m.uDirichlet(pbl, "Holder x-y clamping", "Clamped",1,0.,1,0.,0,0.,0,0.,ensemble_size)
    m.uDirichlet(pbl, "Cooling channels", "Clamped",0,0.,0,0.,0,0.,1,70.-T_ref,ensemble_size)

    zero = np.zeros(ensemble_size)
    hf = (20/0.0121)* np.ones(ensemble_size) #20 W / (0.0121 m^2)

    m.Neumann(pbl, "Mirror surface","heat fluxes",0,zero,0,zero,0,zero,1,hf,ensemble_size)

    m.Source(pbl, "Mirror", "ihg",70E6)
    m.Source(pbl, "Substrate", "ihg",2.45E6)
    m.Source(pbl, "Holder", "ihg",0.7E6)

    m.Source(pbl, "Spacer 1 part 1", "ihg",0.6E6)
    m.Source(pbl, "Spacer 2 part 1", "ihg",0.6E6)
    m.Source(pbl, "Spacer 3 part 1", "ihg",0.6E6)

    m.Source(pbl, "Spacer 1 part 2", "ihg",2.45E6)
    m.Source(pbl, "Spacer 2 part 2", "ihg",2.45E6)
    m.Source(pbl, "Spacer 3 part 2", "ihg",2.45E6)

    m.Source(pbl, "Shaft 1", "ihg",0.8E6)
    m.Source(pbl, "Shaft 2", "ihg",0.8E6)
    m.Source(pbl, "Shaft 3", "ihg",0.8E6)

    m.Source(pbl, "Nut 1", "ihg",0.8E6)
    m.Source(pbl, "Nut 2", "ihg",0.8E6)
    m.Source(pbl, "Nut 3", "ihg",0.8E6)

    m.Source(pbl, "Washer 1", "ihg",0.8E6)
    m.Source(pbl, "Washer 2", "ihg",0.8E6)
    m.Source(pbl, "Washer 3", "ihg",0.8E6)

    slv = m.IterativeSolver(pbl,args.k,solverList,4, False, ensemble_size)

    slv.start()
    
    try:
        from mrstlnos.Juelich.post_process.vtk_post_process import post_process
        post_process()
        from mrstlnos.Juelich.post_process.vtk_post_process_2 import post_process
        post_process()
        from mrstlnos.Juelich.post_process.vtk_post_process_2 import post_process_mirror
        post_process_mirror()
        print "Post processing has been performed"
    except:
        print "No post processing has been performed"

if __name__ == "__main__":
    main()
