#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh
#import matplotlib.pyplot as plt

def read_msh(index_nodes,index_elements,index_tags,filename,filename_output):
    import shlex, subprocess
    import os

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"
    if rank == 0:
        command_line = 'gmsh -3 ' + file_dir + '/' +  filename + ' -o ' + work_dir + '/' +  filename_output
        print command_line
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log','w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
        retcode = p.wait()
        for i in range(1,siz):
            comm.send(1, dest=i, tag=11)
    else:
        data = comm.recv(source=0, tag=11)
    msh = gmsh.MeshLoader(filename_output,work_dir).execute()
    gmshWriter = tbox.GmshExport(msh)
    n_nodes = index_nodes
    for n in msh.nodes:
        n.no = n.no + index_nodes
        n_nodes = n_nodes + 1
    n_elems = index_elements
    for e in msh.elems:
        e.no = e.no + index_elements
        n_elems = n_elems + 1
    n_ptags = index_tags
    for p in msh.ptags:
        if msh.ptags[p].no > 1 :
            msh.ptags[p].no = msh.ptags[p].no+index_tags
        n_ptags = n_ptags+1
    gmshWriter.save(filename_output)

    return n_nodes, n_elems, n_ptags

def fuse_meshes(filename1,filename12,filename2,filename22,filename_output):

    n_nodes, n_elems, n_ptags = read_msh(0,0,0,filename1,filename12)
    n_nodes, n_elems, n_ptags = read_msh(n_nodes, n_elems, n_ptags,filename2,filename22)


    import shlex, subprocess
    import os

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"
    if rank == 0:
        command_line = 'gmsh -3 ' + work_dir + '/' +  filename12 + ' ' + work_dir + '/' +  filename22 + ' -o ' + work_dir + '/' +  filename_output
        print command_line
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log','w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
        retcode = p.wait()
        for i in range(1,siz):
            comm.send(1, dest=i, tag=11)
    else:
        data = comm.recv(source=0, tag=11)

    msh = gmsh.MeshLoader(filename_output,work_dir).execute()

    return msh

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"

    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos

    msh = fuse_meshes('L_shape_curv_1.msh','L_shape_1.msh','L_shape_curv_2.msh','L_shape_2.msh','L_shape_fused.msh')

    import shlex, subprocess
    import os
    import numpy as np

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    ensemble_size = 1

    pbl = m.Problem(msh,comm)

    E = 10. * np.ones(ensemble_size)
    nu = 0.4 * np.ones(ensemble_size)
    k = 0.114
    d = 0.
    m.Medium(pbl, "body", "test",E,nu,k,d)


    m.uDirichlet(pbl, "clamping", "Clamped",1,0.,1,0.,1,0.,1,0.,ensemble_size)
    #m.uDirichlet(pbl, "load", "Clamped",1,0.,1,2.,1,0.,0,0.,ensemble_size)

    norm = tbox.Pt(0,0,-1)
    cont = m.Contact(pbl,"inner edge boundary","contact",norm)
    cont.setMeshTying()
    cont.setMaster(pbl,"outer part boundary",norm)
    f = -0.002 * np.ones(ensemble_size)
    zero = np.zeros(ensemble_size)
    one = np.ones(ensemble_size)

    m.Neumann(pbl, "load","load",0,zero,1,f,0,zero,1,one,ensemble_size)

    solverList = Teuchos.ParameterList()
    solverList['Convergence Tolerance'] = 10**(-8)
    solverList['Maximum Iterations'] = 10
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    solverList["MueLu xml file name"] = file_dir + '/direct_solver.xml'
    solverList["convert MueLu xml file"] = True

    solverList['Use preconditioner'] = True

    solverList['type']= "BlockGmres"
    solverList['Maximum active set iteration'] = 1

    #solverList['Write matrix and vectors'] = True
    solverList['Write txt files'] = True
    
    num_random_variables = 300
    
    for i in range(0,num_random_variables):
        rand = np.random.normal(0, 1, ensemble_size)
        m.RandomVariable(pbl,rand)

    randomParams = Teuchos.ParameterList()
    
    randomParams['Is random'] = True
    randomParams['Begin X'] = 0.
    randomParams['Begin Y'] = 0.
    randomParams['Begin Z'] = 0.
    randomParams['End X'] = 10.
    randomParams['End Y'] = 10.
    randomParams['End Z'] = 2.
    randomParams['Number random variables'] = num_random_variables
    randomParams['Correlation length X'] = 5.
    randomParams['Correlation length Y'] = 5.
    randomParams['Correlation length Z'] = 5.
    randomParams['Maximum Nonlinear Solver Iterations'] = 500

    randomParams['Mean'] = 10./(2.*(1.+0.4))
    randomParams['Dispersion level'] = 0.1
    
    solverList['randomParams'] = randomParams

    useEP = (ensemble_size>1)

    slv = m.IterativeSolver(pbl,args.k,solverList,4, useEP, ensemble_size)

    slv.start()

if __name__ == "__main__":
    main()
