#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh
#import mrstlnos.viewer as mrsv # Error occurs after that

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"


    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos
    import numpy as np
    #import matplotlib.pyplot as plt

    import mrstlnos.utilities as u


    DOFperNode = 3

    ## Mesh

    geo_name = 'beam_4.geo'
    mesh_name = 'beam_4.msh'
    partitioned_mesh_name = 'beam_k.msh'

    par={}
    par['L'] = 50.
    par['W'] = 5.
    par['H'] = 5.
    par['d'] = 1.
    par['n_L'] = 60
    par['n_hW'] = 3
    par['n_hH'] = 3

    ensemble_size = 16

    p_min = 0.5
    p_max = 2.5

    E_min = 20500
    E_max = 21500


    import shlex, subprocess
    import os

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()


    x = tmp = np.load(file_dir+'/QMC_sequence.npy')

    p_gauss = x[0,:]*(p_max - p_min) + p_min
    E_gauss = x[1,:]*(E_max - E_min) + E_min

    M = len(p_gauss)




    import shlex, subprocess
    import os

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    pars={ 'L' : par['L'] , 'W' : par['W'], 'H' : par['H'], 'd' : par['d'], 'n_L' : par['n_L'], 'n_hW' : par['n_hW'], 'n_hH' : par['n_hH']}
    gmsh.MeshLoader(geo_name,__file__).execute(**pars)

    if rank == 0:
        command_line = 'gmsh -3 ' + work_dir + '/' +  mesh_name + ' -o ' + work_dir + '/' +  partitioned_mesh_name +' -part ' + str(siz)
        print command_line
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log','w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
        retcode = p.wait()
        for i in range(1,siz):
            comm.send(1, dest=i, tag=11)
    else:
        data = comm.recv(source=0, tag=11)

    if siz == 1:
        msh = gmsh.MeshLoader(partitioned_mesh_name,work_dir).execute()
    else:
        msh = gmsh.MeshLoader(partitioned_mesh_name,work_dir).myrank_execute(rank)

    for n1 in msh.nodes:
        n1.row = n1.no-1      



    for i in range(0,M,ensemble_size):
        pbl = m.Problem(msh,comm)


        dx = np.zeros(ensemble_size)
        dy = np.zeros(ensemble_size)
        dy2 = np.linspace(0,-2.,ensemble_size)
        dz = 0*np.linspace(0,-2.,ensemble_size)
        dz2 = np.zeros(ensemble_size)
        dT = np.zeros(ensemble_size)
        m.uDirichlet(pbl, "Rigid clamped surface", "Clamped 1",1,dx,1,dy,1,dz,0,dT)
        m.uDirichlet(pbl, "Clamped end", "Clamped 1",1,dx,1,dy,1,dz2,0,dT)
        m.uDirichlet(pbl, "Free end", "Clamped 1",1,0.,0,0.,0,0.,0,0.,ensemble_size)

        norm1 = tbox.Pt(0,0,-1)
        norm2 = tbox.Pt(0,0,1)
        dg = 0. * np.ones(ensemble_size) 

        a = 49.5
        initiallyClosedNodesList = np.array([], dtype=np.float)
        #tag_name = "Center of cross section"
        for el in msh.ntags["Beam contact zone"].elems:
            for n1 in el.nodes:
                node_id = n1.no-1
                if msh.nodes[node_id].pos.x[0] >= a:
                    initiallyClosedNodesList = np.append(initiallyClosedNodesList, node_id)           

        cont1 = m.Contact(pbl,"Beam contact zone","contact",norm1,dg)
        #cont1.setSlaveBoundary(work_dir+"/"+partitioned_mesh_name, "Beam contact boundary")
        #cont1.setMasterBoundary(work_dir+"/"+partitioned_mesh_name, "Rigid surface boundary")
        cont1.setInitialyOpen()
        cont1.setInitiallyClosedNodes(initiallyClosedNodesList)
        cont1.setDualShapeFunctions()
        cont1.setSlaveDirichlet(work_dir+"/"+partitioned_mesh_name, "Beam contact Dirichlet")
        cont1.setSignoriniContact()
        cont1.setMaster(pbl,"Rigid surface",norm2)

        ## Solver

        solverList = Teuchos.ParameterList()
        solverList['Convergence Tolerance'] = 10**(-8)
        solverList['Maximum Iterations'] = 100000
        solverList['Verbosity'] = 33
        solverList['Flexible Gmres'] = False
        solverList['Output Style'] = 1
        solverList['Output Frequency'] = 1

        solverList["MueLu xml file name"] = file_dir + '/nested_multigrid_2_lvls_chebyshev.xml'
        solverList["convert MueLu xml file"] = True

        solverList['Use preconditioner'] = True

        solverList['type']= "BlockGmres"
        solverList['Maximum active set iteration'] = 10
        solverList['Write txt files'] = True

        directory = str(i)
        if not os.path.exists(directory):
            os.makedirs(directory)
        os.chdir(directory)    

        p = p_gauss[i] * np.ones(ensemble_size) 
        for ii in range(0,ensemble_size):
            p[ii] = p_gauss[i+ii]

        zero = np.zeros(ensemble_size)

        m.Neumann(pbl, "Loaded surface", "Load 1",0,zero,0,zero,1,-p,0,zero,ensemble_size)

        E = E_gauss[i] * np.ones(ensemble_size)
        for ii in range(0,ensemble_size):
            E[ii] = E_gauss[i+ii]

        nu = 0.29 * np.ones(ensemble_size)
        k = 0
        d = 0

        m.Medium(pbl, "Beam", "Mat 1",E,nu,k,d)
        m.Medium(pbl, "Rigid body", "Mat 1",E,nu,k,d)

        slv = m.IterativeSolver(pbl,args.k,solverList,DOFperNode, True, ensemble_size)

        slv.start()

        f = open('timers.txt', 'w')
        print >>f, slv.tms
        f.closed

        f = open('parameters_values.txt', 'w')
        print >>f, E
        print >>f, p
        f.closed        

        os.rename('x_mm.txt', 'x_mm_'+str(i)+'.txt')
        os.chdir('..') 

if __name__ == "__main__":
    main()
