#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh
#import matplotlib.pyplot as plt

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"



    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos

    geo_name = 'cube_params.geo'
    mesh_name = 'cube_params.msh'
    partitioned_mesh_name = 'cube.msh'

    nC = 32


    import shlex, subprocess
    import os
    import numpy as np

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    pars={ 'nC' : nC}
    gmsh.MeshLoader(geo_name,__file__).execute(**pars)

    if rank == 0:
        command_line = 'gmsh -3 ' + work_dir + '/' +  mesh_name + ' -o ' + work_dir + '/' +  partitioned_mesh_name +' -part ' + str(siz)
        print command_line
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log','w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
        retcode = p.wait()
        for i in range(1,siz):
            comm.send(1, dest=i, tag=11)
    else:
        data = comm.recv(source=0, tag=11)

    #ks = np.array([1,2,3,4])
    ks = np.array([1])
    n = ks.size
    belos_times = np.zeros((n,), dtype=np.float64)
    prec_times = np.zeros((n,), dtype=np.float64)
    graph_times = np.zeros((n,), dtype=np.float64)
    computeMatrices_times = np.zeros((n,), dtype=np.float64)
    computeLoads_times = np.zeros((n,), dtype=np.float64)

    belos_matrix_times = np.zeros((n,), dtype=np.float64)
    belos_prec_times = np.zeros((n,), dtype=np.float64)
    belos_apply_times = np.zeros((n,), dtype=np.float64) 
    ortho_times = np.zeros((n,), dtype=np.float64) 
    ortho_times_inner = np.zeros((n,), dtype=np.float64) 
    ortho_times_norm = np.zeros((n,), dtype=np.float64)  
    ortho_times_update = np.zeros((n,), dtype=np.float64)    
    total = np.zeros((n,), dtype=np.float64) 

    solver_time = np.zeros((n,), dtype=np.float64) 
    writing_time = np.zeros((n,), dtype=np.float64) 

    discrete_problem_times = np.zeros((n,), dtype=np.float64) 
    map_times = np.zeros((n,), dtype=np.float64) 
    graph_times = np.zeros((n,), dtype=np.float64) 
    update_time = np.zeros((n,), dtype=np.float64) 
    ## Problem
    for e in range(0,n):
            
        if siz == 1:
            msh = gmsh.MeshLoader(partitioned_mesh_name,work_dir).execute()
        else:
            msh = gmsh.MeshLoader(partitioned_mesh_name,work_dir).myrank_execute(rank)
    
        k = ks[e]
        print msh.name
        msh.name = msh.name + "_k_" + str(k)
        print msh.name

        ensemble_size = 1

        pbl = m.Problem(msh,comm)

        #E = np.linspace(0.5,1.5,ensemble_size)
        E = 1. * np.ones(ensemble_size)
        nu = 0.2 * np.ones(ensemble_size)
        m.Medium(pbl, "Body", "test",E,nu)


        m.uDirichlet(pbl, "Surf 1", "Clamped",1,0.,1,0.,1,0.,0,0.,ensemble_size)

        dx = np.zeros(ensemble_size)
        dy = 1. * np.ones(ensemble_size) 
        #dy = np.linspace(0.5,1.5,ensemble_size)
        dz = np.zeros(ensemble_size)
        dT = np.zeros(ensemble_size)
        m.uDirichlet(pbl, "Surf 2", "Moved",0,dx,1,dy,0,dz,0,dT)

        precondList = Teuchos.ParameterList()
        precondList = Teuchos.ParameterList()
        precondSubList = Teuchos.ParameterList()
        precondSubSubList = Teuchos.ParameterList()
        precondSubSubList['fact: iluk level-of-fill'] = 2
        precondSubList['Prec Type'] = "RILUK"
        precondSubList['Overlap'] = 1
        precondSubList['Ifpack2 Settings'] = precondSubSubList
        precondList['Ifpack2'] = precondSubList

        mueluParams = Teuchos.ParameterList()
        #mueluParamsSub1 = Teuchos.ParameterList()
        #mueluParamsSub1['PDE equations'] = 4
        #mueluParams['Matrix'] = mueluParamsSub1
        mueluParams['number of equations'] = 3
        mueluParams['max levels'] = 5
        mueluParams['multigrid algorithm'] = "sa"
        mueluParams['problem: symmetric'] = True
        mueluParams['coarse: max size'] = 10
        #mueluParams['aggregation: type'] = "uncoupled"
        #mueluParams['aggregation: min agg size'] = 3
        #mueluParams['aggregation: max agg size'] = 10
        mueluParams['aggregation: export visualization data'] = True
        mueluParams['smoother: type'] = "RELAXATION"

        mueluParamsSub2 = Teuchos.ParameterList()

        mueluParamsSub2['relaxation: type'] = "Symmetric Gauss-Seidel"
        mueluParamsSub2['relaxation: sweeps'] = 5
        #mueluParamsSub2['relaxation: damping factor'] = 0.9

        mueluParams['coarse: type'] = "Klu"
        mueluParams['coarse: max size'] = 18
        mueluParams['verbosity'] = "low"

        mueluParamsSub3 = Teuchos.ParameterList()

        mueluParamsSub4 = Teuchos.ParameterList()
        mueluParamsSub4['A'] = '{0,1,2}'
        mueluParamsSub4['P'] = '{0,1,2}'
        mueluParamsSub4['R'] = '{0,1}'

        mueluParams['smoother: params'] = mueluParamsSub2
        mueluParams['export data'] = mueluParamsSub4


        solverList = Teuchos.ParameterList()
        solverList['type']= "BlockGmres"
        solverList['Ensemble Convergence Tolerance'] = 10**(-8)
        solverList['Maximum Iterations'] = 500
        solverList['Verbosity'] = 33
        solverList['Flexible Gmres'] = False
        solverList['Output Style'] = 1
        solverList['Output Frequency'] = 1
        solverList['Output Stream file name belos'] = 'belos_'+str(k)+'.text'
        #solverList['Output Stream file name muelu'] = 'muelu_prec.text'
        solverList['Use preconditioner'] = False
        solverList['Write matrix and vectors'] = True
        solverList['ifpackParams'] = precondList
        solverList['mueluParams'] = mueluParams

        solverList['Print Teuchos timers'] = True
        solverList['Write matrix and vectors'] = False

        if ensemble_size>1:
            slv = m.IterativeSolver(pbl,k,solverList,3, True, ensemble_size)
        else:
            slv = m.IterativeSolver(pbl,k,solverList,3, False, ensemble_size)

        slv.start()

        if rank == 0:
            belos_times[e] = slv.tms['time_belos_total'].read().getReal().sec()
            prec_times[e] = slv.tms['prec'].read().getReal().sec()
            graph_times[e] = slv.tms['graph'].read().getReal().sec()
            computeMatrices_times[e] = slv.tms['computeMatrices'].read().getReal().sec()
            computeLoads_times[e] = slv.tms['computeLoads'].read().getReal().sec()

            belos_matrix_times[e] =  slv.tms["time_mat_vec"].read().getReal().sec()
            belos_prec_times[e] = slv.tms["time_prec_setup"].read().getReal().sec()
            belos_apply_times[e] = slv.tms["time_prec_apply"].read().getReal().sec()

            ortho_times[e] = slv.tms["time_ortho"].read().getReal().sec()
            ortho_times_inner[e] = slv.tms["time_ortho_inner"].read().getReal().sec()
            ortho_times_norm[e] = slv.tms["time_ortho_norm"].read().getReal().sec()
            ortho_times_update[e] = slv.tms["time_ortho_update"].read().getReal().sec()

            writing_time[e] = slv.tms["write results"].read().getReal().sec()
            solver_time[e] = slv.tms["solver"].read().getReal().sec()

            total[e] = slv.tms["total"].read().getReal().sec()

            map_times[e] = slv.tms["map"].read().getReal().sec()
            graph_times[e] = slv.tms["graph"].read().getReal().sec()
            discrete_problem_times[e] = slv.tms["dp"].read().getReal().sec()

            update_time[e] = slv.tms["update"].read().getReal().sec()

    if rank == 0:
        print "k:"
        print ks
        print "-------------------"
        print "Total:"
        print total        
        print "-------------------"

        print "Convert to discrete problem:"
        print discrete_problem_times
        print "Map:"
        print map_times
        print "Graph:"
        print graph_times
        print "Matrix assembly:"
        print computeMatrices_times
        print "BCs:"
        print computeLoads_times

        print "Total preprocessing:"
        print discrete_problem_times+computeMatrices_times+computeLoads_times

        print "-------------------"
        print "Solver: active set and gmres combine (= Belos + update cost + writting results)"
        print solver_time
        print "Update:"
        print update_time 
        print "Writing time"
        print writing_time              
        print "-------------------"
        print "Belos total time:"
        print belos_times
        print "Belos apply matrix:"
        print belos_matrix_times        
        print "Belos apply prec:"
        print belos_apply_times
        print "Belos orthogonalization:"
        print ortho_times
        print "Belos orthogonalization: inner product"
        print ortho_times_inner
        print "Belos orthogonalization: norm"
        print ortho_times_norm
        print "Belos orthogonalization: update"
        print ortho_times_update

if __name__ == "__main__":
    main()
