#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"


    n_1D_gauss = 16

    nC = 32

    delta_p = 100.
    p_mean = 0.

    delta_E = 1000.
    E_mean = 21000.

    use_preconditioner = False

    p_min = p_mean-delta_p/2
    p_max = p_mean+delta_p/2

    E_min = E_mean-delta_E/2
    E_max = E_mean+delta_E/2

    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos

    geo_name = 'cube_params_tet.geo'
    mesh_name = 'cube_params_tet.msh'
    partitioned_mesh_name = 'cube_tet.msh'

    


    import shlex, subprocess
    import os
    import numpy as np

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    pars={ 'nC' : nC}
    gmsh.MeshLoader(geo_name,__file__).execute(**pars)

    if rank == 0:
        command_line = 'gmsh -3 ' + work_dir + '/' +  mesh_name + ' -o ' + work_dir + '/' +  partitioned_mesh_name +' -part ' + str(siz)
        print command_line
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log','w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
        retcode = p.wait()
        for i in range(1,siz):
            comm.send(1, dest=i, tag=11)
    else:
        data = comm.recv(source=0, tag=11)

    ensemble_size = 1

    DOFperNode = 3

    x, w = np.polynomial.legendre.leggauss(n_1D_gauss)
    p_gauss = 0.5*(x + 1)*(p_max - p_min) + p_min
    E_gauss = 0.5*(x + 1)*(E_max - E_min) + E_min



    if siz == 1:
        msh = gmsh.MeshLoader(partitioned_mesh_name,work_dir).execute()
    else:
        msh = gmsh.MeshLoader(partitioned_mesh_name,work_dir).myrank_execute(rank)

    for n1 in msh.nodes:
        n1.row = n1.no-1      



    for i in range(0,1):#len(x)):
        for j in range(0,1):#len(x)):
            pbl = m.Problem(msh,comm)

            
            pbl = m.Problem(msh,comm)

            p = p_gauss[i] * np.ones(ensemble_size) 

            zero = np.zeros(ensemble_size)

            m.Neumann(pbl, "Surf 2", "Load 1",0,zero,1,p,0,zero,0,zero,ensemble_size)

            E = E_gauss[j] * np.ones(ensemble_size)
            nu = 0.29 * np.ones(ensemble_size)
            m.Medium(pbl, "Body", "test",E,nu)

            m.uDirichlet(pbl, "Surf 1", "Moved",1,zero ,1,zero ,1,zero ,0,zero )

            precondList = Teuchos.ParameterList()
            precondList = Teuchos.ParameterList()
            precondSubList = Teuchos.ParameterList()
            precondSubSubList = Teuchos.ParameterList()
            precondSubSubList['fact: iluk level-of-fill'] = 2
            precondSubList['Prec Type'] = "RILUK"
            precondSubList['Overlap'] = 1
            precondSubList['Ifpack2 Settings'] = precondSubSubList
            precondList['Ifpack2'] = precondSubList

            mueluParams = Teuchos.ParameterList()
            #mueluParamsSub1 = Teuchos.ParameterList()
            #mueluParamsSub1['PDE equations'] = 4
            #mueluParams['Matrix'] = mueluParamsSub1
            mueluParams['number of equations'] = 3
            mueluParams['max levels'] = 5
            mueluParams['multigrid algorithm'] = "sa"
            mueluParams['problem: symmetric'] = True
            mueluParams['coarse: max size'] = 3000
            #mueluParams['aggregation: type'] = "uncoupled"
            #mueluParams['aggregation: min agg size'] = 3
            #mueluParams['aggregation: max agg size'] = 10
            mueluParams['smoother: type'] = "RELAXATION"

            mueluParamsSub2 = Teuchos.ParameterList()

            mueluParamsSub2['relaxation: type'] = "Symmetric Gauss-Seidel"
            mueluParamsSub2['relaxation: sweeps'] = 5
            #mueluParamsSub2['relaxation: damping factor'] = 0.9

            mueluParams['coarse: type'] = "Klu"
            mueluParams['verbosity'] = "low"


            mueluParams['smoother: params'] = mueluParamsSub2

            solverList = Teuchos.ParameterList()
            solverList['type']= "BlockGmres"
            solverList['Ensemble Convergence Tolerance'] = 10**(-8)
            solverList['Maximum Iterations'] = 10000
            solverList['Verbosity'] = 33
            solverList['Flexible Gmres'] = False
            solverList['Output Style'] = 1
            solverList['Output Frequency'] = 1
            solverList['Output Stream file name belos'] = 'belos_'+str(ensemble_size)+'.text'
            #solverList['Output Stream file name muelu'] = 'muelu_prec.text'
            solverList['Use preconditioner'] = False
            solverList['Write matrix and vectors'] = False
            solverList['ifpackParams'] = precondList
            solverList['mueluParams'] = mueluParams

            solverList['Print Teuchos timers'] = True
            solverList['Write txt files'] = True
            solverList['Write intermediate files'] = True

            slv = m.IterativeSolver(pbl,args.k,solverList,DOFperNode, False, 1)

            directory = str(i)+'_'+str(j)
            if not os.path.exists(directory):
                os.makedirs(directory)
            os.chdir(directory)    

            slv.start()

            f = open('timers.txt', 'w')
            print >>f, slv.tms
            f.closed

            f = open('parameters_values.txt', 'w')
            print >>f, E
            print >>f, p
            f.closed
            

            os.rename('x_mm.txt', 'x_mm_'+str(i)+'_'+str(j)+'.txt')
            os.chdir('..') 

if __name__ == "__main__":
    main()
