#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"



    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos

    import shlex, subprocess
    import os
    import numpy as np

    geo_name = 'cube.geo'
    geo_current = 'cube_current.geo'
    mesh_name = 'cube.msh'

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    d = 13

    nC = np.arange(4,8+d)
    N = np.arange(0,d)
    belos = np.arange(0,d, dtype=np.float)
    prec = np.arange(0,d, dtype=np.float)
    for j in range(0,d):

        if rank == 0:
            replaceData(file_dir + '/' +  geo_name,file_dir + '/' +  geo_current,'nC = 8;','nC = '+str(nC[j])+';')

        if rank == 0:
            command_line = 'gmsh -3 ' + file_dir + '/' +  geo_current + ' -o ' + work_dir + '/' +  mesh_name +' -part ' + str(siz)
            print command_line
            tmp = shlex.split(command_line)
            fileout = open('gmsh.log','w')
            p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
            retcode = p.wait()
            for i in range(1,siz):
                comm.send(1, dest=i, tag=11)
        else:
            data = comm.recv(source=0, tag=11)


        if siz == 1:
            msh = gmsh.MeshLoader(mesh_name,work_dir).execute()
        else:
            msh = gmsh.MeshLoader(mesh_name,work_dir).myrank_execute(rank)

        pbl = m.Problem(msh,comm)

        ensemble_size = 1

        E = 1. * np.ones(ensemble_size)
        nu = 0.2 * np.ones(ensemble_size)
        m.Medium(pbl, "Body", "test",E,nu)


        m.uDirichlet(pbl, "Surf 1", "Clamped",1,0.,1,0.,1,0.,0,0.,ensemble_size)

        dx = np.zeros(ensemble_size)
        dy = 0.1 * np.ones(ensemble_size)
        dz = np.zeros(ensemble_size)
        dT = np.zeros(ensemble_size)
        m.uDirichlet(pbl, "Surf 2", "Moved",0,dx,1,dy,0,dz,0,dT)

        precondList = Teuchos.ParameterList()
        precondList = Teuchos.ParameterList()
        precondSubList = Teuchos.ParameterList()
        precondSubSubList = Teuchos.ParameterList()
        precondSubSubList['fact: iluk level-of-fill'] = 2
        precondSubList['Prec Type'] = "RILUK"
        precondSubList['Overlap'] = 1
        precondSubList['Ifpack2 Settings'] = precondSubSubList
        precondList['Ifpack2'] = precondSubList

        mueluParams = Teuchos.ParameterList()
        #mueluParamsSub1 = Teuchos.ParameterList()
        #mueluParamsSub1['PDE equations'] = 4
        #mueluParams['Matrix'] = mueluParamsSub1
        mueluParams['number of equations'] = 3
        mueluParams['max levels'] = 2
        mueluParams['multigrid algorithm'] = "unsmoothed"
        mueluParams['smoother: type'] = "RELAXATION"

        mueluParamsSub2 = Teuchos.ParameterList()

        mueluParamsSub2['relaxation: type'] = "Symmetric Gauss-Seidel"
        mueluParamsSub2['relaxation: sweeps'] = 4
        mueluParamsSub2['relaxation: damping factor'] = 1.25

        mueluParams['coarse: type'] = "RELAXATION"

        mueluParamsSub3 = Teuchos.ParameterList()

        mueluParamsSub3['relaxation: type'] = "Symmetric Gauss-Seidel"
        mueluParamsSub3['relaxation: sweeps'] = 4
        mueluParamsSub3['relaxation: damping factor'] = 1.25


        mueluParams['smoother: params'] = mueluParamsSub2
        mueluParams['coarse: params'] = mueluParamsSub3

        solverList = Teuchos.ParameterList()
        solverList['type']= "BlockGmres"
        solverList['Convergence Tolerance'] = 10**(-8)
        solverList['Maximum Iterations'] = 1000
        solverList['Verbosity'] = 33
        solverList['Flexible Gmres'] = False
        solverList['Output Style'] = 1
        solverList['Output Frequency'] = 1
        solverList['ifpackParams'] = precondList
        solverList['mueluParams'] = mueluParams
        slv = m.IterativeSolver(pbl,args.k,solverList,3, True, ensemble_size)
        slv.start()

        if rank == 0:
            N[j] = 3*(nC[j]+1)**3
            prec[j] = float(slv.tms['prec'].read().getReal().sec())
            belos[j] = float(slv.tms['belos'].read().getReal().sec())

    if rank == 0:
        print 'finished'
        print nC
        print N
        print prec
        print belos

def replaceData(previousName,newName,previousData,newData):
    f = open(previousName,'r')
    filedata = f.read()
    f.close()

    data = filedata.replace(previousData,newData)

    f = open(newName,'w')
    f.write(data)
    f.close()

if __name__ == "__main__":
    main()
