#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh
#import matplotlib.pyplot as plt

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"



    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos

    geo_name = 'cube_params.geo'
    mesh_name = 'cube_params.msh'
    partitioned_mesh_name = 'cube.msh'

    nC = 16


    import shlex, subprocess
    import os
    import numpy as np

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    pars={ 'nC' : nC}
    gmsh.MeshLoader(geo_name,__file__).execute(**pars)

    if rank == 0:
        command_line = 'gmsh -3 ' + work_dir + '/' +  mesh_name + ' -o ' + work_dir + '/' +  partitioned_mesh_name +' -part ' + str(siz)
        print command_line
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log','w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
        retcode = p.wait()
        for i in range(1,siz):
            comm.send(1, dest=i, tag=11)
    else:
        data = comm.recv(source=0, tag=11)


    if siz == 1:
        msh = gmsh.MeshLoader(mesh_name,work_dir).execute()
    else:
        msh = gmsh.MeshLoader(mesh_name,work_dir).myrank_execute(rank)
 

    ensemble_size = 1

    pbl = m.Problem(msh,comm)

    #E = np.linspace(0.5,1.5,ensemble_size)
    E = 1. * np.ones(ensemble_size)
    nu = 0.2 * np.ones(ensemble_size)
    k = 0.114
    d = 0.
    m.Medium(pbl, "Body", "test",E,nu,k,d)


    m.uDirichlet(pbl, "Surf 1", "Clamped",1,0.,1,0.,1,0.,1,1.,ensemble_size)

    dx = np.zeros(ensemble_size)
    dy = 2. * np.ones(ensemble_size) 
    #dy = np.linspace(0.5,1.5,ensemble_size)
    dz = np.zeros(ensemble_size)
    dT = 10. * np.ones(ensemble_size)
    m.uDirichlet(pbl, "Surf 2", "Moved",0,dx,1,dy,0,dz,1,dT)

    precondList = Teuchos.ParameterList()
    precondList = Teuchos.ParameterList()
    precondSubList = Teuchos.ParameterList()
    precondSubSubList = Teuchos.ParameterList()
    precondSubSubList['fact: iluk level-of-fill'] = 2
    precondSubList['Prec Type'] = "RILUK"
    precondSubList['Overlap'] = 1
    precondSubList['Ifpack2 Settings'] = precondSubSubList
    precondList['Ifpack2'] = precondSubList

    mueluParams = Teuchos.ParameterList()
    #mueluParamsSub1 = Teuchos.ParameterList()
    #mueluParamsSub1['PDE equations'] = 4
    #mueluParams['Matrix'] = mueluParamsSub1
    mueluParams['number of equations'] = 4
    mueluParams['max levels'] = 5
    mueluParams['multigrid algorithm'] = "sa"
    mueluParams['problem: symmetric'] = True
    mueluParams['coarse: max size'] = 10
    #mueluParams['aggregation: type'] = "uncoupled"
    #mueluParams['aggregation: min agg size'] = 3
    #mueluParams['aggregation: max agg size'] = 10
    mueluParams['aggregation: export visualization data'] = True
    mueluParams['smoother: type'] = "RELAXATION"

    mueluParamsSub2 = Teuchos.ParameterList()

    mueluParamsSub2['relaxation: type'] = "Symmetric Gauss-Seidel"
    mueluParamsSub2['relaxation: sweeps'] = 5
    #mueluParamsSub2['relaxation: damping factor'] = 0.9

    mueluParams['coarse: type'] = "Klu"
    mueluParams['coarse: max size'] = 18
    mueluParams['verbosity'] = "low"

    mueluParamsSub3 = Teuchos.ParameterList()

    mueluParamsSub4 = Teuchos.ParameterList()
    mueluParamsSub4['A'] = '{0,1,2}'
    mueluParamsSub4['P'] = '{0,1,2}'
    mueluParamsSub4['R'] = '{0,1}'

    mueluParams['smoother: params'] = mueluParamsSub2
    mueluParams['export data'] = mueluParamsSub4


    solverList = Teuchos.ParameterList()
    solverList['type']= "BlockGmres"
    solverList['Ensemble Convergence Tolerance'] = 10**(-8)
    solverList['Maximum Iterations'] = 500
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1
    solverList['Output Stream file name belos'] = 'belos_'+str(ensemble_size)+'.text'
    #solverList['Output Stream file name muelu'] = 'muelu_prec.text'
    solverList['Use preconditioner'] = False
    solverList['ifpackParams'] = precondList
    solverList['mueluParams'] = mueluParams

    solverList['Print Teuchos timers'] = True
    #solverList['Write matrix and vectors'] = True
    #solverList['Write txt files'] = True

    

    if ensemble_size>1:
        slv = m.IterativeSolver(pbl,args.k,solverList,4, True, ensemble_size)
    else:
        slv = m.IterativeSolver(pbl,args.k,solverList,4, False, ensemble_size)

    slv.start()

if __name__ == "__main__":
    main()
