#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"



    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos

    geo_name = 'cube.geo'
    mesh_name = 'cube.msh'


    import shlex, subprocess
    import os
    import numpy as np

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    if rank == 0:
        command_line = 'gmsh -3 ' + file_dir + '/' +  geo_name + ' -o ' + work_dir + '/' +  mesh_name +' -part ' + str(siz)
        print command_line
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log','w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
        retcode = p.wait()
        for i in range(1,siz):
            comm.send(1, dest=i, tag=11)
    else:
        data = comm.recv(source=0, tag=11)


    if siz == 1:
        msh = gmsh.MeshLoader(mesh_name,work_dir).execute()
    else:
        msh = gmsh.MeshLoader(mesh_name,work_dir).myrank_execute(rank)

    pbl = m.Problem(msh,comm)

    ensemble_size = 1

    E = 1. * np.ones(ensemble_size)
    nu = 0.2 * np.ones(ensemble_size)
    m.Medium(pbl, "Body", "test",E,nu)


    m.uDirichlet(pbl, "Surf 1", "Clamped",1,0.,1,0.,1,0.,0,0.,ensemble_size)

    dx = np.zeros(ensemble_size)
    dy = 0.1 * np.ones(ensemble_size)
    dz = np.zeros(ensemble_size)
    dT = np.zeros(ensemble_size)
    m.uDirichlet(pbl, "Surf 2", "Moved",0,dx,1,dy,0,dz,0,dT)

    solverList = Teuchos.ParameterList()
    slv = m.DirectSolver(pbl,args.k,solverList,3, True, ensemble_size)

    if not args.nogui:
        if rank == 0:
            all_msh = gmsh.MeshLoader(mesh_name,work_dir).execute()
            import mrstlnos.viewer as v
            gui = v.MeshViewer(pbl,slv,all_msh)
            gui.vmin = 0
            gui.vmax = 1
            gui.start()
        else:
            slv.start()
    else:
        slv.start()

    timers = slv.tms;
    if rank == 0:
        print 'Map real CPU cost'
        print timers['map'].read().getReal()
        print 'Graph real CPU cost'
        print timers['graph'].read().getReal()
        print 'Matrix computation on each mpi process real CPU cost'
        print timers['computeMatrices'].read().getReal()
        print 'MPI sharing CPU cost'
        print timers['assembly-MPI'].read().getReal()
        print 'Amesos2'
        print timers['Amesos2'].read().getReal()

if __name__ == "__main__":
    main()
