#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh

def readBelos(filename):  
    import numpy as np    
    it = np.array([],int)
    res = np.array([])
    with open(filename) as fid:
        for i, line in enumerate(fid):
            if i < 8:
                continue
            if len(line.strip()) == 0:
                break
            if line[0:4] == 'Iter':
                # get the iteration id 
                i_1 = 4
                i_2 = 4
                while line[i_2] != ',':
                    i_2 = i_2+1
                it = np.insert(it,len(it),int(line[i_1:i_2]) )    
                # get the res   
                i_1 = i_2
                i_2 = len(line)-1
                while line[i_1] != ":":
                    i_1 = i_1+1
                i_1 = i_1+1    
                res = np.insert(res,len(res),float(line[i_1:i_2]) )  

    return it, res

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"



    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos

    geo_name = 'cube.geo'
    mesh_name = 'cube.msh'


    import shlex, subprocess
    import os
    import numpy as np

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    if rank == 0:
        command_line = 'gmsh -3 ' + file_dir + '/' +  geo_name + ' -o ' + work_dir + '/' +  mesh_name +' -part ' + str(siz)
        print command_line
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log','w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
        retcode = p.wait()
        for i in range(1,siz):
            comm.send(1, dest=i, tag=11)
    else:
        data = comm.recv(source=0, tag=11)


    if siz == 1:
        msh = gmsh.MeshLoader(mesh_name,work_dir).execute()
    else:
        msh = gmsh.MeshLoader(mesh_name,work_dir).myrank_execute(rank)

    pbl = m.Problem(msh,comm)

    ensemble_size = 1

    E = 1. * np.ones(ensemble_size)
    nu = 0.2 * np.ones(ensemble_size)
    m.Medium(pbl, "Body", "test",E,nu)


    m.uDirichlet(pbl, "Surf 1", "Clamped",1,0.,1,0.,1,0.,0,0.,ensemble_size)

    dx = np.zeros(ensemble_size)
    dy = 0. * np.ones(ensemble_size)
    dz = np.zeros(ensemble_size)
    dT = np.zeros(ensemble_size)
    m.uDirichlet(pbl, "Surf 2", "Moved",0,dx,1,dy,0,dz,0,dT)

    precondList = Teuchos.ParameterList()
    precondList = Teuchos.ParameterList()
    precondSubList = Teuchos.ParameterList()
    precondSubSubList = Teuchos.ParameterList()
    precondSubSubList['fact: iluk level-of-fill'] = 2
    precondSubList['Prec Type'] = "RILUK"
    precondSubList['Overlap'] = 1
    precondSubList['Ifpack2 Settings'] = precondSubSubList
    precondList['Ifpack2'] = precondSubList

    mueluParams = Teuchos.ParameterList()
    #mueluParamsSub1 = Teuchos.ParameterList()
    #mueluParamsSub1['PDE equations'] = 4
    #mueluParams['Matrix'] = mueluParamsSub1
    mueluParams['number of equations'] = 3
    mueluParams['max levels'] = 2
    mueluParams['multigrid algorithm'] = "unsmoothed"
    mueluParams['smoother: type'] = "RELAXATION"

    mueluParamsSub2 = Teuchos.ParameterList()

    mueluParamsSub2['relaxation: type'] = "Symmetric Gauss-Seidel"
    mueluParamsSub2['relaxation: sweeps'] = 4
    mueluParamsSub2['relaxation: damping factor'] = 1.25

    mueluParams['coarse: type'] = "RELAXATION"

    mueluParams['verbosity'] = "low"

    mueluParamsSub3 = Teuchos.ParameterList()

    mueluParamsSub3['relaxation: type'] = "Symmetric Gauss-Seidel"
    mueluParamsSub3['relaxation: sweeps'] = 4
    mueluParamsSub3['relaxation: damping factor'] = 1.25


    mueluParams['smoother: params'] = mueluParamsSub2
    mueluParams['coarse: params'] = mueluParamsSub3

    solverList = Teuchos.ParameterList()
    solverList['type']= "BlockGmres"
    solverList['Convergence Tolerance'] = 10**(-8)
    solverList['Maximum Iterations'] = 100
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1
    solverList['Output Stream file name belos'] = 'multiF2Guess_belos.text'
    #solverList['Output Stream file name muelu'] = 'muelu_prec.text'
    solverList['Use preconditioner'] = False
    solverList['ifpackParams'] = precondList
    solverList['mueluParams'] = mueluParams
    solverList['Initial guess'] = 'multiF2Guess.text'
    slv = m.IterativeSolver(pbl,args.k,solverList,3, False, ensemble_size)

    d = 10
    max_it = np.zeros((d,), dtype =np.int)
    sum_res = np.zeros((d,), dtype =np.float)
    expo = np.linspace(0.72,0.73,d)

    V = np.genfromtxt("V.text", dtype=None)
    ordered_d = np.genfromtxt("ordered_d.text", dtype=None)
    K11Indices = np.genfromtxt("K11Indices.text", dtype=None)
    indices = np.genfromtxt("indices.text", dtype=None)

    filename1 = solverList['Initial guess']
    filename2 = solverList['Output Stream file name belos']

    for i in range(0,d):
        initialGuess = np.zeros(np.shape(V[:,0]))
        for j in range(0,len(K11Indices )-1):
            initialGuess = initialGuess + (np.sqrt(ordered_d[K11Indices[1]])/np.sqrt(ordered_d[K11Indices[j]]))**expo[i]*V[:,indices[K11Indices[j]]]

        np.savetxt(filename1, initialGuess, fmt='%1.4e')
        slv.start()

        timers = slv.tms;
        if rank == 0:
            it, res = readBelos(filename2)
            max_it[i] = it[len(it)-1]
            sum_res[i] = sum(res)

    if rank == 0:
        print expo
        print max_it
        print sum_res
if __name__ == "__main__":
    main()

