#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"


    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos
    import numpy as np

    preCondList = Teuchos.ParameterList()
    preCondList['type'] = "RILUK"
    preCondList['fact: iluk level-of-fill'] = 1

    solverList = Teuchos.ParameterList()
    solverList['type']= "KLU2"
    solverList['Convergence Tolerance'] = 10**(-10)
    solverList['Maximum Iterations'] = 1000
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 20
    solverList['precondList'] = preCondList

    geo_name = 'model_problem.geo'
    mesh_name = 'model_problem.msh'

    import shlex, subprocess
    import os

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    if rank == 0:
        command_line = 'gmsh -3 ' + file_dir + '/' +  geo_name + ' -o ' + work_dir + '/' +  mesh_name +' -part ' + str(siz)
        print command_line
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log','w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
        retcode = p.wait()
        for i in range(1,siz):
            comm.send(1, dest=i, tag=11)
    else:
        data = comm.recv(source=0, tag=11)


    if siz == 1:
        msh = gmsh.MeshLoader(mesh_name,work_dir).execute()
    else:
        msh = gmsh.MeshLoader(mesh_name,work_dir).myrank_execute(rank)


    pbl = m.Problem(msh,comm)

    ensemble_size = 1
    E = 330000 * np.ones(ensemble_size)
    nu = 0.36 * np.ones(ensemble_size)
    k = 0.114
    d = 0.5E-05

    m.Medium(pbl, "Holder", "TZM",E,nu,k,d)
    m.Medium(pbl, "Holder 2", "TZM",E,nu,k,d)
    m.Medium(pbl, "Mirror", "ScMo",E,nu,k,d)

    m.uDirichlet(pbl, "Holder 2 back", "Horizontal clamped line_name",1,0.,1,0.,1,0.,1,0.)
    m.uDirichlet(pbl, "Holder front", "moved",1,0.,0,0.,0,0.,1,2500.)
    m.uDirichlet(pbl, "y sym","y symetry bc",0,0.,1,0.,0,0.,0,0.)
    m.uDirichlet(pbl, "Holder below","z symetry bc",0,0.,0,0.,1,0.,0,0.)


    norm1 = tbox.Pt(-1,0,0)
    norm2 = tbox.Pt(-1,0,0)
    cont = m.Contact(pbl,"Holder back","contact",norm1)

    cont.setMaster(pbl,"Holder 2 front",norm2)

    slv = m.Solver(pbl,args.k,solverList,4)

    if not args.nogui:
        if rank == 0:
            all_msh = gmsh.MeshLoader(mesh_name,work_dir).execute()
            import mrstlnos.viewer as v
            gui = v.MeshViewer(pbl,slv,all_msh)
            gui.vmin = 0
            gui.vmax = 1
            gui.start()
        else:
            slv.start()
    else:
        slv.start()

    from fwk.testing import *
    tests = CTests()
    tests.add(CTest('mean(Tbottom)', 0, 0))
    tests.run()

if __name__ == "__main__":
    main()
