#include "mrstlnos.h"
#include "wMshData.h"
#include "wProblem.h"
#include "wMedium.h"
#include "wuDirichlet.h"
#include "wNeumann.h"
#include "wContact.h"
#include "wPt.h"
#include "IterativeSolver.h"
#include <mpi.h>

#include "wGmshImport.h"
#include "wNode.h"
#include "Discrete_problem.h"

#include <iostream>
#include <cmath>

#include <Teuchos_ParameterList.hpp>
#include <Teuchos_RCP.hpp>

#include <stdlib.h>

#include <xmmintrin.h>
#include <cmath>

#include "Belos_Tpetra_MP_Vector.hpp"
#include "BelosLinearProblem.hpp"
#include "BelosPseudoBlockGmresSolMgr.hpp"

const double pi = std::acos(-1);


/*
int main(int argc, char** argv)
{

    _MM_SET_EXCEPTION_MASK(_MM_GET_EXCEPTION_MASK()
    & ~( _MM_MASK_INVALID 
         | _MM_MASK_DIV_ZERO 
         | _MM_MASK_OVERFLOW 
//           | _MM_MASK_UNDERFLOW 
       )
    );



    using Teuchos::RCP;
    using Teuchos::rcp;

    // Initialize the MPI environment
    MPI_Init(NULL, NULL); 

    int worldsize;
    MPI_Comm_size(MPI_COMM_WORLD, &worldsize);

    int myrank;
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

    std::cout << "start" << std::endl;

    int case_id;
    if(argc > 1) 
        case_id = atoi(argv[1]);
    else
        case_id = 1;

    if( case_id == 1)
    {
        int ensemble_size;
        if(argc > 2) 
            ensemble_size = atoi(argv[2]);
        else
            ensemble_size = 1;

        int DOFperNode = 3;

        std::string mesh_name;
        mesh_name = "beam_k.msh";

        auto msh = std::make_shared<MshData>();
        if (worldsize > 1)
            msh->load(mesh_name, myrank);
        else
            msh->load(mesh_name);

        mrstlnos::Problem pbl = mrstlnos::Problem(msh, MPI_COMM_WORLD);

        double E_value = 21000.;
        double nu_value = 0.29;
        double k = 0.;
        double d = 0.;

        std::vector< double > E, nu;
        for (auto i=0; i< ensemble_size; ++i)
        {
            E.push_back(E_value);
            nu.push_back(nu_value);
        }

        auto md0 = mrstlnos::Medium(pbl, std::string("Beam"), std::string("Mat 1"),E,nu,k,d);
        auto md1 = mrstlnos::Medium(pbl, std::string("Rigid body"), std::string("Mat 2"),E,nu,k,d);

        pbl.media[0] = &md0;
        pbl.media[1] = &md1;

        double dr_value = 0.1;

        std::vector< double > zero, p;
        for (auto i=0; i< ensemble_size; ++i)
        {
            zero.push_back( (double) 0.);
            p.push_back( (double) -1.5);
        }

        auto ud0 = mrstlnos::uDirichlet(pbl, std::string("Rigid clamped surface"), std::string("Clamped 1"),1,zero,1,zero,1,zero,0,zero);
        auto ud1 = mrstlnos::uDirichlet(pbl, std::string("Clamped end"),  std::string("Clamped 1"),1,zero,1,zero,1,zero,0,zero);
        auto ud2 = mrstlnos::uDirichlet(pbl, std::string("Free end"),   std::string("Clamped 1"),1,zero,0,zero,0,zero,0,zero);

        pbl.duBCs[0] = &ud0;
        pbl.duBCs[1] = &ud1;
        pbl.duBCs[2] = &ud2;

        auto v0 = mrstlnos::Neumann(pbl, std::string("Loaded surface"), std::string("Load 1"),0,zero,0,zero,1,p,0,zero,ensemble_size);

        Pt norm1 = Pt(0,0,-1);
        Pt norm2 = Pt(0,0,1);

        auto cont1 = mrstlnos::Contact(pbl,std::string("Beam contact zone"),std::string("contact"),norm1,zero);
        cont1.setInitialyOpen();
        cont1.setMaster(pbl,std::string("Rigid surface"),norm2);

        pbl.Contacts[0] = &cont1;

        Teuchos::ParameterList solverList;
        
        solverList.set(std::string("Convergence Tolerance"),            1E-06);
        solverList.set(std::string("Maximum Iterations"),               1000);
        solverList.set(std::string("Verbosity"),                        33);
        solverList.set(std::string("Flexible Gmres"),                   false);
        solverList.set(std::string("Output Style"),                     1);
        solverList.set(std::string("Output Frequency"),                 1);
        solverList.set(std::string("MueLu xml file name"),              std::string("nested_multigrid_2_lvls.xml"));
        solverList.set(std::string("Output Style"),                     1);
        solverList.set(std::string("convert MueLu xml file"),           true);
        solverList.set(std::string("Use preconditioner"),               true);
        solverList.set(std::string("Maximum active set iteration"),     2);
        solverList.set(std::string("type"),                             std::string("BlockGmres"));
    
        // std::cout << solverList << std::endl;
        // std::cout << pbl.duBCs.size() << std::endl;
    
        auto slv = mrstlnos::IterativeSolver(pbl, 1, solverList, DOFperNode, (ensemble_size>1), ensemble_size);
        slv.start();
    } 
    else if ( case_id == 2)
    {
        int ensemble_size;
        if(argc > 2) 
            ensemble_size = atoi(argv[2]);
        else
            ensemble_size = 1;

        int DOFperNode = 3;

        std::string mesh_name;
        mesh_name = "cylinder.msh";

        auto msh = std::make_shared<MshData>();
        if (worldsize > 1)
            msh->load(mesh_name, 1-myrank);
        else
            msh->load(mesh_name);

        mrstlnos::Problem pbl = mrstlnos::Problem(msh, MPI_COMM_WORLD);

        double E_value = 21000.;
        double nu_value = 0.29;
        double k = 0.;
        double d = 0.;

        std::vector< double > E, nu;
        for (auto i=0; i< ensemble_size; ++i)
        {
            E.push_back(E_value);
            nu.push_back(nu_value);
        }

        auto md0 = mrstlnos::Medium(pbl, std::string("cylinder"), std::string("Mat 1"),E,nu,k,d);
        auto md1 = mrstlnos::Medium(pbl, std::string("block"), std::string("Mat 2"),E,nu,k,d);

        pbl.media[0] = &md0;
        pbl.media[1] = &md1;

        double dr_value = 0.1;

        std::vector< double > dr, theta, dx, dy, zero;
        for (auto i=0; i< ensemble_size; ++i)
        {
            dr.push_back(dr_value);
            theta.push_back(i*2.*pi/ensemble_size);
            dx.push_back(dr[i]*std::cos(theta[i]));
            dy.push_back(dr[i]*std::sin(theta[i]));
            zero.push_back( (double) 0.);
        }

        auto ud0 = mrstlnos::uDirichlet(pbl, std::string("clamped"), std::string("Clamped 1"),1,zero,1,zero,1,zero,0,zero);
        auto ud1 = mrstlnos::uDirichlet(pbl, std::string("sym xy"),  std::string("Clamped 1"),0,zero,0,zero,1,zero,0,zero);
        auto ud2 = mrstlnos::uDirichlet(pbl, std::string("moved"),   std::string("Clamped 1"),1,dx,  1,dy,  1,zero,0,zero);

        pbl.duBCs[0] = &ud0;
        pbl.duBCs[1] = &ud1;
        pbl.duBCs[2] = &ud2;

        Pt norm1 = Pt(1,0,0);
        Pt norm2 = Pt(1,0,0);

        auto cont1 = mrstlnos::Contact(pbl,std::string("cylinder contact"),std::string("contact"),norm2,zero);
        //cont1.setInitialyOpen();
        cont1.setMaster(pbl,std::string("block contact"),norm1);

        pbl.Contacts[0] = &cont1;

        Teuchos::ParameterList solverList;
        
        solverList.set(std::string("Convergence Tolerance"),            1E-06);
        solverList.set(std::string("Maximum Iterations"),               1000);
        solverList.set(std::string("Verbosity"),                        33);
        solverList.set(std::string("Flexible Gmres"),                   false);
        solverList.set(std::string("Output Style"),                     1);
        solverList.set(std::string("Output Frequency"),                 1);
        solverList.set(std::string("MueLu xml file name"),              std::string("nested_multigrid_2_lvls.xml"));
        solverList.set(std::string("Output Style"),                     1);
        solverList.set(std::string("convert MueLu xml file"),           true);
        solverList.set(std::string("Use preconditioner"),               true);
        solverList.set(std::string("Maximum active set iteration"),     2);
        solverList.set(std::string("type"),                             std::string("BlockGmres"));
    
        // std::cout << solverList << std::endl;
        // std::cout << pbl.duBCs.size() << std::endl;
    
        auto slv = mrstlnos::IterativeSolver(pbl, 1, solverList, DOFperNode, (ensemble_size>1), ensemble_size);
        slv.start();
    }    
    // Finalize the MPI environment.
    MPI_Finalize();

    return 0;
}
*/

template<typename scalar> void test_function(int ensemble_size)
{

    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    //typedef Tpetra::Map<> map_type;
    //typedef Tpetra::Vector<>::scalar_type scalar_type;
    //typedef Tpetra::Vector<>::mag_type magnitude_type;
    typedef Tpetra::MultiVector<scalar,local_ordinal_type,global_ordinal_type> multivector_type;
    //typedef Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> crs_matrix_type;
    //typedef Tpetra::CrsGraph<local_ordinal_type,global_ordinal_type> crs_graph_type;
    //typedef Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> vector_type;

    int worldsize;
    MPI_Comm_size(MPI_COMM_WORLD, &worldsize);

    int myrank;
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

    int DOFperNode = 3;

    std::string mesh_name;
    mesh_name = "cube.msh";


    auto msh = std::make_shared<MshData>();
    auto gmshLoader = std::make_shared<GmshImport>(msh);
    if (worldsize > 1)
        gmshLoader->load(mesh_name, myrank);
    else
        gmshLoader->load(mesh_name, -1);

    // assign a global row for each global node
    for(auto n : msh->nodes)
        n->row = n->no-1;

    //auto numNodes = msh->nodes.size();
    
    
    mrstlnos::Problem pbl(msh, MPI_COMM_WORLD);

    
    double E_value = 21000.;
    double nu_value = 0.29;
    double k = 0.;
    double d = 0.;

    std::vector< double > E, nu;
    for (auto i=0; i< ensemble_size; ++i)
    {
        E.push_back(E_value);
        nu.push_back(nu_value);
    }    

    mrstlnos::Medium md0(pbl, std::string("Body"), std::string("Mat 1"),E,nu,k,d);
    pbl.media[0] = &md0;

    std::vector< double > zero, dy;
    for (auto i=0; i< ensemble_size; ++i)
    {
        zero.push_back( (double) 0.);
        dy.push_back( (double) 1.);
    }

    mrstlnos::uDirichlet ud0(pbl, std::string("Surf 1"),  std::string("Clamped 1"),1,zero,1,zero,1,zero,0,zero);
    mrstlnos::uDirichlet ud1(pbl, std::string("Surf 2"),   std::string("Clamped 1"),0,zero,1,dy,0,zero,0,zero);

    pbl.duBCs[0] = &ud0;
    pbl.duBCs[1] = &ud1;

    Teuchos::ParameterList solverList;

    scalar tol = (scalar) 1E-06;
    solverList.set(std::string("Convergence Tolerance"),            tol);
    solverList.set(std::string("Maximum Iterations"),               1000);
    solverList.set(std::string("Verbosity"),                        33);
    solverList.set(std::string("Flexible Gmres"),                   false);
    solverList.set(std::string("Output Style"),                     1);
    solverList.set(std::string("Output Frequency"),                 1);
    //solverList.set(std::string("MueLu xml file name"),              std::string("nested_multigrid_2_lvls.xml"));
    solverList.set(std::string("Output Style"),                     1);
    //solverList.set(std::string("convert MueLu xml file"),           true);
    solverList.set(std::string("Use preconditioner"),               false);
    solverList.set(std::string("Maximum active set iteration"),     1);
    solverList.set(std::string("type"),                             std::string("BlockGmres"));

    RCP<Teuchos::ParameterList> solverList2 = rcp(new Teuchos::ParameterList(solverList));
    mrstlnos::IterativeSolver slv(pbl, 1, solverList2, DOFperNode, (ensemble_size>1), ensemble_size);
    
    auto begin = std::chrono::high_resolution_clock::now();

    Kokkos::initialize (1);

    /*
    mrstlnos::discrete_problem::Map map = mrstlnos::discrete_problem::Map(pbl,DOFperNode);
    mrstlnos::discrete_problem::Domain<scalar> domain = mrstlnos::discrete_problem::Domain<scalar>(pbl,map);
    mrstlnos::discrete_problem::Loads<scalar> loads = mrstlnos::discrete_problem::Loads<scalar> (pbl,map,domain);
    mrstlnos::discrete_problem::Graph graph = mrstlnos::discrete_problem::Graph(map, domain.elements_list);
    typedef Tpetra::CrsMatrix<>::node_type node_type;
    auto AWO = rcp(new Tpetra::CrsMatrix<scalar,int,int,node_type> (graph.AWO));
    */
    
    //mrstlnos::discrete_problem::Algebraic<scalar> algebraic = mrstlnos::discrete_problem::Algebraic<scalar> (map, domain, loads);
    //auto AWO = rcp(new Tpetra::CrsMatrix<scalar,int,int> (map.mapDofs));

    scalar a[0];
    mrstlnos::Discrete_problem<scalar> discrete_problem = mrstlnos::Discrete_problem<scalar>(pbl, DOFperNode, Teuchos::null, a);
    discrete_problem.computeMatrices();
    discrete_problem.computeLoads();
    //std::string name_0;
    //name_0  = "matrix"+ std::to_string(ensemble_size)+ ".text";
    //discrete_problem.algebraic->writeCrsMatrixToFile(name_0,myrank);

    //RCP<vector_type> r (new vector_type(discrete_problem.algebraic->map.mapDofs, true));

    //discrete_problem.algebraic->matrices->A->apply(*(discrete_problem.algebraic->vectors->x),*r,Teuchos::NO_TRANS,(scalar) -1.,(scalar) 1.);

    
    typedef multivector_type  MV;
    typedef typename multivector_type::dot_type belos_scalar;
    typedef Tpetra::Operator<scalar >     OP;
    typedef Belos::LinearProblem<belos_scalar,MV,OP> BLinProb;
    RCP< BLinProb > blinproblem = rcp(new BLinProb(discrete_problem.algebraic->matrices->A, discrete_problem.algebraic->vectors->x, discrete_problem.algebraic->vectors->b));
    blinproblem->setProblem();
    RCP<Teuchos::ParameterList> belosParams = rcp(new Teuchos::ParameterList(solverList));
    RCP<Belos::SolverManager<belos_scalar,MV,OP> > blinsolver =
                rcp(new Belos::PseudoBlockGmresSolMgr<belos_scalar,MV,OP>(blinproblem, belosParams));
    //Belos::ReturnType ret = blinsolver->solve();
    blinsolver->solve();  
    
    auto end = std::chrono::high_resolution_clock::now();
    double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end-begin).count();
    std::cout << ensemble_size << " " << total_time << " nanosec" << std::endl;
    Kokkos::finalize ();
}

template<typename scalar> void test_function2(int ensemble_size)
{

    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    //typedef Tpetra::Map<> map_type;
    //typedef Tpetra::Vector<>::scalar_type scalar_type;
    //typedef Tpetra::Vector<>::mag_type magnitude_type;
    //typedef Tpetra::MultiVector<scalar,local_ordinal_type,global_ordinal_type> multivector_type;
    //typedef Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> crs_matrix_type;
    //typedef Tpetra::CrsGraph<local_ordinal_type,global_ordinal_type> crs_graph_type;
    //typedef Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> vector_type;

    int worldsize;
    MPI_Comm_size(MPI_COMM_WORLD, &worldsize);

    int myrank;
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

    int DOFperNode = 3;

    std::string mesh_name;
    mesh_name = "cube.msh";


    auto msh = std::make_shared<MshData>();
    auto gmshLoader = std::make_shared<GmshImport>(msh);
    if (worldsize > 1)
        gmshLoader->load(mesh_name, myrank);
    else
        gmshLoader->load(mesh_name, -1);

    // assign a global row for each global node
    for(auto n : msh->nodes)
        n->row = n->no-1;

    //auto numNodes = msh->nodes.size();
    
    
    mrstlnos::Problem pbl(msh, MPI_COMM_WORLD);

    
    double E_value = 21000.;
    double nu_value = 0.29;
    double k = 0.;
    double d = 0.;

    std::vector< double > E, nu;
    for (auto i=0; i< ensemble_size; ++i)
    {
        E.push_back(E_value);
        nu.push_back(nu_value);
    }    

    mrstlnos::Medium md0(pbl, std::string("Body"), std::string("Mat 1"),E,nu,k,d);
    pbl.media[0] = &md0;

    std::vector< double > zero, dy;
    for (auto i=0; i< ensemble_size; ++i)
    {
        zero.push_back( (double) 0.);
        dy.push_back( (double) 1.);
    }

    mrstlnos::uDirichlet ud0(pbl, std::string("Surf 1"),  std::string("Clamped 1"),1,zero,1,zero,1,zero,0,zero);
    mrstlnos::uDirichlet ud1(pbl, std::string("Surf 2"),   std::string("Clamped 1"),0,zero,1,dy,0,zero,0,zero);

    pbl.duBCs[0] = &ud0;
    pbl.duBCs[1] = &ud1;

    Teuchos::ParameterList solverList;

    solverList.set(std::string("Ensemble Convergence Tolerance"),            1E-06);
    solverList.set(std::string("Maximum Iterations"),               1000);
    solverList.set(std::string("Verbosity"),                        33);
    solverList.set(std::string("Flexible Gmres"),                   false);
    solverList.set(std::string("Output Style"),                     1);
    solverList.set(std::string("Output Frequency"),                 1);
    //solverList.set(std::string("MueLu xml file name"),              std::string("nested_multigrid_2_lvls.xml"));
    solverList.set(std::string("Output Style"),                     1);
    //solverList.set(std::string("convert MueLu xml file"),           true);
    solverList.set(std::string("Use preconditioner"),               false);
    solverList.set(std::string("Maximum active set iteration"),     1);
    solverList.set(std::string("type"),                             std::string("BlockGmres"));

    RCP<Teuchos::ParameterList> solverList2 = rcp(new Teuchos::ParameterList(solverList));
    mrstlnos::IterativeSolver slv(pbl, 1, solverList2, DOFperNode, (ensemble_size>1), ensemble_size);
    

    auto begin2 = std::chrono::high_resolution_clock::now();
    slv.start();
    auto end2 = std::chrono::high_resolution_clock::now();
    double total_time2 = std::chrono::duration_cast<std::chrono::nanoseconds>(end2-begin2).count();
    std::cout << ensemble_size << " " << total_time2 << " nanosec" << std::endl;
}

int main(int argc, char** argv)
{
    int case_id;
    if(argc > 1) 
        case_id = atoi(argv[1]);
    else
        case_id = 1;

    MPI_Init(NULL, NULL); 
    //Kokkos::initialize (1);

    {
    const int ensemble_size = 1;
    typedef double scalar;
    if (case_id ==1)        
    test_function<scalar>(ensemble_size);
    else
    test_function2<scalar>(ensemble_size);
    }
    {
    const int ensemble_size = 8;
    typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;
    if (case_id ==1)        
    test_function<scalar>(ensemble_size);
    else
    test_function2<scalar>(ensemble_size);
    }
    {
    const int ensemble_size = 16;
    typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;
    if (case_id ==1)        
    test_function<scalar>(ensemble_size);
    else
    test_function2<scalar>(ensemble_size);
    }
    //Kokkos::finalize ();
    MPI_Finalize();
}
