#include "Stokhos_Sacado_Kokkos_MP_Vector.hpp"
//#include "Teuchos_ScalarTraits.hpp"
#include "EnsembleTraits.h"
#include <iostream>
#include <mkl.h>
#include <utility>

//#include "Quantile.h"
#include "InputParser.h"

#include "ROTG_decl.hpp"

using Teuchos::ScalarTraits;


template <typename ScalarType> void set_da_db
(ScalarType &da, ScalarType &db, ScalarType &c, ScalarType &s) {
    
    typedef EnsembleTraits<ScalarType> ET;
    typedef ScalarTraits<ScalarType> STS;
    
    da = 1.589*STS::one();
    db = -3.685*STS::one();
    
    if(ET::size > 4)
    {
        ET::coeff(da,2) = 0.;
        ET::coeff(db,2) = 0.;
      
        ET::coeff(db,3) = 0.;
        
        ET::coeff(da,4) = 0.;
    }
}

template <typename ScalarType> void test_da_db
(ScalarType &da, ScalarType &db, ScalarType &c, ScalarType &s) {
    
    std::cout<< da << " " << db << " " << c << " " << s << std::endl;
}

void test_function_double(std::string filename, int N)
{
    typedef double scalar;
    typedef EnsembleTraits<scalar> ET;
    scalar da, db, c, s;
    
    set_da_db(da,db,c,s);
    
    std::ofstream timers;
    std::string name_file;
    name_file = filename+"_"+std::to_string(ET::size)+".txt";
    timers.open(name_file);
    timers << "N " << N << std::endl;

    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            cblas_drotg(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "cblas_drotg "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
    set_da_db(da,db,c,s);
    cblas_drotg(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_ensemble_traits(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_ensemble_traits "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
    set_da_db(da,db,c,s);
    ROTG_ensemble_traits(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_ensemble_traits_pragma(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_ensemble_traits_pragma "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
    set_da_db(da,db,c,s);
    ROTG_ensemble_traits_pragma(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_as_double(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_as_double "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
    
    set_da_db(da,db,c,s);
    ROTG_mask_as_double(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
  
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_1(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_1 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
    
    set_da_db(da,db,c,s);
    ROTG_mask_assign_1(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_2(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_2 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
    
    set_da_db(da,db,c,s);
    ROTG_mask_assign_2(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_3(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_3 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
  
    set_da_db(da,db,c,s);
    ROTG_mask_assign_3(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_4(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_4 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
  
    set_da_db(da,db,c,s);
    ROTG_mask_assign_4(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_5(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_5 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
  
    set_da_db(da,db,c,s);
    ROTG_mask_assign_5(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_6(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_6 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
  
    set_da_db(da,db,c,s);
    ROTG_mask_assign_6(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_7(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_7 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
  
    set_da_db(da,db,c,s);
    ROTG_mask_assign_7(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
  
    timers.close();
}

template<typename scalar> void test_function(std::string filename, int N)
{
/*
  const unsigned long knl_features =
      (_FEATURE_AVX512F | _FEATURE_AVX512ER |
       _FEATURE_AVX512PF | _FEATURE_AVX512CD );
 */
    typedef EnsembleTraits<scalar> ET;
    scalar da, db, c, s;
    
    set_da_db(da,db,c,s);
    
    std::ofstream timers;
    std::string name_file;
    name_file = filename+"_"+std::to_string(ET::size)+".txt";
    timers.open(name_file);
    timers << "N " << N << std::endl;
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_ensemble_traits(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_ensemble_traits "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
    set_da_db(da,db,c,s);
    ROTG_ensemble_traits(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_ensemble_traits_pragma(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_ensemble_traits_pragma "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
    set_da_db(da,db,c,s);
    ROTG_ensemble_traits_pragma(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_as_double(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_as_double "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
    
    set_da_db(da,db,c,s);
    ROTG_mask_as_double(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_1(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_1 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
    
    set_da_db(da,db,c,s);
    ROTG_mask_assign_1(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_2(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_2 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
    
    set_da_db(da,db,c,s);
    ROTG_mask_assign_2(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_3(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_3 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
  
    set_da_db(da,db,c,s);
    ROTG_mask_assign_3(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_4(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_4 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
  
    set_da_db(da,db,c,s);
    ROTG_mask_assign_4(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_5(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_5 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
  
    set_da_db(da,db,c,s);
    ROTG_mask_assign_5(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_6(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_6 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
  
    set_da_db(da,db,c,s);
    ROTG_mask_assign_6(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    {
        auto begin_time = std::chrono::high_resolution_clock::now();
        for(int i=0; i<N; ++i)
        {
            ROTG_mask_assign_7(&da,&db,&c,&s);
        }
        auto end_time = std::chrono::high_resolution_clock::now();
        double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
        timers << "ROTG_mask_assign_7 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
    }
  
    set_da_db(da,db,c,s);
    ROTG_mask_assign_7(&da,&db,&c,&s);
    test_da_db(da,db,c,s);
    set_da_db(da,db,c,s);
    
    if(0) {//if ( _may_i_use_cpu_feature( knl_features ) ){
      set_da_db(da,db,c,s);
      {
          auto begin_time = std::chrono::high_resolution_clock::now();
          for(int i=0; i<N; ++i)
          {
              ROTG_Intel_Intrinsics(&da,&db,&c,&s);
          }
          auto end_time = std::chrono::high_resolution_clock::now();
          double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
          timers << "ROTG_Intel_Intrinsics "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
      }
      
      set_da_db(da,db,c,s);
      ROTG_Intel_Intrinsics(&da,&db,&c,&s);
      test_da_db(da,db,c,s);
      set_da_db(da,db,c,s);
      {
          auto begin_time = std::chrono::high_resolution_clock::now();
          for(int i=0; i<N; ++i)
          {
              ROTG_Intel_Intrinsics_2(&da,&db,&c,&s);
          }
          auto end_time = std::chrono::high_resolution_clock::now();
          double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
          timers << "ROTG_Intel_Intrinsics_2 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
      }
      
      set_da_db(da,db,c,s);
      ROTG_Intel_Intrinsics_2(&da,&db,&c,&s);
      test_da_db(da,db,c,s);
      set_da_db(da,db,c,s);
      {
          auto begin_time = std::chrono::high_resolution_clock::now();
          for(int i=0; i<N; ++i)
          {
              ROTG_Intel_Intrinsics_3(&da,&db,&c,&s);
          }
          auto end_time = std::chrono::high_resolution_clock::now();
          double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
          timers << "ROTG_Intel_Intrinsics_3 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
      }
    
      set_da_db(da,db,c,s);
      ROTG_Intel_Intrinsics_3(&da,&db,&c,&s);
      test_da_db(da,db,c,s);
      set_da_db(da,db,c,s);
      {
          auto begin_time = std::chrono::high_resolution_clock::now();
          for(int i=0; i<N; ++i)
          {
              ROTG_Intel_Intrinsics_4(&da,&db,&c,&s);
          }
          auto end_time = std::chrono::high_resolution_clock::now();
          double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
          timers << "ROTG_Intel_Intrinsics_4 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
      }
    
      set_da_db(da,db,c,s);
      ROTG_Intel_Intrinsics_4(&da,&db,&c,&s);
      test_da_db(da,db,c,s);
      set_da_db(da,db,c,s);
      {
          auto begin_time = std::chrono::high_resolution_clock::now();
          for(int i=0; i<N; ++i)
          {
              ROTG_Intel_Intrinsics_5(&da,&db,&c,&s);
          }
          auto end_time = std::chrono::high_resolution_clock::now();
          double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
          timers << "ROTG_Intel_Intrinsics_5 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
      }
    
      set_da_db(da,db,c,s);
      ROTG_Intel_Intrinsics_5(&da,&db,&c,&s);
      test_da_db(da,db,c,s);
      set_da_db(da,db,c,s);
      {
          auto begin_time = std::chrono::high_resolution_clock::now();
          for(int i=0; i<N; ++i)
          {
              ROTG_Intel_Intrinsics_6(&da,&db,&c,&s);
          }
          auto end_time = std::chrono::high_resolution_clock::now();
          double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
          timers << "ROTG_Intel_Intrinsics_6 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
      }
    
      set_da_db(da,db,c,s);
      ROTG_Intel_Intrinsics_6(&da,&db,&c,&s);
      test_da_db(da,db,c,s);
      set_da_db(da,db,c,s);
      {
          auto begin_time = std::chrono::high_resolution_clock::now();
          for(int i=0; i<N; ++i)
          {
              ROTG_Intel_Intrinsics_7(&da,&db,&c,&s);
          }
          auto end_time = std::chrono::high_resolution_clock::now();
          double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
          timers << "ROTG_Intel_Intrinsics_7 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
      }
    
      set_da_db(da,db,c,s);
      ROTG_Intel_Intrinsics_7(&da,&db,&c,&s);
      test_da_db(da,db,c,s);
      set_da_db(da,db,c,s);
      {
          auto begin_time = std::chrono::high_resolution_clock::now();
          for(int i=0; i<N; ++i)
          {
              ROTG_Intel_Intrinsics_8(&da,&db,&c,&s);
          }
          auto end_time = std::chrono::high_resolution_clock::now();
          double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
          timers << "ROTG_Intel_Intrinsics_8 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
      }
    
      set_da_db(da,db,c,s);
      ROTG_Intel_Intrinsics_8(&da,&db,&c,&s);
      test_da_db(da,db,c,s);
      set_da_db(da,db,c,s);
      {
          auto begin_time = std::chrono::high_resolution_clock::now();
          for(int i=0; i<N; ++i)
          {
              ROTG_Intel_Intrinsics_9(&da,&db,&c,&s);
          }
          auto end_time = std::chrono::high_resolution_clock::now();
          double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
          timers << "ROTG_Intel_Intrinsics_9 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
      }
    
      set_da_db(da,db,c,s);
      ROTG_Intel_Intrinsics_9(&da,&db,&c,&s);
      test_da_db(da,db,c,s);
      set_da_db(da,db,c,s);
      {
          auto begin_time = std::chrono::high_resolution_clock::now();
          for(int i=0; i<N; ++i)
          {
              ROTG_Intel_Intrinsics_10(&da,&db,&c,&s);
          }
          auto end_time = std::chrono::high_resolution_clock::now();
          double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end_time-begin_time).count();
          timers << "ROTG_Intel_Intrinsics_10 "<< (total_time/1e09) << " " << (total_time/(N*1e09))  << std::endl;
      }
    
      set_da_db(da,db,c,s);
      ROTG_Intel_Intrinsics_10(&da,&db,&c,&s);
      test_da_db(da,db,c,s);
      set_da_db(da,db,c,s);
    }
    timers.close();
}


int main(int argc, char **argv){
    
    std::string filename;
    int N;
    
    readInput(argc, argv, filename, N);
    
    if( N == -1)
        return 0;
    
    using namespace std;
    {
        test_function_double(filename,N);
    }
    {
        constexpr int ensemble_size = 8;
        typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;
        test_function<scalar>(filename,N);
    }
    {
        constexpr int ensemble_size = 16;
        typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;
        test_function<scalar>(filename,N);
    }
    {
        constexpr int ensemble_size = 32;
        typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;
        test_function<scalar>(filename,N);
    }

    return 0;
}
