// Tpetra
#include "Stokhos_Tpetra_MP_Vector.hpp"
#include "Stokhos_Tpetra_Utilities_MP_Vector.hpp"
#include "Tpetra_ConfigDefs.hpp"
#include "Tpetra_DefaultPlatform.hpp"
#include "Tpetra_Map.hpp"
#include "Tpetra_MultiVector.hpp"
#include "Tpetra_Vector.hpp"
#include "Tpetra_CrsGraph.hpp"
#include "Tpetra_CrsMatrix.hpp"
#include "Stokhos_Tpetra_CG.hpp"

#include <MatrixMarket_Tpetra.hpp>

#include <mpi.h>

#include <iostream>
#include <cmath>

#include <Teuchos_ParameterList.hpp>
#include <Teuchos_RCP.hpp>

#include <stdlib.h>

#include "Belos_Tpetra_MP_Vector.hpp"
#include "BelosLinearProblem.hpp"
#include "BelosPseudoBlockGmresSolMgr.hpp"

#include "Belos_PseudoBlockGmresIter_MP_Vector.hpp"
#include "Belos_StatusTest_ImpResNorm_MP_Vector.hpp"
#include "Belos_DGKSOrthoManager_MP_Vector.hpp"


#include "EnsembleTraits.h"

#include "BLAS_MP_Vector.hpp"

const double pi = std::acos(-1);

void  run_test_double()
{
    typedef double scalar;
    typedef EnsembleTraits<scalar> ET;
    const int ensemble_size = ET::size;

    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    typedef Tpetra::Map<> map_type;
    typedef Tpetra::Vector<>::scalar_type scalar_type;
    typedef Tpetra::Vector<>::mag_type magnitude_type;
    typedef Tpetra::MultiVector<scalar,local_ordinal_type,global_ordinal_type> multivector_type;
    typedef Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> crs_matrix_type;
    typedef Tpetra::CrsGraph<local_ordinal_type,global_ordinal_type> crs_graph_type;

    typedef Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> vector_type;

    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;

    int worldsize;
    MPI_Comm_size(MPI_COMM_WORLD, &worldsize);

    int myrank;
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

    RCP<const Teuchos::Comm<int> > comm = rcp (new Teuchos::MpiComm<int> (MPI_COMM_WORLD));

    Teuchos::ParameterList solverList;

    scalar tol = (scalar) 1E-06;
    solverList.set(std::string("Convergence Tolerance"),            tol);
    solverList.set(std::string("Maximum Iterations"),               10);
    solverList.set(std::string("Verbosity"),                        33);
    solverList.set(std::string("Flexible Gmres"),                   false);
    solverList.set(std::string("Output Style"),                     1);
    solverList.set(std::string("Output Frequency"),                 1);
  
    const int n = 3;
  
    Kokkos::initialize (1);
    {
        for (int index_i = 0; index_i <8; ++index_i)
        {
            const global_ordinal_type indexBase = 0;

            RCP<const map_type> mapNodes = rcp (new map_type (n, n, indexBase, comm));


            RCP<crs_graph_type> graphA (new crs_graph_type (mapNodes, 0));
            for ( global_ordinal_type i = 0; i<n; ++i)
                for ( global_ordinal_type j = 0; j<n; ++j)
                    graphA->insertGlobalIndices(i, tuple<global_ordinal_type> (j));
          
            graphA->fillComplete ();

            RCP<crs_matrix_type> A (new crs_matrix_type (graphA));

            A->replaceGlobalValues(0,tuple<global_ordinal_type> (0,1,2), tuple<scalar> (1,0,0));
            A->replaceGlobalValues(1,tuple<global_ordinal_type> (0,1,2), tuple<scalar> (1,2,0));
            A->replaceGlobalValues(2,tuple<global_ordinal_type> (0,1,2), tuple<scalar> (2,3,3));
        
            A->fillComplete ();

            RCP<vector_type> x (new vector_type(mapNodes, true));

            RCP<vector_type> b (new vector_type(mapNodes, true));
          
            scalar b_value = (scalar) 0;
            if (index_i==0)
                b_value = 1.;
          
            b->sumIntoGlobalValue(0,b_value);
          
            b_value = (scalar) 1;
            if (index_i==1)
                b_value = 0.;
            if (index_i==2)
                b_value = 0.;
          
            b->sumIntoGlobalValue(1,b_value);

            b_value = (scalar) 2;
            if (index_i==2)
                b_value = 0.;
          
            b->sumIntoGlobalValue(2,b_value);

            typedef multivector_type  MV;
            typedef typename multivector_type::dot_type belos_scalar;
            typedef Tpetra::Operator<scalar >     OP;

            typedef Belos::LinearProblem<belos_scalar,MV,OP> BLinProb;
            RCP< BLinProb > blinproblem = rcp(new BLinProb(A, x, b));
            blinproblem->setProblem();
          
            RCP<Teuchos::ParameterList> belosParams = rcp(new Teuchos::ParameterList(solverList));
          
            belosParams->set("Orthogonalization","DGKS");
          
            RCP<Belos::PseudoBlockGmresSolMgr<belos_scalar,MV,OP> > blinsolver =
                        rcp(new Belos::PseudoBlockGmresSolMgr<belos_scalar,MV,OP>(blinproblem, belosParams));
            Belos::ReturnType ret = blinsolver->solve();
          
            auto tmp = blinsolver->getResidualStatusTest()->getEnsembleIterations();
            for (auto in : tmp)
              std::cout << in << " ";
            std::cout << std::endl;
            std::cout << tmp.size() << std::endl;
          
            //std::cout << tmp << std::endl;
            //std::cout << tmp->getStatus() << std::endl;
            //Teuchos::RCP<Belos::StatusTest<belos_scalar,MV,OP>> tmp2 = static_cast<Teuchos::RCP<Belos::StatusTest<belos_scalar,MV,OP>>>(tmp);
            //std::vector<int> ensemble_iterations = static_cast<Teuchos::RCP<Belos::StatusTestImpResNorm<belos_scalar,MV,OP>>>(blinsolver->getResidualStatusTest())->getEnsembleIterations();

            Tpetra::MatrixMarket::Writer< crs_matrix_type >::writeSparseFile(std::string("A_" + std::to_string(index_i)+".txt"),A);
            Tpetra::MatrixMarket::Writer< crs_matrix_type >::writeDenseFile(std::string("b_" + std::to_string(index_i)+".txt"),b);
            Tpetra::MatrixMarket::Writer< crs_matrix_type >::writeDenseFile(std::string("x_" + std::to_string(index_i)+".txt"),x);
        
        }
    
    }
    Kokkos::finalize ();
}

template<typename scalar> void  run_test()
{
    typedef EnsembleTraits<scalar> ET;
    const int ensemble_size = ET::size;

    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    typedef Tpetra::Map<> map_type;
    typedef Tpetra::Vector<>::scalar_type scalar_type;
    typedef Tpetra::Vector<>::mag_type magnitude_type;
    typedef Tpetra::MultiVector<scalar,local_ordinal_type,global_ordinal_type> multivector_type;
    typedef Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> crs_matrix_type;
    typedef Tpetra::CrsGraph<local_ordinal_type,global_ordinal_type> crs_graph_type;

    typedef Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> vector_type;

    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;

    int worldsize;
    MPI_Comm_size(MPI_COMM_WORLD, &worldsize);

    int myrank;
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

    RCP<const Teuchos::Comm<int> > comm = rcp (new Teuchos::MpiComm<int> (MPI_COMM_WORLD));

    Teuchos::ParameterList solverList;
  
    magnitude_type tol = (magnitude_type) 1E-06;
    solverList.set(std::string("Convergence Tolerance"),            tol);
    solverList.set(std::string("Maximum Iterations"),               10);
    solverList.set(std::string("Verbosity"),                        33);
    solverList.set(std::string("Flexible Gmres"),                   false);
    solverList.set(std::string("Output Style"),                     1);
    solverList.set(std::string("Output Frequency"),                 1);
  
    const int n = 3;
  
    Kokkos::initialize (1);
    {
        const global_ordinal_type indexBase = 0;

        RCP<const map_type> mapNodes = rcp (new map_type (n, n, indexBase, comm));


        RCP<crs_graph_type> graphA (new crs_graph_type (mapNodes, 0));
        for ( global_ordinal_type i = 0; i<n; ++i)
            for ( global_ordinal_type j = 0; j<n; ++j)
                graphA->insertGlobalIndices(i, tuple<global_ordinal_type> (j));
        
        graphA->fillComplete ();

        RCP<crs_matrix_type> A (new crs_matrix_type (graphA));

        A->replaceGlobalValues(0,tuple<global_ordinal_type> (0,1,2), tuple<scalar> (1,0,0));
        A->replaceGlobalValues(1,tuple<global_ordinal_type> (0,1,2), tuple<scalar> (1,2,0));
        A->replaceGlobalValues(2,tuple<global_ordinal_type> (0,1,2), tuple<scalar> (2,3,3));
    
        A->fillComplete ();

        RCP<vector_type> x (new vector_type(mapNodes, true));

        RCP<vector_type> b (new vector_type(mapNodes, true));
      
        scalar b_value = (scalar) 0;
        b_value[0] = 1.;
      
        b->sumIntoGlobalValue(0,b_value);
      
        b_value = (scalar) 1;
        b_value[1] = 0.;
        b_value[2] = 0.;
      
        b->sumIntoGlobalValue(1,b_value);

        b_value = (scalar) 2;
        b_value[2] = 0.;
      
        b->sumIntoGlobalValue(2,b_value);

        typedef multivector_type  MV;
        typedef typename multivector_type::dot_type belos_scalar;
        typedef Tpetra::Operator<scalar >     OP;

        typedef Belos::LinearProblem<belos_scalar,MV,OP> BLinProb;
        RCP< BLinProb > blinproblem = rcp(new BLinProb(A, x, b));
        blinproblem->setProblem();
      
        RCP<Teuchos::ParameterList> belosParams = rcp(new Teuchos::ParameterList(solverList));
      
        belosParams->set("Orthogonalization","DGKS");
      
        RCP<Belos::PseudoBlockGmresSolMgr<belos_scalar,MV,OP> > blinsolver =
                    rcp(new Belos::PseudoBlockGmresSolMgr<belos_scalar,MV,OP>(blinproblem, belosParams));
        //Belos::ReturnType ret = blinsolver->solve();
        blinsolver->solve();
      
        auto tmp = blinsolver->getResidualStatusTest()->getEnsembleIterations();
        for (auto in : tmp)
          std::cout << in << " ";
        std::cout << std::endl;
        std::cout << tmp.size() << std::endl;
        
        //std::cout << tmp << std::endl;
        //std::cout << tmp->getStatus() << std::endl;
        //Teuchos::RCP<Belos::StatusTest<belos_scalar,MV,OP>> tmp2 = static_cast<Teuchos::RCP<Belos::StatusTest<belos_scalar,MV,OP>>>(tmp);
        //std::vector<int> ensemble_iterations = static_cast<Teuchos::RCP<Belos::StatusTestImpResNorm<belos_scalar,MV,OP>>>(blinsolver->getResidualStatusTest())->getEnsembleIterations();

        Tpetra::MatrixMarket::Writer< crs_matrix_type >::writeSparseFile(std::string("A_" + std::to_string(ensemble_size)+".txt"),A);
        Tpetra::MatrixMarket::Writer< crs_matrix_type >::writeDenseFile(std::string("b_" + std::to_string(ensemble_size)+".txt"),b);
        Tpetra::MatrixMarket::Writer< crs_matrix_type >::writeDenseFile(std::string("x_" + std::to_string(ensemble_size)+".txt"),x);
    
    }
    Kokkos::finalize ();
}

int main(int argc, char** argv)
{
    MPI_Init(NULL, NULL);

    run_test_double();
  
    const int ensemble_size = 8;
    typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;

    run_test<scalar>();

    MPI_Finalize();
}
