
#include <iostream>
#include "Kokkos_Core.hpp"
#include "Stokhos_Sacado_Kokkos_MP_Vector.hpp"
using Teuchos::ScalarTraits;

int main(int argc, char** argv)
{
  constexpr int ensemble_size = 8;
  typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;
  typedef ScalarTraits<scalar> STS;
  
  const size_t scalar_size = sizeof(Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>>);
  
  std::cout << scalar_size << " " << sizeof(double) << std::endl;
  
  typedef Kokkos::DefaultExecutionSpace::scratch_memory_space
  ScratchSpace;
  // Define a view type in ScratchSpace
  typedef Kokkos::View<scalar*[4],ScratchSpace,
          Kokkos::MemoryTraits<Kokkos::Unmanaged>> ScalarViewType;
  
  Kokkos::initialize (argc,argv);
  {
    int N = 4;
    
    int NA = 1;

    int team_size = 4;
    int workset_size = 20;
    printf("----------\n");

    size_t shared_size = ScalarViewType::shmem_size(NA,team_size);
    
    using Kokkos::TeamThreadRange;

    Kokkos::TeamPolicy<> policy = Kokkos::TeamPolicy<>( N, team_size).set_scratch_size(0, Kokkos::PerTeam(shared_size));
    typedef Kokkos::TeamPolicy<>::member_type member_type;
    parallel_for (policy, KOKKOS_LAMBDA (member_type team_member) {
      ScalarViewType team_shared_a(team_member.team_scratch(0),NA,team_size);
      Kokkos::deep_copy(team_shared_a, STS::zero());
      
      if(team_member.team_rank ()==0)
        printf("1. [%f , %f, %f, %f, %f];\n", team_shared_a(0,0)[0], team_shared_a(0,0)[1], team_shared_a(0,0)[2], team_shared_a(0,0)[3], team_shared_a(0,0)[4]);

      parallel_for (TeamThreadRange (team_member, workset_size),[&] (int& j){
        printf("%d , %d, %d, %d, %d;\n", team_member.league_rank (),team_member.team_size (),team_member.team_rank (), Kokkos::DefaultExecutionSpace::thread_pool_rank(), j);
        for (int i=0; i< NA; ++i)
          team_shared_a(i,team_member.team_rank ()) += team_member.league_rank () * STS::one();
      });
      team_member.team_barrier();
      parallel_for (TeamThreadRange (team_member, NA),[&] (int& j){
        for (int i=1; i< team_member.team_size (); ++i)
          team_shared_a(0,j) += team_shared_a(i,j);
      });
      team_member.team_barrier();
      if(team_member.team_rank ()==0)
        printf("2. [%f , %f, %f, %f, %f];\n", team_shared_a(0,0)[0], team_shared_a(0,0)[1], team_shared_a(0,0)[2], team_shared_a(0,0)[3], team_shared_a(0,0)[4]);
    });
    printf("----------\n");
  }
  Kokkos::finalize();
  
  return 0;
}
