//#define __KOKKOSBATCHED_INTEL_MKL__
//#define __KOKKOSBATCHED_INTEL_MKL_BATCHED__
//#define __KOKKOSBATCHED_INTEL_MKL_COMPACT_BATCHED__

//#include "mkl.h"

#include <iostream>
#include "Stokhos_Sacado_Kokkos_MP_Vector.hpp"
#include "Kokkos_Core.hpp"

#include "KokkosBatched_Vector.hpp"

#include "KokkosBatched_Gemm_Decl.hpp"
#include "KokkosBatched_Gemm_Serial_Impl.hpp"

#include "TpetraCore_config.h"
#include "Kokkos_ArithTraits.hpp"
#include "Kokkos_Complex.hpp"
#include <type_traits>



#include "EnsembleTraits.h"

// Tpetra
#include "Stokhos_Tpetra_MP_Vector.hpp"
#include "Stokhos_Tpetra_Utilities_MP_Vector.hpp"
#include "Tpetra_ConfigDefs.hpp"
#include "Tpetra_DefaultPlatform.hpp"
#include "Tpetra_Map.hpp"
#include "Tpetra_MultiVector.hpp"
#include "Tpetra_Vector.hpp"
#include "Tpetra_CrsGraph.hpp"
#include "Tpetra_CrsMatrix.hpp"
#include "Stokhos_Tpetra_CG.hpp"


//#include "Tpetra_Details_defaultGemm.hpp"
//#include "Tpetra_Details_defaultGemm_MP_Vector.hpp"





//#include "KokkosBatched_Gemm_Decl.hpp"
//#include "KokkosBatched_Gemm_Serial_Impl.hpp"


namespace { // (anonymous)

  template<class ValueType>
  ValueType
  conditionalConjugate (const ValueType& x, const bool conj)
  {
    return conj ? Kokkos::Details::ArithTraits<ValueType>::conj (x) : x;
  }

} // namespace (anonymous)



template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class CoefficientType,
         class IndexType = int>
void
test_gemm_default (const char transA,
      const char transB,
      const CoefficientType& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const CoefficientType& beta,
      const ViewType3& C)
{

  //std::cout<< "gemm 1" << std::endl;
  // Assert that A, B, and C are in fact matrices
  static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
  static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
  static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

  typedef typename ViewType3::non_const_value_type c_value_type;
  typedef Kokkos::Details::ArithTraits<CoefficientType> STS;
  const CoefficientType ZERO = STS::zero ();
  const CoefficientType ONE = STS::one ();

  // Get the dimensions
  const IndexType m = C.dimension_0 ();
  const IndexType n = C.dimension_1 ();
  const IndexType k = (transA == 'N' || transA == 'n') ?
    A.dimension_1 () : A.dimension_0 ();

  //std::cout << "m = " << m << ", n = "<< n << ", k = " << k << std::endl;
  //std::cout << "A not transposed: " << (transA == 'N' || transA == 'n') << std::endl;
  //std::cout << "B not transposed: " << (transB == 'N' || transB == 'n') << std::endl;

  const bool conjA = transA == 'C' || transA == 'c';
  const bool conjB = transB == 'C' || transB == 'c';

  // quick return if possible
  if (alpha == ZERO && beta == ONE) {
    return;
  }

  // And if alpha equals zero...
  if (alpha == ZERO) {
    if (beta == ZERO) {
      for (IndexType i = 0; i < m; ++i) {
        for (IndexType j = 0; j < n; ++j) {
          C(i,j) = ZERO;
        }
      }
    }
    else {
      for (IndexType i = 0; i < m; ++i) {
        for (IndexType j = 0; j < n; ++j) {
          C(i,j) = beta*C(i,j);
        }
      }
    }
    return;
  }
  // Start the operations
  if (transB == 'n' || transB == 'N') {
    if (transA == 'n' || transA == 'N') {
      // Form C = alpha*A*B + beta*C
      for (IndexType j = 0; j < n; ++j) {
        if (beta == ZERO) {
          for (IndexType i = 0; i < m; ++i) {
            C(i,j) = ZERO;
          }
        }
        else if (beta != ONE) {
          for (IndexType i = 0; i < m; ++i) {
            C(i,j) = beta*C(i,j);
          }
        }
        for (IndexType l = 0; l < k; ++l) {
          // Don't use c_value_type here, since it unnecessarily
          // forces type conversion before we assign to C(i,j).
          auto temp = alpha * conditionalConjugate (B(l,j), conjB);
          for (IndexType i = 0; i < m; ++i) {
            C(i,j) = C(i,j) + temp * conditionalConjugate (A(i,l), conjA);
          }
        }
      }
    }
    else {
      // Form C = alpha*A**T*B + beta*C
      for (IndexType j = 0; j < n; ++j) {
        for (IndexType i = 0; i < m; ++i) {
          c_value_type temp = ZERO;
          for (IndexType l = 0; l < k; ++l) {
            temp = temp + conditionalConjugate (A(l,i), conjA) *
              conditionalConjugate (B(l,j), conjB);
          }
          if (beta == ZERO) {
            C(i,j) = alpha*temp;
          }
          else {
            C(i,j) = alpha*temp + beta * C(i,j);
          }
        }
      }
    }
  }
  else {
    if (transA == 'n' || transA == 'N') {
      // Form C = alpha*A*B**T + beta*C
      for (IndexType j = 0; j < n; ++j) {
        if (beta == ZERO) {
          for (IndexType i = 0; i < m; ++i) {
            C(i,j) = ZERO;
          }
        }
        else if (beta != ONE) {
          for (IndexType i = 0; i < m; ++i) {
            C(i,j) = beta*C(i,j);
          }
        }
        for (IndexType l = 0; l < k; ++l) {
          // Don't use c_value_type here, since it unnecessarily
          // forces type conversion before we assign to C(i,j).
          auto temp = alpha * conditionalConjugate (B(j,l), conjB);
          for (IndexType i = 0; i < m; ++i) {
            C(i,j) = C(i,j) + temp * conditionalConjugate (A(i,l), conjA);
          }
        }
      }
    }
    else {
      // Form C = alpha*A**T*B**T + beta*C
      for (IndexType j = 0; j < n; ++j) {
        for (IndexType i = 0; i < m; ++i) {
          c_value_type temp = ZERO;
          for (IndexType l = 0; l < k; ++l) {
            temp = temp + conditionalConjugate (A(l,i), conjA) *
              conditionalConjugate (B(j,l), conjB);
          }
          if (beta == ZERO) {
            C(i,j) = alpha*temp;
          }
          else {
            C(i,j) = alpha*temp + beta*C(i,j);
          }
        }
      }
    }
  }
}


template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class CoefficientType,
         class IndexType = int>
void
test_compact (const char transA,
      const char transB,
      const CoefficientType& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const CoefficientType& beta,
      const ViewType3& C)
{

  //std::cout<< "gemm 1" << std::endl;
  // Assert that A, B, and C are in fact matrices
  static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
  static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
  static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

  typedef typename ViewType3::non_const_value_type c_value_type;
  typedef Kokkos::Details::ArithTraits<CoefficientType> STS;
  const CoefficientType ZERO = STS::zero ();
  const CoefficientType ONE = STS::one ();

  // Get the dimensions
  const IndexType m = C.dimension_0 ();
  const IndexType n = C.dimension_1 ();
  const IndexType k = (transA == 'N' || transA == 'n') ?
    A.dimension_1 () : A.dimension_0 ();

  //std::cout << "m = " << m << ", n = "<< n << ", k = " << k << std::endl;
  //std::cout << "A not transposed: " << (transA == 'N' || transA == 'n') << std::endl;
  //std::cout << "B not transposed: " << (transB == 'N' || transB == 'n') << std::endl;

  const bool conjA = transA == 'C' || transA == 'c';
  const bool conjB = transB == 'C' || transB == 'c';

  // quick return if possible
  if (alpha == ZERO && beta == ONE) {
    return;
  }

  // And if alpha equals zero...
  if (alpha == ZERO) {
    if (beta == ZERO) {
      for (IndexType i = 0; i < m; ++i) {
        for (IndexType j = 0; j < n; ++j) {
          C(i,j) = ZERO;
        }
      }
    }
    else {
      for (IndexType i = 0; i < m; ++i) {
        for (IndexType j = 0; j < n; ++j) {
          C(i,j) = beta*C(i,j);
        }
      }
    }
    return;
  }
 
  using KokkosBatched::Experimental::Trans;
  using KokkosBatched::Experimental::Algo;

  if(transA == 'N' || transA == 'n')
  {
    if (transB == 'N' || transB == 'n')
        KokkosBatched::Experimental::SerialGemm<Trans::NoTranspose,Trans::NoTranspose,Algo::Gemm::Blocked>
        ::invoke(alpha, A, B, beta, C);
    else
        KokkosBatched::Experimental::SerialGemm<Trans::NoTranspose,Trans::Transpose,Algo::Gemm::Blocked>
        ::invoke(alpha, A, B, beta, C);
  }
  else
  {
    if (transB == 'N' || transB == 'n')
        KokkosBatched::Experimental::SerialGemm<Trans::Transpose,Trans::NoTranspose,Algo::Gemm::Blocked>
        ::invoke(alpha, A, B, beta, C);
    else
        KokkosBatched::Experimental::SerialGemm<Trans::Transpose,Trans::Transpose,Algo::Gemm::Blocked>
        ::invoke(alpha, A, B, beta, C);
  }  
}


template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class CoefficientType,
         class IndexType = int>
void
test_team (const char transA,
      const char transB,
      const CoefficientType& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const CoefficientType& beta,
      const ViewType3& C)
{

  // Assert that A, B, and C are in fact matrices
  static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
  static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
  static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

  // Get the dimensions
  const IndexType m = C.dimension_0 ();
  const IndexType n = C.dimension_1 ();
  const IndexType k = (transA == 'N' || transA == 'n') ?
    A.dimension_1 () : A.dimension_0 ();

  using KokkosBatched::Experimental::Trans;
  using KokkosBatched::Experimental::Algo;

  const int pool_size = Kokkos::DefaultExecutionSpace::thread_pool_size();

  if(m < k)
  {
    //Inner product
    const IndexType k_per_thread = ceil(k/pool_size);
    Kokkos::parallel_for (pool_size, KOKKOS_LAMBDA (const int i) {
      IndexType k_max = ((i+1)*k_per_thread > k)? k:(i+1)*k_per_thread ;
      auto A_sub = subview (A, Kokkos::ALL(), Kokkos::make_pair (i*k_per_thread,  k_max));
      auto B_sub = subview (B, Kokkos::make_pair (i*k_per_thread,  k_max), Kokkos::ALL());
      KokkosBatched::Experimental::SerialGemm<Trans::NoTranspose,Trans::NoTranspose,Algo::Gemm::Blocked>
        ::invoke(alpha, A_sub, B_sub, beta, C);    
    });  

  }
  else
  {
    //Update
    const IndexType m_per_thread = ceil(m/pool_size);
    Kokkos::parallel_for (pool_size, KOKKOS_LAMBDA (const int i) {
      IndexType m_max = ((i+1)*m_per_thread > m)? m:(i+1)*m_per_thread ;
      auto A_sub = subview (A, Kokkos::make_pair (i*m_per_thread,  m_max), Kokkos::ALL());
      auto C_sub = subview (C, Kokkos::make_pair (i*m_per_thread,  m_max), Kokkos::ALL());
      KokkosBatched::Experimental::SerialGemm<Trans::NoTranspose,Trans::NoTranspose,Algo::Gemm::Blocked>
        ::invoke(alpha, A_sub, B, beta, C_sub);    
    });   
  }

/*
  std::cout << Kokkos::DefaultExecutionSpace::thread_pool_size() << std::endl;

  Kokkos::parallel_for (15, KOKKOS_LAMBDA (const int i) {
    // printf works in a CUDA parallel kernel; std::ostream does not.
    printf ("Hello from i = %i\n",  Kokkos::DefaultExecutionSpace::thread_pool_rank());
  });  
  */
}


template<typename scalar> void test_function(int argc, char** argv, int ensemble_size)
{
    using namespace KokkosBatched::Experimental;

    Kokkos::initialize (argc,argv);
    {

        const int m = 30000;
        const int n = 1;
        const int k = 300;

        const int N = 10;


        typedef int IndexType;


        typedef Kokkos::DefaultHostExecutionSpace HostSpaceType ;
        typedef Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::Device<Kokkos::OpenMP, Kokkos::HostSpace>, Kokkos::MemoryTraits<(unsigned int)0>> DoubleViewType;
        typedef Kokkos::View<scalar**, Kokkos::LayoutRight, Kokkos::Device<Kokkos::OpenMP, Kokkos::HostSpace>, Kokkos::MemoryTraits<(unsigned int)0>> ScalarViewType;
        ScalarViewType A("A", m,k), B("B", k,n), C("C", m,n);
        DoubleViewType Ad("Ad", ensemble_size,m,k), Bd("Bd", ensemble_size,k,n), Cd("Cd", ensemble_size,m,n);
        
        scalar lin_one = (scalar) 1.;
        for(auto s=0; s<ensemble_size; ++s)
            lin_one[s] = lin_one[s]*(s+1)/ensemble_size;

        scalar alpha = (scalar) 1. * lin_one;
        scalar beta = (scalar) 0.;
        std::cout<< alpha << std::endl;
        std::cout<< beta << std::endl;


        Kokkos::deep_copy(A, 1.5*lin_one);
        Kokkos::deep_copy(B, 2.1*lin_one);
        Kokkos::deep_copy(C, 1.*lin_one);

        const char transA = 'n';
        const char transB = 'n';

        auto begin_default = std::chrono::high_resolution_clock::now();

        for (IndexType i = 0; i < N; ++i)
          test_gemm_default<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);

        auto end_default = std::chrono::high_resolution_clock::now();
        std::ofstream goodC;
        goodC.open("good_C.txt");

        for (IndexType i = 0; i < m; ++i) {
        for (IndexType j = 0; j < n; ++j) {
            goodC << C(i,j) << " ";
        }
        goodC << std::endl;
        }
        goodC.close();

        Kokkos::deep_copy(C, 1.*lin_one);

        auto begin_compact = std::chrono::high_resolution_clock::now();

        for (IndexType i = 0; i < N; ++i)
          test_compact<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
        auto end_compact = std::chrono::high_resolution_clock::now();

        std::ofstream badC;
        badC.open("bad_C.txt");

        for (IndexType i = 0; i < m; ++i) {
        for (IndexType j = 0; j < n; ++j) {
            badC << C(i,j) << " ";
        }
        badC << std::endl;
        }
        badC.close();

        Kokkos::deep_copy(C, 1.*lin_one);

        auto begin_team = std::chrono::high_resolution_clock::now();

        for (IndexType i = 0; i < N; ++i)
          test_team<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
        auto end_team = std::chrono::high_resolution_clock::now();

        std::ofstream teamC;
        teamC.open("team_C.txt");

        for (IndexType i = 0; i < m; ++i) {
        for (IndexType j = 0; j < n; ++j) {
            teamC << C(i,j) << " ";
        }
        teamC << std::endl;
        }
        teamC.close();

        double total_time_default = std::chrono::duration_cast<std::chrono::nanoseconds>(end_default-begin_default).count();
        std::cout << ensemble_size << " " << total_time_default/(N*1e09) << " sec" << std::endl;

        double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();
        std::cout << ensemble_size << " " << total_time_compact/(N*1e09) << " sec" << std::endl;

        double total_time_team = std::chrono::duration_cast<std::chrono::nanoseconds>(end_team-begin_team).count();
        std::cout << ensemble_size << " " << total_time_team/(N*1e09) << " sec" << std::endl;        

#if 0       
    
        
        for(size_t i=0; i<A.dimension_0 ();++i)
            for(size_t j=0; j<A.dimension_1 ();++j)  
                for(size_t s=0; s<ensemble_size;++s)
                    Ad(s,i,j) = A(i,j)[s];   

        for(size_t i=0; i<B.dimension_0 ();++i)
            for(size_t j=0; j<B.dimension_1 ();++j)  
                for(size_t s=0; s<ensemble_size;++s)
                    Bd(s,i,j) = B(i,j)[s];   

        for(size_t i=0; i<C.dimension_0 ();++i)
            for(size_t j=0; j<C.dimension_1 ();++j)  
                for(size_t s=0; s<ensemble_size;++s)
                    Cd(s,i,j) = C(i,j)[s];                                                              



        bool conjA = false;
        bool conjB = false;

        using Kokkos::LayoutRight;
        using Kokkos::LayoutLeft;
        int group_count = 1;


        MKL_INT blksize_m[1] = { m };
        MKL_INT blksize_n[1] = { n };
        MKL_INT blksize_k[1] = { k };

        MKL_INT lda[1] = { k };
        MKL_INT ldb[1] = { n };
        MKL_INT ldc[1] = { n };

        CBLAS_LAYOUT layout = CblasRowMajor;

        CBLAS_TRANSPOSE ctransA[1] = {CblasNoTrans};
        CBLAS_TRANSPOSE ctransB[1] = {CblasNoTrans};
        
/*
            typedef typename ScalarViewType::array_layout array_layout;
            if (std::is_same<array_layout, LayoutLeft>::value)
            {
                std::cout << "layout = CblasColMajor;" << std::endl;
                layout = CblasColMajor;

                ldc[0]= (MKL_INT) ((1<m)?m:1);


                for (auto i=0; i<group_count; ++i)
                {
                    
                if (transA == 'N' || transA == 'n')
                    ctransA[i] = CblasNoTrans;
                else if(conjA)
                    ctransA[i] = CblasConjTrans;
                else  
                    ctransA[i] = CblasTrans;

                if (transB == 'N' || transB == 'n')
                    ctransB[i] = CblasNoTrans;
                else if(conjB)
                    ctransB[i] = CblasConjTrans;
                else  
                    ctransB[i] = CblasTrans;

                if(ctransA[i] == CblasNoTrans)
                    lda[i]=(1<m)?m:1;
                else   
                    lda[i]=(1<k)?k:1;

                if(ctransB[i] == CblasNoTrans)
                    ldb[i]=(1<k)?k:1;
                else   
                    ldb[i]=(1<n)?n:1;        
                }
            }
            else
            {
                std::cout << "layout = CblasRowMajor;" << std::endl;
                layout = CblasRowMajor;
                ldc[0]= (MKL_INT) ((1<n)?n:1);

                for (auto i=0; i<group_count; ++i)
                {
                if (transA == 'N' || transA == 'n')
                    ctransA[i] = CblasNoTrans;
                else if(conjA)
                    ctransA[i] = CblasConjTrans;
                else  
                    ctransA[i] = CblasTrans;

                if (transB == 'N' || transB == 'n')
                    ctransB[i] = CblasNoTrans;
                else if(conjB)
                    ctransB[i] = CblasConjTrans;
                else  
                    ctransB[i] = CblasTrans;

                if(ctransA[i] == CblasNoTrans)
                    lda[i]=(1<k)?k:1;
                else   
                    lda[i]=(1<m)?m:1;

                if(ctransB[i] == CblasNoTrans)
                    ldb[i]=(1<n)?n:1;
                else   
                    ldb[i]=(1<k)?k:1;        
                }
            }
 */          

        double calpha[1] = { alpha[0] };
        double cbeta[1] = { beta[0] };
        double one[1] = { 1.0 };
        MKL_INT size_per_grp[1] = { ensemble_size};

        double
            *aa[ensemble_size],
            *bb[ensemble_size],
            *cc[ensemble_size];

        for (int s=0;s<ensemble_size;++s) {
            aa[s] = &Ad(s, 0, 0);
            bb[s] = &Bd(s, 0, 0);
            cc[s] = &Cd(s, 0, 0);
        }
        
        std::cout << lda[0] << std::endl;
        std::cout << ldb[0] << std::endl;
        std::cout << ldc[0] << std::endl;
/*
        cblas_dgemm_batch(layout, 
                            ctransA,
                            ctransB,
                            blksize_m, blksize_n, blksize_k, 
                            calpha,
                            (const double**)aa, lda,
                            (const double**)bb, ldb,
                            cbeta,
                            cc, ldc,
                            1, size_per_grp);    
*/        
        for(size_t i=0; i<C.dimension_0 ();++i)
            for(size_t j=0; j<C.dimension_1 ();++j)  
                for(size_t s=0; s<ensemble_size;++s)
                    C(i,j)[s] =  Cd(s,i,j); 

        std::cout << "==========" << std::endl; 
        for(size_t i=0; i<C.dimension_0 ();++i)
        {
            for(size_t j=0; j<C.dimension_1 ();++j)       
                std::cout << C(i,j) << " ";
            std::cout << std::endl;    
        }
        std::cout << "==========" << std::endl; 

        {
            using KokkosBatched::Experimental::Trans;
            using KokkosBatched::Experimental::Algo;

            Kokkos::deep_copy(C, 1.*lin_one);
            std::cout << "==========" << std::endl; 
            for(size_t i=0; i<C.dimension_0 ();++i)
            {
                for(size_t j=0; j<C.dimension_1 ();++j)       
                    std::cout << C(i,j) << " ";
                std::cout << std::endl;    
            }
            std::cout << "==========" << std::endl;        


            KokkosBatched::Experimental::SerialGemm<Trans::NoTranspose,Trans::NoTranspose,Algo::Gemm::Blocked>
            ::invoke(alpha, A, B, beta, C);    
            
            std::cout << "==========" << std::endl; 
            for(size_t i=0; i<C.dimension_0 ();++i)
            {
                for(size_t j=0; j<C.dimension_1 ();++j)       
                    std::cout << C(i,j) << " ";
                std::cout << std::endl;    
            }
            std::cout << "==========" << std::endl;                     
        }

#endif        
    }
    Kokkos::finalize();
}

int main(int argc, char** argv)
{

  {
    constexpr int ensemble_size = 8;

    typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;

    test_function<scalar>(argc,argv,ensemble_size);
  }
  {
    constexpr int ensemble_size = 16;

    typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;

    test_function<scalar>(argc,argv,ensemble_size);
  }
  {
    constexpr int ensemble_size = 32;

    typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;

    test_function<scalar>(argc,argv,ensemble_size);
  }

    return 0;
}
