#ifndef BlisGEMV_DEF_HPP
#define BlisGEMV_DEF_HPP

//#include <boost/math/common_factor.hpp>
#include <xmmintrin.h>
//#include <Kokkos_ScatterView.hpp>

#include <EnsembleTraits.h>

namespace BlisGEMV{

template<typename S>
KOKKOS_INLINE_FUNCTION __attribute__((always_inline)) __m512d M512D_ENSEMBLE_STREAM_LOAD(const Sacado::MP::Vector<S>  &ensemble,const int &i)
{
  return _mm512_stream_load_si512(&(ensemble[i*STOKHOS_MASK_AVX_VECTOR_SIZE]));
}

template< typename T, int unrolling_factor >
KOKKOS_INLINE_FUNCTION
void inner_product_inner_kernel(T * A, T * b, T * c){
#pragma vector aligned
#pragma ivdep
#pragma unroll
  for (int j=0; j<unrolling_factor; ++j)
    c[0] += A[j]*b[j];
};

template< typename T, int unrolling_factor >
KOKKOS_INLINE_FUNCTION
void update_inner_kernel(T * A, T * b, T * c, T alpha){
#pragma vector aligned
#pragma ivdep
#pragma unroll
  for (int j=0; j<unrolling_factor; ++j)
    c[j] += alpha*A[j]*b[0];
};

#define Sacado_MP_Vector_GEMV_Unrolling_Factor_s_1  20
#define Sacado_MP_Vector_GEMV_Unrolling_Factor_s_8  1
#define Sacado_MP_Vector_GEMV_Unrolling_Factor_s_16 1
#define Sacado_MP_Vector_GEMV_Unrolling_Factor_s_24 1
#define Sacado_MP_Vector_GEMV_Unrolling_Factor_s_32 1



#define Sacado_MP_Vector_GEMV_Has_Intel_Intrinsics

#define Sacado_MP_Vector_GEMV_Unrolling_Factor(size)                  \
{                                                                     \
  (size==32*8 ? Sacado_MP_Vector_GEMV_Unrolling_Factor_s_32 :         \
    (size==24*8 ? Sacado_MP_Vector_GEMV_Unrolling_Factor_s_24 :       \
      (size==16*8 ? Sacado_MP_Vector_GEMV_Unrolling_Factor_s_16 :     \
        (size==8*8 ? Sacado_MP_Vector_GEMV_Unrolling_Factor_s_8 :     \
          Sacado_MP_Vector_GEMV_Unrolling_Factor_s_1 ))))             \
}

#define Sacado_MP_Vector_GEMV_m_n_per_tile_s_1  128
#define Sacado_MP_Vector_GEMV_m_n_per_tile_s_8  128
#define Sacado_MP_Vector_GEMV_m_n_per_tile_s_16 128
#define Sacado_MP_Vector_GEMV_m_n_per_tile_s_24 128
#define Sacado_MP_Vector_GEMV_m_n_per_tile_s_32 128

#define Sacado_MP_Vector_GEMV_m_n_per_tile(size)                  \
{                                                                 \
  (size==32*8 ? Sacado_MP_Vector_GEMV_m_n_per_tile_s_32 :         \
    (size==24*8 ? Sacado_MP_Vector_GEMV_m_n_per_tile_s_24 :       \
      (size==16*8 ? Sacado_MP_Vector_GEMV_m_n_per_tile_s_16 :     \
        (size==8*8 ? Sacado_MP_Vector_GEMV_m_n_per_tile_s_8 :     \
          Sacado_MP_Vector_GEMV_m_n_per_tile_s_1 ))))             \
}

#define Sacado_MP_Vector_GEMV_Unrolling_Factor_after_peeling_s_1  Sacado_MP_Vector_GEMV_m_n_per_tile_s_1%Sacado_MP_Vector_GEMV_Unrolling_Factor_s_1 == 0? Sacado_MP_Vector_GEMV_Unrolling_Factor_s_1 : Sacado_MP_Vector_GEMV_m_n_per_tile_s_1%Sacado_MP_Vector_GEMV_Unrolling_Factor_s_1
#define Sacado_MP_Vector_GEMV_Unrolling_Factor_after_peeling_s_8  Sacado_MP_Vector_GEMV_m_n_per_tile_s_8%Sacado_MP_Vector_GEMV_Unrolling_Factor_s_8 == 0? Sacado_MP_Vector_GEMV_Unrolling_Factor_s_8 : Sacado_MP_Vector_GEMV_m_n_per_tile_s_8%Sacado_MP_Vector_GEMV_Unrolling_Factor_s_8
#define Sacado_MP_Vector_GEMV_Unrolling_Factor_after_peeling_s_16 Sacado_MP_Vector_GEMV_m_n_per_tile_s_16%Sacado_MP_Vector_GEMV_Unrolling_Factor_s_16 == 0? Sacado_MP_Vector_GEMV_Unrolling_Factor_s_16 : Sacado_MP_Vector_GEMV_m_n_per_tile_s_16%Sacado_MP_Vector_GEMV_Unrolling_Factor_s_16
#define Sacado_MP_Vector_GEMV_Unrolling_Factor_after_peeling_s_24 Sacado_MP_Vector_GEMV_m_n_per_tile_s_24%Sacado_MP_Vector_GEMV_Unrolling_Factor_s_24 == 0? Sacado_MP_Vector_GEMV_Unrolling_Factor_s_24 : Sacado_MP_Vector_GEMV_m_n_per_tile_s_24%Sacado_MP_Vector_GEMV_Unrolling_Factor_s_24
#define Sacado_MP_Vector_GEMV_Unrolling_Factor_after_peeling_s_32 Sacado_MP_Vector_GEMV_m_n_per_tile_s_32%Sacado_MP_Vector_GEMV_Unrolling_Factor_s_32 == 0? Sacado_MP_Vector_GEMV_Unrolling_Factor_s_32 : Sacado_MP_Vector_GEMV_m_n_per_tile_s_32%Sacado_MP_Vector_GEMV_Unrolling_Factor_s_32


#define Sacado_MP_Vector_GEMV_Unrolling_Factor_After_Peeling(size)                  \
{                                                                                   \
  (size==32*8 ? Sacado_MP_Vector_GEMV_Unrolling_Factor_after_peeling_s_32 :         \
    (size==24*8 ? Sacado_MP_Vector_GEMV_Unrolling_Factor_after_peeling_s_24 :       \
      (size==16*8 ? Sacado_MP_Vector_GEMV_Unrolling_Factor_after_peeling_s_16 :     \
        (size==8*8 ? Sacado_MP_Vector_GEMV_Unrolling_Factor_after_peeling_s_8 :     \
          Sacado_MP_Vector_GEMV_Unrolling_Factor_after_peeling_s_1 ))))             \
}

#define size 8
#define n_vectors 1
#include "inner_kernels.hpp"
#undef size
#undef n_vectors

#define size 16
#define n_vectors 2
#include "inner_kernels.hpp"
#undef size
#undef n_vectors

#define size 24
#define n_vectors 3
#include "inner_kernels.hpp"
#undef size
#undef n_vectors

#define size 32
#define n_vectors 4
#include "inner_kernels.hpp"
#undef size
#undef n_vectors

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void DOTXV(
      const char transA,
      Scalar & rho,
      const Scalar& alpha,
      const Scalar& beta,
      const ViewType1& A,
      const ViewType2& B,
      const IndexType i,
      const IndexType n)
{

/*
    Implement the Generalized dot product.
*/

    rho *= beta;

    IndexType j;

    if (transA == 'N' || transA == 'n')
        for ( j=0; j<n; ++j )
        {   
            rho += alpha*A(i,j)*B(j,0); 
        }
    else    
        for ( j=0; j<n; ++j )
        {   
            rho += alpha*A(j,i)*B(j,0); 
        }    
}    

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void DOTXV_j(
      const char transA,
      Scalar & rho,
      const Scalar& alpha,
      const Scalar& beta,
      const ViewType1& A,
      const ViewType2& B,
      const IndexType i,
      const IndexType j_0,
      const IndexType n)
{

/*
    Implement the Generalized dot product.
*/

    rho *= beta;

    IndexType j;

    if (transA == 'N' || transA == 'n')
        #pragma unroll
        for ( j=0; j<n; ++j )
        {
            rho += alpha*A(i,j_0+j)*B(j_0+j,0);
        }
    else
        #pragma unroll
        for ( j=0; j<n; ++j )
        {
            rho += alpha*A(j_0+j,i)*B(j_0+j,0);
        }
}

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void DOTXV_j2(
      const char transA,
      Scalar *c,
      const IndexType i_row,
      const IndexType i_col,
      const Scalar& alpha,
      const Scalar& beta,
      const Scalar *a,
      const IndexType lda,
      const ViewType2& B,
      const IndexType i,
      const IndexType j_0,
      const IndexType n_row,
      const IndexType n,
      const IndexType n_jump)
{
    Scalar tmp = 0.;
    for (IndexType j=0; j<n_row; ++j )
    {
        const int i_C = floor(((double) j_0+j)/n);
        const int j_b = j_0+j-n*i_C;
        c[i_C] += alpha*a[j*lda]*B(j_b,0);
    }
}

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void AXPYV(
      const char transA,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const ViewType3& C,
      const IndexType j,
      const IndexType m)
{

/*
    Implement the Accumulate scaled vector.
*/    

    IndexType i;
    Scalar c;

    if (transA == 'N' || transA == 'n')
        for ( i=0; i<m; ++i )
        {   
            c += alpha*B(j,0)*A(i,j);
        }
    else    
        for ( i=0; i<m; ++i )
        {   
            c += alpha*B(j,0)*A(j,i);
        }   

} 


template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void AXPYV_atom(
      const char transA,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const ViewType3& C,
      const IndexType j,
      const IndexType m)
{

/*
    Implement the Accumulate scaled vector.
*/    

    IndexType i;

    Scalar tmp;

    if (transA == 'N' || transA == 'n')
        for ( i=0; i<m; ++i )
        {   
            tmp = alpha*B(j,0)*A(i,j);
            Kokkos::atomic_add(&C(i,0),tmp); 
        }
    else    
        for ( i=0; i<m; ++i )
        {   
            tmp = alpha*B(j,0)*A(j,i);
            Kokkos::atomic_add(&C(i,0),tmp); 
        }   

} 

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void DOTXF(
      const char transA,
      Scalar * rho,
      const Scalar& alpha,
      const Scalar& beta,
      const ViewType1& A,
      const ViewType2& B,
      const IndexType i,
      const IndexType n)
{

/*
    Implement the Generalized dot product.
*/

    rho[0] *= beta;
    rho[1] *= beta;
    rho[2] *= beta;
    rho[3] *= beta;

    IndexType j;

    if (transA == 'N' || transA == 'n')
        for ( j=0; j<n; ++j )
        {   
            rho[0] += alpha*A(i,j)*B(j,0); 
            rho[1] += alpha*A(i+1,j)*B(j,0); 
            rho[2] += alpha*A(i+2,j)*B(j,0); 
            rho[3] += alpha*A(i+3,j)*B(j,0); 
        }
    else    
        for ( j=0; j<n; ++j )
        {   
            rho[0] += alpha*A(j,i)*B(j,0); 
            rho[1] += alpha*A(j,i+1)*B(j,0); 
            rho[2] += alpha*A(j,i+2)*B(j,0); 
            rho[3] += alpha*A(j,i+3)*B(j,0); 
        }    
}    

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void AXPYF(
      const char transA,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const ViewType3& C,
      const IndexType j,
      const IndexType m)
{

/*
    Implement the Accumulate scaled vector.
*/    

    IndexType i;
    Scalar c;

    if (transA == 'N' || transA == 'n')
        for ( i=0; i<m; ++i )
        {   
            c += alpha*(B(j,0)*A(i,j)+B(j+1,0)*A(i,j+1)+B(j+2,0)*A(i,j+2)+B(j+3,0)*A(i,j+3));
        }
    else    
        for ( i=0; i<m; ++i )
        {   
            c += alpha*(B(j,0)*A(j,i)+B(j+1,0)*A(j+1,i)+B(j+2,0)*A(j+2,i)+B(j+3,0)*A(j+3,i));
        }   

}

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void AXPYF_atom(
      const char transA,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const ViewType3& C,
      const IndexType j,
      const IndexType m)
{

/*
    Implement the Accumulate scaled vector.
*/    

    IndexType i;

    Scalar tmp;

    if (transA == 'N' || transA == 'n')
        for ( i=0; i<m; ++i )
        {   
            tmp = alpha*(B(j,0)*A(i,j)+B(j+1,0)*A(i,j+1)+B(j+2,0)*A(i,j+2)+B(j+3,0)*A(i,j+3));
            Kokkos::atomic_add(&C(i,0),tmp); 
        }
    else    
        for ( i=0; i<m; ++i )
        {   
            tmp = alpha*(B(j,0)*A(j,i)+B(j+1,0)*A(j+1,i)+B(j+2,0)*A(j+2,i)+B(j+3,0)*A(j+3,i));
            Kokkos::atomic_add(&C(i,0),tmp); 
        }   

} 

template< typename T >
KOKKOS_INLINE_FUNCTION
void prefetch_scalar(volatile T * const dest, int i){
_mm_prefetch((const char*) dest,i);
};


template <typename S>
KOKKOS_INLINE_FUNCTION
void prefetch_scalar(volatile Sacado::MP::Vector<S> * const dest, int i)
{
#pragma unroll
for (int j=0; j<S::static_size; j+=8)
  _mm_prefetch((const char*) (dest)+j*8,i); // sizeof(char) = 1, sizeof(double) = 8
};

  
template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void GEMV_row_based(
      const char transA,
      const char transB,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const Scalar& beta,
      const ViewType3& C)
{
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n_rhs = C.dimension_1 ();
    const IndexType n = (transA == 'N' || transA == 'n') ?
        A.dimension_1 () : A.dimension_0 ();

    IndexType i;

    for ( i=0; i<m; ++i )
    {   
        DOTXV<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,C(i,0),alpha,beta,A,B,i,n);
    }
}

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void GEMV_col_based(
      const char transA,
      const char transB,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const Scalar& beta,
      const ViewType3& C)
{
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n_rhs = C.dimension_1 ();
    const IndexType n = (transA == 'N' || transA == 'n') ?
        A.dimension_1 () : A.dimension_0 ();

    IndexType i, j;

    for ( i=0; i<m; ++i )
    {   
        C(i,0) *= beta; 
    }

    for ( j=0; j<n; ++j )
    {   
        AXPYV<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,alpha,A,B,C,j,m);
    }
}

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void GEMV_fused_row_based(
      const char transA,
      const char transB,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const Scalar& beta,
      const ViewType3& C)
{
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n_rhs = C.dimension_1 ();
    const IndexType n = (transA == 'N' || transA == 'n') ?
        A.dimension_1 () : A.dimension_0 ();

    IndexType i;

    for ( i=0; i<m; i+=4 )
    {   
        if(i+4<m)
            DOTXF<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,&C(i,0),alpha,beta,A,B,i,n);
        else
            for ( ; i<m; ++i )
                DOTXV<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,C(i,0),alpha,beta,A,B,i,n);
    }
}

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void GEMV_fused_col_based(
      const char transA,
      const char transB,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const Scalar& beta,
      const ViewType3& C)
{
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n_rhs = C.dimension_1 ();
    const IndexType n = (transA == 'N' || transA == 'n') ?
        A.dimension_1 () : A.dimension_0 ();

    IndexType i, j;

    for ( i=0; i<m; ++i )
    {   
        C(i,0) *= beta; 
    }

    for ( j=0; j<n; j+=4 )
    {   
        if(j+4<n)
            AXPYF<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,alpha,A,B,C,j,m);
        else
            for ( ; j<n; ++j )
                AXPYV<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,alpha,A,B,C,j,m);
    }
}


template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void GEMV_threaded_row_based(
      const char transA,
      const char transB,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const Scalar& beta,
      const ViewType3& C)
{
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n = B.dimension_0 ();
  
    const int pool_size = Kokkos::DefaultExecutionSpace::thread_pool_size();
    const double mul_pool_size = 2;
    if(((double) m)>pool_size*mul_pool_size)
    {
      Kokkos::parallel_for (m, KOKKOS_LAMBDA (const int i)
      {
          DOTXV<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,C(i,0),alpha,beta,A,B,i,n);
      });
    }
    else
    {
      const IndexType n_per_tile = Sacado_MP_Vector_GEMV_m_n_per_tile(sizeof(Scalar));//(sizeof(Scalar) == 32*8) ? 80 : ( (sizeof(Scalar) == 24*8) ? 98 : ( (sizeof(Scalar) == 16*8) ? 132 : 136 ) ) ;

      const int pool_size = Kokkos::DefaultExecutionSpace::thread_pool_size();

      const int team_size = 4;
      #if 1
        const IndexType n_per_tile2 = n_per_tile*team_size;
      #else
        const double work_per_thread = ((double) n)/(pool_size*n_per_tile);
        const IndexType n_per_tile2 = ceil((double) n)/(pool_size*ceil(work_per_thread))*team_size;
      #endif

      const int unrolling_factor = Sacado_MP_Vector_GEMV_Unrolling_Factor(sizeof(Scalar)); //(sizeof(Scalar) == 8*8) ? Sacado_MP_Vector_GEMV_Unrolling_Factor_s_8 : ((sizeof(Scalar) == 16*8) ? Sacado_MP_Vector_GEMV_Unrolling_Factor_s_16 : Sacado_MP_Vector_GEMV_Unrolling_Factor_s_32);

      const int n_i2 = ceil(((double) n)/n_per_tile2);
      using Kokkos::TeamThreadRange;

      Kokkos::TeamPolicy<> policy = Kokkos::TeamPolicy<>( n_i2, team_size);
      typedef Kokkos::TeamPolicy<>::member_type member_type;
    
      Kokkos::parallel_for (policy, KOKKOS_LAMBDA (member_type team_member)
      {
        const IndexType j = team_member.league_rank();
        const IndexType j_min = n_per_tile2*j;
        const IndexType j_max = (j_min+n_per_tile2 > n) ? n : j_min+n_per_tile2;
        #if 1
          const IndexType i_min = (j<m) ? j:(j%m);
        #else
          const int i_team = (Kokkos::DefaultExecutionSpace::thread_pool_rank()-team_member.team_rank ())/team_size;
          const IndexType i_min = (i_team<m) ? i_team:(i_team%m);
        #endif
        
        const int nj = floor((j_max-j_min)/unrolling_factor);
        
        if (transA == 'N' || transA == 'n'){
          for ( IndexType i=i_min; i<m; ++i ){
            Scalar tmp = 0.;
            Kokkos::parallel_reduce (TeamThreadRange (team_member, nj),[=] (IndexType jj, Scalar& tmp_sum){
              inner_product_inner_kernel<Scalar,unrolling_factor> (&A(i,jj*unrolling_factor+j_min),&B(jj*unrolling_factor+j_min,0),&tmp_sum);
            },tmp);
            if ( team_member.team_rank () == 0) {
              tmp *= alpha;
              Kokkos::atomic_add(&C(i,0),tmp);
            }
          }
        }
        else{
          for ( IndexType i=i_min; i<m; ++i ){
            Scalar tmp = 0.;
            Kokkos::parallel_reduce (TeamThreadRange (team_member, nj),[=] (IndexType jj, Scalar& tmp_sum){
              inner_product_inner_kernel<Scalar,unrolling_factor> (&A(jj*unrolling_factor+j_min,i),&B(jj*unrolling_factor+j_min,0),&tmp_sum);
            },tmp);
            if ( team_member.team_rank () == 0) {
              tmp *= alpha;
              Kokkos::atomic_add(&C(i,0),tmp);
            }
          }
        }
        if (transA == 'N' || transA == 'n'){
          for ( IndexType i=0; i<i_min; ++i ){
            Scalar tmp = 0.;
            Kokkos::parallel_reduce (TeamThreadRange (team_member, nj),[=] (IndexType jj, Scalar& tmp_sum){
              inner_product_inner_kernel<Scalar,unrolling_factor> (&A(i,jj*unrolling_factor+j_min),&B(jj*unrolling_factor+j_min,0),&tmp_sum);
            },tmp);
            if ( team_member.team_rank () == 0) {
              tmp *= alpha;
              Kokkos::atomic_add(&C(i,0),tmp);
            }
          }
        }
        else{
          for ( IndexType i=0; i<i_min; ++i ){
            Scalar tmp = 0.;
            Kokkos::parallel_reduce (TeamThreadRange (team_member, nj),[=] (IndexType jj, Scalar& tmp_sum){
              inner_product_inner_kernel<Scalar,unrolling_factor> (&A(jj*unrolling_factor+j_min,i),&B(jj*unrolling_factor+j_min,0),&tmp_sum);
            },tmp);
            if ( team_member.team_rank () == 0) {
              tmp *= alpha;
              Kokkos::atomic_add(&C(i,0),tmp);
            }
          }
        }
      });
    }
}

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void GEMV_threaded_col_based(
      const char transA,
      const char transB,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const Scalar& beta,
      const ViewType3& C)
{
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n = B.dimension_0 ();

    IndexType i;

    for ( i=0; i<m; ++i )
    {   
        C(i,0) *= beta; 
    }

    Kokkos::parallel_for (n, KOKKOS_LAMBDA (const int j) 
    {   
        AXPYV<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,alpha,A,B,C,j,m);
    });
}


template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void GEMV_threaded_fused_row_based_buffer(
      const char transA,
      const char transB,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const Scalar& beta,
      const ViewType3& C)
{
    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n = B.dimension_0 ();
  
    const IndexType n_per_tile = (sizeof(Scalar) == 32*8) ? 50 : ( (sizeof(Scalar) == 16*8) ? 100 : 100 ) ;
  
    const int n_i = ceil(((double) n)/n_per_tile);

    Kokkos::parallel_for (m, KOKKOS_LAMBDA (const int i)
    {
        C(i,0) *= beta;
    });
#if 0
    Kokkos::parallel_for (n_i, KOKKOS_LAMBDA (const int j)
    {
      const IndexType j_min = n_per_tile*j;
      const IndexType j_max = (n_per_tile*(j+1) > n) ? n : n_per_tile*(j+1);
      
      if (transA == 'N' || transA == 'n'){
        for ( IndexType i=0; i<m; ++i ){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(i,jj);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i,0),tmp);
        }
      }
      else{
        for ( IndexType i=0; i<m; ++i ){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(jj,i);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i,0),tmp);
        }
      }
    });
#else
const int pool_size = Kokkos::DefaultExecutionSpace::thread_pool_size();


if (sizeof(Scalar)!=8){ //0.5

  
  
    const int team_size = 4;
    const IndexType n_per_tile2 = n_per_tile*team_size;
    const int n_i2 = ceil(((double) n)/n_per_tile2);
    using Kokkos::TeamThreadRange;

/*
    Kokkos::TeamPolicy<> policy = Kokkos::TeamPolicy<>( m, team_size);
    typedef Kokkos::TeamPolicy<>::member_type member_type;
    Kokkos::parallel_for (policy, KOKKOS_LAMBDA (member_type team_member) {
      int i = team_member.league_rank();
      Scalar tmp = 0.;
      Kokkos::parallel_reduce (TeamThreadRange (team_member, n),[=] (IndexType j, Scalar& tmp_sum){
        if (transA == 'N' || transA == 'n')
          tmp_sum += B(j,0)*A(i,j);
        else
          tmp_sum += B(j,0)*A(j,i);
      },tmp);
      
      if ( team_member.team_rank () == 0) {
        C(i,0) += alpha*tmp;
      }
    });
*/

    Kokkos::TeamPolicy<> policy = Kokkos::TeamPolicy<>( n_i2, team_size);
    typedef Kokkos::TeamPolicy<>::member_type member_type;
  
    Kokkos::parallel_for (policy, KOKKOS_LAMBDA (member_type team_member)
    {
      const IndexType j = team_member.league_rank();
      const IndexType j_min = n_per_tile2*j;
      const IndexType j_max = (j_min+n_per_tile2 > n) ? n : j_min+n_per_tile2;

      //const int i_thread = Kokkos::DefaultExecutionSpace::thread_pool_rank();
      const IndexType i_min = (j<m) ? j:(j%m);
      
      if (transA == 'N' || transA == 'n'){
        for ( IndexType i=i_min; i<m; ++i ){
          Scalar tmp = 0.;
          Kokkos::parallel_reduce (TeamThreadRange (team_member, j_max-j_min),[=] (IndexType jj, Scalar& tmp_sum){
            tmp_sum += B(jj+j_min,0)*A(i,jj+j_min);
          },tmp);
          tmp *= alpha;
          if ( team_member.team_rank () == 0) {
            Kokkos::atomic_add(&C(i,0),tmp);
          }
        }
      }
      else{
        for ( IndexType i=i_min; i<m; ++i ){
          Scalar tmp = 0.;
          Kokkos::parallel_reduce (TeamThreadRange (team_member, j_max-j_min),[=] (IndexType jj, Scalar& tmp_sum){
            tmp_sum += B(jj+j_min,0)*A(jj+j_min,i);
          },tmp);
          tmp *= alpha;
          if ( team_member.team_rank () == 0) {
            Kokkos::atomic_add(&C(i,0),tmp);
          }
        }
      }
      if (transA == 'N' || transA == 'n'){
        for ( IndexType i=0; i<i_min; ++i ){
          Scalar tmp = 0.;
          Kokkos::parallel_reduce (TeamThreadRange (team_member, j_max-j_min),[=] (IndexType jj, Scalar& tmp_sum){
            tmp_sum += B(jj+j_min,0)*A(i,jj+j_min);
          },tmp);
          tmp *= alpha;
          if ( team_member.team_rank () == 0) {
            Kokkos::atomic_add(&C(i,0),tmp);
          }
        }
      }
      else{
        for ( IndexType i=0; i<i_min; ++i ){
          Scalar tmp = 0.;
          Kokkos::parallel_reduce (TeamThreadRange (team_member, j_max-j_min),[=] (IndexType jj, Scalar& tmp_sum){
            tmp_sum += B(jj+j_min,0)*A(jj+j_min,i);
          },tmp);
          tmp *= alpha;
          if ( team_member.team_rank () == 0) {
            Kokkos::atomic_add(&C(i,0),tmp);
          }
        }
      }
    });
/*
    Kokkos::TeamPolicy<> policy = Kokkos::TeamPolicy<>( m, team_size);
    typedef Kokkos::TeamPolicy<>::member_type member_type;
    Kokkos::parallel_for (policy, KOKKOS_LAMBDA (member_type team_member) {
      int i = team_member.league_rank();
      Scalar tmp = 0.;
      Kokkos::parallel_reduce (TeamThreadRange (team_member, n),[=] (IndexType j, Scalar& tmp_sum){
        if (transA == 'N' || transA == 'n')
          tmp_sum += B(j,0)*A(i,j);
        else
          tmp_sum += B(j,0)*A(j,i);
      },tmp);
      
      if ( team_member.team_rank () == 0) {
        C(i,0) += alpha*tmp;
      }
    });
*/

    /*
    //Scalar result;
    Kokkos::parallel_reduce(pool_size, KOKKOS_LAMBDA (const int& j, Scalar& tmp_sum)
    {
      const IndexType j_min = n_per_tile2*j;
      const IndexType j_max = (j_min+n_per_tile2 > n) ? n : j_min+n_per_tile2;
    
      Scalar tmp = 0.;
      #pragma unroll
      if (transA == 'N' || transA == 'n'){
        for ( IndexType jj=j_min; jj<j_max; ++jj ){
          tmp += B(jj,0)*A(i,jj);
        }
      }
      else{
        for ( IndexType jj=j_min; jj<j_max; ++jj ){
          tmp += B(jj,0)*A(jj,i);
        }
      }
      tmp_sum += alpha*tmp;
    },C(i,0));
    */
}
/*
if (m*1.<=pool_size*0.){ //0.5

    const IndexType n_per_tile2 = ceil(((double) n)/pool_size);
  
    for(IndexType i=0;i<m; ++i)
    {
        Scalar result;
        Kokkos::parallel_reduce(pool_size, KOKKOS_LAMBDA (const int& j, Scalar& tmp_sum)
        {
          const IndexType j_min = n_per_tile2*j;
          const IndexType j_max = (j_min+n_per_tile2 > n) ? n : j_min+n_per_tile2;
        
          Scalar tmp = 0.;
          #pragma unroll
          if (transA == 'N' || transA == 'n'){
            for ( IndexType jj=j_min; jj<j_max; ++jj ){
              tmp += B(jj,0)*A(i,jj);
            }
          }
          else{
            for ( IndexType jj=j_min; jj<j_max; ++jj ){
              tmp += B(jj,0)*A(jj,i);
            }
          }
          tmp_sum += tmp;
        },result);
      
        C(i,0) += alpha*result;
    }
}
*/
else{
    const int m_i = ceil(((double) pool_size)/m);
  
    Kokkos::parallel_for (n_i, KOKKOS_LAMBDA (const int j)
    {
      const IndexType j_min = n_per_tile*j;
      const IndexType j_max = (n_per_tile*(j+1) > n) ? n : n_per_tile*(j+1);

      const int i_thread = Kokkos::DefaultExecutionSpace::thread_pool_rank();
      const IndexType i_min = (i_thread<m) ? i_thread:(i_thread%m);
      
      const IndexType j_min2 = j_min;// (m_i>1) ? j_min + floor(((double) i_thread)/m)*floor(((double) (j_max-1-j_min))/(m_i)) : j_min; // OK!

      if (transA == 'N' || transA == 'n'){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min2; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(i_min,jj);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i_min,0),tmp);
      }
      else{
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min2; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(jj,i_min);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i_min,0),tmp);
      }
      
      if (transA == 'N' || transA == 'n'){
        for ( IndexType i=i_min+1; i<m; ++i ){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(i,jj);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i,0),tmp);
        }
      }
      else{
        for ( IndexType i=i_min+1; i<m; ++i ){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(jj,i);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i,0),tmp);
        }
      }
      if (transA == 'N' || transA == 'n'){
        for ( IndexType i=0; i<i_min; ++i ){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(i,jj);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i,0),tmp);
        }
      }
      else{
        for ( IndexType i=0; i<i_min; ++i ){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(jj,i);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i,0),tmp);
        }
      }
      
      if (transA == 'N' || transA == 'n'){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_min2; ++jj ){
            tmp += B(jj,0)*A(i_min,jj);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i_min,0),tmp);
      }
      else{
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_min2; ++jj ){
            tmp += B(jj,0)*A(jj,i_min);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i_min,0),tmp);
      }
      
    });
}
#endif
/*
      const int i_thread = Kokkos::DefaultExecutionSpace::thread_pool_rank();
 
      const IndexType i_min = (i_thread<m) ? i_thread:0;
 
      if (transA == 'N' || transA == 'n'){
        for ( IndexType i=i_min; i<m; ++i ){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(i,jj);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i,0),tmp);
        }
        for ( IndexType i=0; i<i_min; ++i ){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(i,jj);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i,0),tmp);
        }
      }
      else{
        for ( IndexType i=i_min; i<m; ++i ){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(jj,i);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i,0),tmp);
        }
        for ( IndexType i=0; i<i_min; ++i ){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(jj,i);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i,0),tmp);
        }
      }
    });
    */
/*
      if (transA == 'N' || transA == 'n')
        for ( IndexType i=i_min; i<m; ++i ){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(i,jj);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i,0),tmp);
        }
      else
        for ( IndexType i=i_thread; i<m; ++i ){
          Scalar tmp = 0.;
          #pragma unroll
          for ( IndexType jj=j_min; jj<j_max; ++jj ){
            tmp += B(jj,0)*A(jj,i);
          }
          tmp *= alpha;
          Kokkos::atomic_add(&C(i,0),tmp);
        }
      if (thread_id_smaller){
        if (transA == 'N' || transA == 'n')
          for ( IndexType i=0; i<i_min; ++i ){
            Scalar tmp = 0.;
            #pragma unroll
            for ( IndexType jj=j_min; jj<j_max; ++jj ){
              tmp += B(jj,0)*A(i,jj);
            }
            tmp *= alpha;
            Kokkos::atomic_add(&C(i,0),tmp);
          }
        else
          for ( IndexType i=0; i<i_min; ++i ){
            Scalar tmp = 0.;
            #pragma unroll
            for ( IndexType jj=j_min; jj<j_max; ++jj ){
              tmp += B(jj,0)*A(jj,i);
            }
            tmp *= alpha;
            Kokkos::atomic_add(&C(i,0),tmp);
          }
      }
*/

  
}

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
std::vector<double> GEMV_threaded_fused_row_based(
      const char transA,
      const char transB,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const Scalar& beta,
      const ViewType3& C)
{
    std::vector<double> time_vector;

    auto begin_total = std::chrono::high_resolution_clock::now();
/*
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n = B.dimension_0 ();

    const int m_i = ceil(((double) m)/4);

    Kokkos::parallel_for (m_i, KOKKOS_LAMBDA (const int i) 
    {   
        if(4*i+4<m)
            DOTXF<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,&C(4*i,0),alpha,beta,A,B,4*i,n);
        else
            for (IndexType ii = 4*i; ii<m; ++ii )
                DOTXV<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,C(ii,0),alpha,beta,A,B,ii,n);
    });
    */
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n = B.dimension_0 ();
  
    const int pool_size = Kokkos::DefaultExecutionSpace::thread_pool_size();
    const double mul_pool_size = 2;
    if(((double) m)>pool_size*mul_pool_size)
    {
      Kokkos::parallel_for (m, KOKKOS_LAMBDA (const int i)
      {
          DOTXV<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,C(i,0),alpha,beta,A,B,i,n);
      });
    }
    else if (0)
    {
        auto begin_setup = std::chrono::high_resolution_clock::now();
      
        //const IndexType n_per_tile = (sizeof(Scalar) == 32*8) ? 2 : ( (sizeof(Scalar) == 16*8) ? 6 : 14 ) ;
        const IndexType n_per_tile = (sizeof(Scalar) == 32*8) ? 50 : ( (sizeof(Scalar) == 16*8) ? 100 : 100 ) ;
      
        const int n_i = ceil(((double) n)/n_per_tile);

        #if 0
        typedef Kokkos::View<Scalar**, Kokkos::LayoutLeft, Kokkos::Device<Kokkos::OpenMP, Kokkos::HostSpace>, Kokkos::MemoryTraits<(unsigned int)0>> ScalarLeftViewType;
        #else
        //typedef Kokkos::View<Scalar*, Kokkos::LayoutRight, Kokkos::Device<Kokkos::OpenMP, Kokkos::HostSpace>, Kokkos::MemoryTraits<(unsigned int)0>> ScalarRightViewType;
        //typedef Kokkos::View<Scalar**> ScalarRightViewType;
        #endif
        auto end_setup = std::chrono::high_resolution_clock::now();
      
        auto begin_view = std::chrono::high_resolution_clock::now();
        #if 0
        ScalarLeftViewType C_thread = ScalarLeftViewType("C_thread",m,pool_size);
        #else
        //ScalarRightViewType C_thread = ScalarRightViewType(Kokkos::ViewAllocateWithoutInitializing ("C_thread"),pool_size,m);
        Kokkos::View<Scalar*> C_thread = Kokkos::View<Scalar*>(Kokkos::ViewAllocateWithoutInitializing ("C_thread"),pool_size*m);
      
        Scalar zero = 0.;
      
        Kokkos::parallel_for (pool_size*m, KOKKOS_LAMBDA (const int i)
        {
            C_thread(i) = zero;
        });
      
        #endif
      
        auto end_view = std::chrono::high_resolution_clock::now();
      
        auto begin_tile_parallel_for = std::chrono::high_resolution_clock::now();
        Kokkos::parallel_for (n_i, KOKKOS_LAMBDA (const int j)
        {
          IndexType j_min = n_per_tile*j;
          IndexType j_max = (n_per_tile*(j+1) > n) ? n : n_per_tile*(j+1);
          
          int i_thread = Kokkos::DefaultExecutionSpace::thread_pool_rank();
        
          if (transA == 'N' || transA == 'n')
            for ( IndexType i=0; i<m; ++i ){
              #pragma unroll
              for ( IndexType jj=j_min; jj<j_max; ++jj ){
              #if 0
                C_thread(i,i_thread) += alpha*B(jj,0)*A(i,jj);
              #else
                C_thread(i_thread*m+i) += alpha*B(jj,0)*A(i,jj);
              #endif
              }
            }
          else
            for ( IndexType i=0; i<m; ++i ){
              #pragma unroll
              for ( IndexType jj=j_min; jj<j_max; ++jj ){
              #if 0
                C_thread(i,i_thread) += alpha*B(jj,0)*A(jj,i);
              #else
                C_thread(i_thread*m+i) += alpha*B(jj,0)*A(jj,i);
              #endif
              }
            }
        });
        auto end_tile_parallel_for = std::chrono::high_resolution_clock::now();
      
        auto begin_reduction_parallel_for = std::chrono::high_resolution_clock::now();
        Kokkos::parallel_for (m, KOKKOS_LAMBDA (const int i)
        {
            C(i,0) *= beta;
            #pragma unroll
            for (int j =0; j<pool_size; ++j)
            #if 0
              C(i,0) += C_thread(i,j);
            #else
              C(i,0) += C_thread(j*m+i);
            #endif
        });
        auto end_reduction_parallel_for = std::chrono::high_resolution_clock::now();

        time_vector.push_back(std::chrono::duration_cast<std::chrono::nanoseconds>(end_setup-begin_setup).count());
      
        time_vector.push_back(std::chrono::duration_cast<std::chrono::nanoseconds>(end_view-begin_view).count());
      
        time_vector.push_back(std::chrono::duration_cast<std::chrono::nanoseconds>(end_tile_parallel_for-begin_tile_parallel_for).count());
      
        time_vector.push_back(std::chrono::duration_cast<std::chrono::nanoseconds>(end_reduction_parallel_for-begin_reduction_parallel_for).count());
    }
    else{
    GEMV_threaded_fused_row_based_buffer<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,transB,alpha,A,B,beta,C);
    }
    auto end_total = std::chrono::high_resolution_clock::now();
  
    time_vector.push_back(std::chrono::duration_cast<std::chrono::nanoseconds>(end_total-begin_total).count());
  
    return time_vector;
}      


template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void GEMV_threaded_fused_col_based(
      const char transA,
      const char transB,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const Scalar& beta,
      const ViewType3& C)
{
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n = B.dimension_0 ();

    IndexType i;

    for ( i=0; i<m; ++i )
    {   
        C(i,0) *= beta; 
    }

    const int n_j = ceil(((double) n)/4);

    Kokkos::parallel_for (n_j, KOKKOS_LAMBDA (const int j) 
    {   
        if(4*j+4<n)
            AXPYF<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,alpha,A,B,C,4*j,m);
        else
            for ( IndexType jj = 4*j; jj<n; ++jj )
                AXPYV<ViewType1,ViewType2,ViewType3,Scalar,IndexType>(transA,alpha,A,B,C,jj,m);
    });
}

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void GEMV_threaded_tiled_col_based(
      const char transA,
      const char transB,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const Scalar& beta,
      const ViewType3& C)
{
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n = B.dimension_0 ();
  
    const IndexType m_per_tile = 100;//ceil(((double)100*32*8)/(sizeof(Scalar)));

    const int n_i = ceil(((double) m)/m_per_tile);

    Kokkos::parallel_for (n_i, KOKKOS_LAMBDA (const int i)
    {
      IndexType i_min = m_per_tile*i;
      IndexType i_max = (m_per_tile*(i+1) > m) ? m : m_per_tile*(i+1);
    
      for ( IndexType i=i_min; i<i_max; ++i )
        C(i,0) *= beta;
      
    
      if (transA == 'N' || transA == 'n')
        for ( IndexType j=0; j<n; ++j )
          for ( IndexType i=i_min; i<i_max; ++i ){
            //_mm_prefetch(&A(i+1,j),_MM_HINT_T1);
            //_mm_prefetch(&C(i+1,0),_MM_HINT_T1);
            //_mm_prefetch(&A(i,j+1),_MM_HINT_T1);
            C(i,0) += alpha*B(j,0)*A(i,j);
          }
      else
        for ( IndexType j=0; j<n; ++j )
          for ( IndexType i=i_min; i<i_max; ++i ){
            //_mm_prefetch(&A(j,i+1),_MM_HINT_T1);
            //_mm_prefetch(&C(i+1,0),_MM_HINT_T1);
            //_mm_prefetch(&A(j+1,i),_MM_HINT_T1);
            C(i,0) += alpha*B(j,0)*A(j,i);
          }
    });
}


template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void GEMV_threaded_row_tiles(
      const char transA,
      const char transB,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const Scalar& beta,
      const ViewType3& C,
      const int m_per_tile)
{
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n = B.dimension_0 ();

    const int unrolling_factor = Sacado_MP_Vector_GEMV_Unrolling_Factor(sizeof(Scalar));
  
    const int n_i = ceil(((double) m)/m_per_tile);

    Kokkos::parallel_for (n_i, KOKKOS_LAMBDA (const int i)
    {
      IndexType i_min = m_per_tile*i;
      IndexType i_max = (m_per_tile*(i+1) > m) ? m : m_per_tile*(i+1);
      
      const int ni = floor((i_max-i_min)/unrolling_factor);
    
      #pragma unroll
      for ( IndexType i=i_min; i<i_max; ++i )
        C(i,0) *= beta;
  
      if (transA == 'N' || transA == 'n')
        for ( IndexType j=0; j<n; ++j )
        {
          for ( IndexType i=i_min; i<i_max; i+=unrolling_factor )
            update_inner_kernel<Scalar,unrolling_factor> (&A(i,j),&B(j,0),&C(i,0),alpha);
        }
      else
        for ( IndexType j=0; j<n; ++j )
        {
          for ( IndexType i=i_min; i<i_max; i+=unrolling_factor )
            update_inner_kernel<Scalar,unrolling_factor> (&A(j,i),&B(j,0),&C(i,0),alpha);
        }
    });
}



template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void GEMV_threaded_col_tiles(
      const char transA,
      const char transB,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const Scalar& beta,
      const ViewType3& C,
      const int n_per_tile)
{
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n = B.dimension_0 ();
  
    const int pool_size = Kokkos::DefaultExecutionSpace::thread_pool_size();

    const int team_size = 4;
    #if 1
      const IndexType n_per_tile2 = n_per_tile*team_size;
    #else
      const double work_per_thread = ((double) n)/(pool_size*n_per_tile);
      const IndexType n_per_tile2 = ceil((double) n)/(pool_size*ceil(work_per_thread))*team_size;
    #endif

    const int unrolling_factor = Sacado_MP_Vector_GEMV_Unrolling_Factor(sizeof(Scalar)); //(sizeof(Scalar) == 64) ? Sacado_MP_Vector_GEMV_Unrolling_Factor_s_8 : ((sizeof(Scalar) == 128) ? Sacado_MP_Vector_GEMV_Unrolling_Factor_s_16 : Sacado_MP_Vector_GEMV_Unrolling_Factor_s_32);

    const int n_i2 = ceil(((double) n)/n_per_tile2);
    using Kokkos::TeamThreadRange;

    Kokkos::TeamPolicy<> policy = Kokkos::TeamPolicy<>( n_i2, team_size);
    typedef Kokkos::TeamPolicy<>::member_type member_type;
  
    Kokkos::parallel_for (policy, KOKKOS_LAMBDA (member_type team_member)
    {
      const IndexType j = team_member.league_rank();
      const IndexType j_min = n_per_tile2*j;
      const IndexType j_max = (j_min+n_per_tile2 > n) ? n : j_min+n_per_tile2;
      #if 1
        const IndexType i_min = (j<m) ? j:(j%m);
      #else
        const int i_team = (Kokkos::DefaultExecutionSpace::thread_pool_rank()-team_member.team_rank ())/team_size;
        const IndexType i_min = (i_team<m) ? i_team:(i_team%m);
      #endif
      
      if (transA == 'N' || transA == 'n'){
        for ( IndexType i=i_min; i<m; ++i ){
          Scalar tmp = 0.;
          Kokkos::parallel_reduce (TeamThreadRange (team_member, (j_max-j_min)/unrolling_factor),[=] (IndexType jj, Scalar& tmp_sum){
            inner_product_inner_kernel<Scalar,unrolling_factor> (&A(i,jj*unrolling_factor+j_min),&B(jj*unrolling_factor+j_min,0),&tmp_sum);
            //tmp_sum += B(jj+j_min,0)*A(i,jj+j_min);
          },tmp);
          tmp *= alpha;
          if ( team_member.team_rank () == 0) {
            Kokkos::atomic_add(&C(i,0),tmp);
          }
        }
      }
      else{
        for ( IndexType i=i_min; i<m; ++i ){
          Scalar tmp = 0.;
          Kokkos::parallel_reduce (TeamThreadRange (team_member, (j_max-j_min)/unrolling_factor),[=] (IndexType jj, Scalar& tmp_sum){
            inner_product_inner_kernel<Scalar,unrolling_factor> (&A(jj*unrolling_factor+j_min,i),&B(jj*unrolling_factor+j_min,0),&tmp_sum);
            //tmp_sum += B(jj+j_min,0)*A(jj+j_min,i);
          },tmp);
          tmp *= alpha;
          if ( team_member.team_rank () == 0) {
            Kokkos::atomic_add(&C(i,0),tmp);
          }
        }
      }
      if (transA == 'N' || transA == 'n'){
        for ( IndexType i=0; i<i_min; ++i ){
          Scalar tmp = 0.;
          Kokkos::parallel_reduce (TeamThreadRange (team_member, (j_max-j_min)/unrolling_factor),[=] (IndexType jj, Scalar& tmp_sum){
            inner_product_inner_kernel<Scalar,unrolling_factor> (&A(i,jj*unrolling_factor+j_min),&B(jj*unrolling_factor+j_min,0),&tmp_sum);
            //tmp_sum += B(jj+j_min,0)*A(i,jj+j_min);
          },tmp);
          tmp *= alpha;
          if ( team_member.team_rank () == 0) {
            Kokkos::atomic_add(&C(i,0),tmp);
          }
        }
      }
      else{
        for ( IndexType i=0; i<i_min; ++i ){
          Scalar tmp = 0.;
          Kokkos::parallel_reduce (TeamThreadRange (team_member, (j_max-j_min)/unrolling_factor),[=] (IndexType jj, Scalar& tmp_sum){
            inner_product_inner_kernel<Scalar,unrolling_factor> (&A(jj*unrolling_factor+j_min,i),&B(jj*unrolling_factor+j_min,0),&tmp_sum);
            //tmp_sum += B(jj+j_min,0)*A(jj+j_min,i);
          },tmp);
          tmp *= alpha;
          if ( team_member.team_rank () == 0) {
            Kokkos::atomic_add(&C(i,0),tmp);
          }
        }
      }
    });
}

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void GEMV_threaded_tiled_2_col_based(
      const char transA,
      const char transB,
      const Scalar& alpha,
      const ViewType1& A,
      const ViewType2& B,
      const Scalar& beta,
      const ViewType3& C)
{
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    const IndexType m = C.dimension_0 ();
    const IndexType n = B.dimension_0 ();
  
    const IndexType m_per_tile = Sacado_MP_Vector_GEMV_m_n_per_tile(sizeof(Scalar));

    const int unrolling_factor = Sacado_MP_Vector_GEMV_Unrolling_Factor(sizeof(Scalar));
    const int unrolling_factor_after_peeling = Sacado_MP_Vector_GEMV_Unrolling_Factor_After_Peeling(sizeof(Scalar));
  
    //static_assert (m_per_tile%unrolling_factor == 0, "GEMM: Tile size must be a multiple of the unrolling factor.");

    typedef EnsembleTraits<Scalar> ET;
    const int s = ET::size;

    const int n_i = ceil(((double) m)/m_per_tile);
 
    Kokkos::parallel_for (n_i, KOKKOS_LAMBDA (const int i)
    {
      IndexType i_min = m_per_tile*i;
      IndexType i_max = (m_per_tile*(i+1) > m) ? m : m_per_tile*(i+1);
    
      #pragma unroll
      for ( IndexType i=i_min; i<i_max; ++i )
        C(i,0) *= beta;
  
      i_max -= unrolling_factor_after_peeling;
      if (transA == 'N' || transA == 'n')
        for ( IndexType j=0; j<n; ++j )
        {
          for ( IndexType i=i_min; i<i_max; i+=unrolling_factor )
            update_inner_kernel<Scalar,unrolling_factor> (&A(i,j),&B(j,0),&C(i,0),alpha);
          update_inner_kernel<Scalar,unrolling_factor_after_peeling> (&A(i_max,j),&B(j,0),&C(i_max,0),alpha);
        }
      else
        for ( IndexType j=0; j<n; ++j )
        {
          for ( IndexType i=i_min; i<i_max; i+=unrolling_factor )
            update_inner_kernel<Scalar,unrolling_factor> (&A(j,i),&B(j,0),&C(i,0),alpha);
          update_inner_kernel<Scalar,unrolling_factor_after_peeling> (&A(j,i_max),&B(j,0),&C(i_max,0),alpha);
        }
    });
}
};

#endif
