//#define __KOKKOSBATCHED_INTEL_MKL__
//#define __KOKKOSBATCHED_INTEL_MKL_BATCHED__
//#define __KOKKOSBATCHED_INTEL_MKL_COMPACT_BATCHED__


#define dm_c 120
#define dk_c 256 // normally 240
#define dm_r 30

//#include "mkl.h"

#include <iostream>
#include "Stokhos_Sacado_Kokkos_MP_Vector.hpp"
#include "Kokkos_Core.hpp"

#include "KokkosBatched_Vector.hpp"

#include "KokkosBatched_Gemm_Decl.hpp"
#include "KokkosBatched_Gemm_Serial_Impl.hpp"

#include "TpetraCore_config.h"
#include "Kokkos_ArithTraits.hpp"
#include "Kokkos_Complex.hpp"
#include <type_traits>

//#include <boost/math/common_factor.hpp>

#include "EnsembleTraits.h"

// Tpetra
#include "Stokhos_Tpetra_MP_Vector.hpp"
#include "Stokhos_Tpetra_Utilities_MP_Vector.hpp"
#include "Tpetra_ConfigDefs.hpp"
#include "Tpetra_DefaultPlatform.hpp"
#include "Tpetra_Map.hpp"
#include "Tpetra_MultiVector.hpp"
#include "Tpetra_Vector.hpp"
#include "Tpetra_CrsGraph.hpp"
#include "Tpetra_CrsMatrix.hpp"
#include "Stokhos_Tpetra_CG.hpp"



#include "Quantile.h"
#include "InputParser.h"

#include "DefaultGEMV_decl.hpp"
#include "BlisGEMV_decl.hpp"
#include "MKLGEMV.h"

#include "Tiled_compact_blas.hpp"

template <typename Scalar, typename ScalarViewType> void set_matrices
(ScalarViewType A, ScalarViewType B, ScalarViewType C, Scalar &alpha, Scalar &beta) {
  
  typedef EnsembleTraits<Scalar> ET;
  size_t ensemble_size = ET::size;
  
  Scalar lin_one = (Scalar) 1.;
  for(auto s=0; s<ensemble_size; ++s)
    ET::coeff(lin_one,s) = ET::coeff(lin_one,s)*(s+1)/ensemble_size;

  Kokkos::deep_copy(A, 1.5*lin_one);
  Kokkos::deep_copy(B, 2.1*lin_one);
  Kokkos::deep_copy(C, 0*lin_one);
  
  alpha = (Scalar) 1.;
  beta = (Scalar) 0.;
}

template <typename Scalar, typename ScalarViewType> bool test_matrices
(ScalarViewType A, ScalarViewType B, ScalarViewType C, Scalar alpha, Scalar beta) {
  
  using Teuchos::ScalarTraits;
  int k = B.dimension_0 ();
  
  typedef EnsembleTraits<Scalar> ET;
  typedef ScalarTraits<Scalar> STS;
  size_t ensemble_size = ET::size;
  
  Scalar lin_one = (Scalar) 1.;
  for(auto s=0; s<ensemble_size; ++s)
    ET::coeff(lin_one,s) = ET::coeff(lin_one,s)*(s+1)/ensemble_size;
  
  bool output = MaskLogic::AND(STS::magnitude(C(0,0)-lin_one*lin_one*1.5*2.1*k)<=(1e-10*lin_one*lin_one*1.5*2.1*k));
  
  if(!output)
    std::cout << C(0,0) << " " << lin_one*lin_one*1.5*2.1*k << " " << C(0,0)-lin_one*lin_one*1.5*2.1*k << " " << (ET::coeff(C(0,0),0) == ET::coeff(lin_one*1.5*2.1*k,0)) << std::endl;
  
  return output;
}

void flush_memory(void)
{
  const int leng_a = 1073741824;
  double tmp = 0;
  //double a[leng_a];
  Kokkos::View<double*> a ("some label",leng_a);
  Kokkos::parallel_for (leng_a, KOKKOS_LAMBDA (const int i)
  {
    a(i) = (double) i;
  });
  std::cout<< " flushed " << leng_a*8E-09 << " GB" << std::endl;
}

template<typename layout> void test_function_double(int argc, char** argv, int ensemble_size, const int m, const int n, const int N, const int k_min, const int k_max, const int d_k, const std::string &filename, const int per_tile)
{
    using namespace KokkosBatched::Experimental;

    InputParser input(argc, argv);
    int algo;   

    const std::string &algo_string = input.getCmdOption("-algo");
    if(algo_string.empty())
      algo = 1;
    else
      algo = std::stoi(algo_string);   

    bool write = input.cmdOptionExists("-w");  
    bool flush = input.cmdOptionExists("-flush");
    bool transpose = input.cmdOptionExists("-t");     

    typedef double scalar;
    typedef EnsembleTraits<scalar> ET;          

    Kokkos::initialize (argc,argv);
    {
      typedef int IndexType;

      typedef Kokkos::DefaultHostExecutionSpace HostSpaceType ;
      typedef Kokkos::View<double***, layout, Kokkos::Device<Kokkos::OpenMP, Kokkos::HostSpace>, Kokkos::MemoryTraits<(unsigned int)0>> DoubleViewType;
      typedef Kokkos::View<scalar**, layout, Kokkos::Device<Kokkos::OpenMP, Kokkos::HostSpace>, Kokkos::MemoryTraits<(unsigned int)0>> ScalarViewType;


      std::ofstream timers, timers_2;
      std::string name_file, name_file_2;
      name_file = transpose ? filename+"_transposed_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+"_"+std::to_string(per_tile)+".txt": filename+"_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+"_"+std::to_string(per_tile)+".txt";
      name_file_2 = transpose ? filename+"_transposed_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+"_"+std::to_string(per_tile)+"_2.txt": filename+"_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+"_"+std::to_string(per_tile)+"_2.txt";
      timers.open(name_file);
      timers << "% m = "<< m << " n = "<< n << " " << transpose<<std::endl;

      timers_2.open(name_file_2);
      timers_2 << "% m = "<< m << " n = "<< n << " " << transpose<<std::endl;
      for (IndexType k = k_min; k < k_max; k+=d_k)
      {
        
        IndexType Am = transpose? k:m;
        IndexType Ak = transpose? m:k;

        //ScalarViewType A("A", Am,Ak), B("B", k,n), C("C", m,n);
        ScalarViewType A("A", m,k), B("B", Ak,n), C("C", Am,n);
        
        //std::cout << transpose << " " << Am << " " << Ak << std::endl;
        //std::cout << "A : (" << A.dimension_0 () << "," << A.dimension_1 () << ") " << "B : (" << B.dimension_0 () << "," << B.dimension_1 () << ") "  << "C : (" << C.dimension_0 () << "," << C.dimension_1 () << ") " << std::endl;

        scalar alpha, beta;
/*
        scalar lin_one = (scalar) 1.;
        for(auto s=0; s<ensemble_size; ++s)
          ET::coeff(lin_one,s) = ET::coeff(lin_one,s)*(s+1)/ensemble_size;

        scalar alpha = (scalar) 1.;
        scalar beta = (scalar) 0.;

        Kokkos::deep_copy(A, 1.5*lin_one);
        Kokkos::deep_copy(B, 2.1*lin_one);
        //Kokkos::deep_copy(C, 0.*lin_one);

*/
        //int strides[2]; // any integer type works in stride()
        //A.stride (strides);

        //typedef typename ScalarViewType::array_layout array_layout;
        //bool layoutL = std::is_same<array_layout, Kokkos::LayoutLeft>::value; 

        //std::cout << strides[0] << " " << strides[1] << " " << Am << " " << A.dimension_0 () << " " << Ak << " " << A.dimension_1 () << " " << transpose << " " << layoutL << std::endl;

        const char transA = transpose? 't':'n';
        const char transB = 'n';
        
        bool has_passed;

        //Kokkos::deep_copy(C, 0.*lin_one);
        double gflops = 2.0 * m * n * k * 1.0e-09 * ensemble_size;
        if (algo==0 && ensemble_size==1)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            MKLGEMV::GEMV<ScalarViewType,ScalarViewType,ScalarViewType,double,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }        
        if (algo==1)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            DefaultGEMV::GEMV_row_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==2)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            DefaultGEMV::GEMV_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        } 
        if (algo==3)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_row_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==4)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        } 
        if (algo==5)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_fused_row_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==6)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_fused_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }  
        if (algo==7)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_row_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          const int pool_size = Kokkos::DefaultExecutionSpace::thread_pool_size();
          
          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
	  /*
          if ( k <= pool_size)
            timers << " " << (boost::math::lcm(k, pool_size)/k) << std::endl;
          else
            timers << " 1" << std::endl;
	  */
        } 
        if (algo==8)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        } 
        if (algo==9)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            std::vector<double> time_vector_2 =  BlisGEMV::GEMV_threaded_fused_row_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            if ( time_vector_2.size() > 1 && i >= 2)
            {
              //timers << "% k " << k << " " << i << std::endl;
              timers_2 << k << " " << i << " " << time_vector_2[4] << " " << std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count() << " " << time_vector_2[0] << " " << time_vector_2[1] << " " << time_vector_2[2] << " " << time_vector_2[3] << " " << time_vector_2[4] << std::endl;
              //timers << "% total 1 " << time_vector_2[4] << " nanosec, total 2 " << std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count() << " nanosec, setup " << time_vector_2[0] << " nanosec, view " << time_vector_2[1] << " nanosec, tile_parallel_for " << time_vector_2[2] << " nanosec, reduction_parallel_for " << time_vector_2[3] << " nanosec, remaining " << time_vector_2[4]-time_vector_2[0]-time_vector_2[1]-time_vector_2[2]-time_vector_2[3] << " nanosec" << std::endl;
              //timers_2 << "% setup " << time_vector_2[0]/time_vector_2[4] << " view " << time_vector_2[1]/time_vector_2[4] << " tile_parallel_for " << time_vector_2[2]/time_vector_2[4] << " reduction_parallel_for " << time_vector_2[3]/time_vector_2[4] << " remaining " << (time_vector_2[4]-time_vector_2[0]-time_vector_2[1]-time_vector_2[2]-time_vector_2[3])/time_vector_2[4] << std::endl;
            }
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        } 
        if (algo==10)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_fused_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==11)
        {
          std::vector<double> time_vector;
          
          //std::cout << "A"<< std::endl;
          
          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            tiled_compact_GEMV<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();
              
              if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }
          
          auto quantiles = Quantile<double>(time_vector, dquantiles);
          
          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
              timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==12)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_tiled_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==13)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_tiled_2_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==14)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_row_tiles<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C, per_tile);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==15)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_col_tiles<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C, per_tile);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (write)
        {
          std::ofstream m_stream;  
          std::string name =  "C_";
          name = name +std::to_string(k)+"_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+".txt"; 
          m_stream.open(name);
          for(auto i=0; i<m; ++i){
            for(auto j=0; j<n; ++j){              
              m_stream << ET::coeff(C(i,j),ensemble_size-1) << " ";
            }
            m_stream << std::endl;
          }  
          m_stream.close();  
          name =  "A_";
          name = name +std::to_string(k)+"_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+".txt"; 
          m_stream.open(name);
          for(auto i=0; i<m; ++i){
            for(auto j=0; j<k; ++j){
              m_stream << ET::coeff(A(i,j),ensemble_size-1) << " ";
            }
            m_stream << std::endl;
          }  
          m_stream.close();   
          name =  "B_";
          name = name +std::to_string(k)+"_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+".txt"; 
          m_stream.open(name);
          for(auto i=0; i<k; ++i){
            for(auto j=0; j<n; ++j){
              m_stream << ET::coeff(B(i,j),ensemble_size-1) << " ";
            }
            m_stream << std::endl;
          }  
          m_stream.close();               
        }                          
      }
      timers.close();
      timers_2.close();
    }
    Kokkos::finalize();
}

template<typename scalar,typename layout> void test_function(int argc, char** argv, int ensemble_size, const int m, const int n, const int N, const int k_min, const int k_max, const int d_k, const std::string &filename, const int per_tile)
{
    using namespace KokkosBatched::Experimental;

    InputParser input(argc, argv);
    int algo;   

    const std::string &algo_string = input.getCmdOption("-algo");
    if(algo_string.empty())
      algo = 1;
    else
      algo = std::stoi(algo_string);   

    bool write = input.cmdOptionExists("-w");  
    bool flush = input.cmdOptionExists("-flush");
    bool transpose = input.cmdOptionExists("-t");     

    typedef EnsembleTraits<scalar> ET;          

    Kokkos::initialize (argc,argv);
    {
      typedef int IndexType;

      typedef Kokkos::DefaultHostExecutionSpace HostSpaceType ;
      typedef Kokkos::View<double***, layout, Kokkos::Device<Kokkos::OpenMP, Kokkos::HostSpace>, Kokkos::MemoryTraits<(unsigned int)0>> DoubleViewType;
      typedef Kokkos::View<scalar**, layout, Kokkos::Device<Kokkos::OpenMP, Kokkos::HostSpace>, Kokkos::MemoryTraits<(unsigned int)0>> ScalarViewType;


      std::ofstream timers, timers_2;
      std::string name_file, name_file_2;
      name_file = transpose ? filename+"_transposed_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+"_"+std::to_string(per_tile)+".txt": filename+"_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+"_"+std::to_string(per_tile)+".txt";
      name_file_2 = transpose ? filename+"_transposed_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+"_"+std::to_string(per_tile)+"_2.txt": filename+"_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+"_"+std::to_string(per_tile)+"_2.txt";
      timers.open(name_file);
      timers << "% m = "<< m << " n = "<< n << " " << transpose<<std::endl;

      timers_2.open(name_file_2);
      timers_2 << "% m = "<< m << " n = "<< n << " " << transpose<<std::endl;


      for (IndexType k = k_min; k < k_max; k+=d_k)
      {
        

        IndexType Am = transpose? k:m;
        IndexType Ak = transpose? m:k;

        //ScalarViewType A("A", Am,Ak), B("B", k,n), C("C", m,n);
        ScalarViewType A("A", m,k), B("B", Ak,n), C("C", Am,n);
        
        scalar alpha, beta;
/*
        scalar lin_one = (scalar) 1.;
        for(auto s=0; s<ensemble_size; ++s)
          ET::coeff(lin_one,s) = ET::coeff(lin_one,s)*(s+1)/ensemble_size;

        scalar alpha = (scalar) 1.;
        scalar beta = (scalar) 0.;

        Kokkos::deep_copy(A, 1.5*lin_one);
        Kokkos::deep_copy(B, 2.1*lin_one);
*/
        //Kokkos::deep_copy(C, 0.*lin_one);


        const char transA = transpose? 't':'n';
        const char transB = 'n';
        
        bool has_passed;

        //Kokkos::deep_copy(C, 0.*lin_one);
        double gflops = 2.0 * m * n * k * 1.0e-09 * ensemble_size;       
        if (algo==0)
        {
          write = false;
        }        
        if (algo==1)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            DefaultGEMV::GEMV_row_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==2)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            DefaultGEMV::GEMV_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        } 
        if (algo==3)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_row_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==4)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        } 
        if (algo==5)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_fused_row_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==6)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_fused_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        } 
        if (algo==7)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_row_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);
          
          const int pool_size = Kokkos::DefaultExecutionSpace::thread_pool_size();

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
	  /*
          if ( k <= pool_size)
            timers << " " << (boost::math::lcm(k, pool_size)/k) << std::endl;
          else
            timers << " 1" << std::endl;
	  */
        } 
        if (algo==8)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        } 
        if (algo==9)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            std::vector<double> time_vector_2 =  BlisGEMV::GEMV_threaded_fused_row_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            if ( time_vector_2.size() > 1 && i >= 2)
            {
              //timers << "% k " << k << " " << i << std::endl;
              timers_2 << k << " " << i << " " << time_vector_2[4] << " " << std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count() << " " << time_vector_2[0] << " " << time_vector_2[1] << " " << time_vector_2[2] << " " << time_vector_2[3] << " " << time_vector_2[4] << std::endl;
              //timers << "% total 1 " << time_vector_2[4] << " nanosec, total 2 " << std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count() << " nanosec, setup " << time_vector_2[0] << " nanosec, view " << time_vector_2[1] << " nanosec, tile_parallel_for " << time_vector_2[2] << " nanosec, reduction_parallel_for " << time_vector_2[3] << " nanosec, remaining " << time_vector_2[4]-time_vector_2[0]-time_vector_2[1]-time_vector_2[2]-time_vector_2[3] << " nanosec" << std::endl;
              timers_2 << "% setup " << time_vector_2[0]/time_vector_2[4] << " view " << time_vector_2[1]/time_vector_2[4] << " tile_parallel_for " << time_vector_2[2]/time_vector_2[4] << " reduction_parallel_for " << time_vector_2[3]/time_vector_2[4] << " remaining " << (time_vector_2[4]-time_vector_2[0]-time_vector_2[1]-time_vector_2[2]-time_vector_2[3])/time_vector_2[4] << std::endl;
            }
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        } 
        if (algo==10)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_fused_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==11)
        {
          std::vector<double> time_vector;
          
          //std::cout << "A"<< std::endl;
          
          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            tiled_compact_GEMV<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();
            
            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }
          
          auto quantiles = Quantile<double>(time_vector, dquantiles);
          
          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
              timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==12)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_tiled_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==13)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_tiled_2_col_based<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==14)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_row_tiles<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C, per_tile);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (algo==15)
        {
          std::vector<double> time_vector;

          //std::cout << "A"<< std::endl;

          has_passed = true;
          for (IndexType i = 0; i <= N; ++i)
          {
            set_matrices<scalar,ScalarViewType>(A,B,C,alpha,beta);
            if(flush) flush_memory();
            auto begin_compact = std::chrono::high_resolution_clock::now();
            BlisGEMV::GEMV_threaded_col_tiles<ScalarViewType,ScalarViewType,ScalarViewType,scalar,int> (transA, transB, alpha, A, B, beta, C, per_tile);
            auto end_compact = std::chrono::high_resolution_clock::now();
            has_passed = (has_passed && test_matrices(A,B,C,alpha,beta) );
            double total_time_compact = std::chrono::duration_cast<std::chrono::nanoseconds>(end_compact-begin_compact).count();

            if (i!=0) time_vector.push_back(total_time_compact/1e09);
          }

          auto quantiles = Quantile<double>(time_vector, dquantiles);

          timers << k << " " << gflops << " ";
          for (auto i=0; i<quantiles.size(); ++i)
            timers << gflops/quantiles[i] << " ";
          timers << " " << has_passed << std::endl;
        }
        if (write)
        {
          std::ofstream m_stream;  
          std::string name =  "C_";
          name = name +std::to_string(k)+"_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+".txt"; 
          m_stream.open(name);
          for(auto i=0; i<m; ++i){
            for(auto j=0; j<n; ++j){              
              m_stream << ET::coeff(C(i,j),ensemble_size-1) << " ";
            }
            m_stream << std::endl;
          }  
          m_stream.close();  
          name =  "A_";
          name = name +std::to_string(k)+"_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+".txt"; 
          m_stream.open(name);
          for(auto i=0; i<m; ++i){
            for(auto j=0; j<k; ++j){
              m_stream << ET::coeff(A(i,j),ensemble_size-1) << " ";
            }
            m_stream << std::endl;
          }  
          m_stream.close();   
          name =  "B_";
          name = name +std::to_string(k)+"_algo_"+std::to_string(algo)+"_"+std::to_string(ensemble_size)+".txt"; 
          m_stream.open(name);
          for(auto i=0; i<k; ++i){
            for(auto j=0; j<n; ++j){
              m_stream << ET::coeff(B(i,j),ensemble_size-1) << " ";
            }
            m_stream << std::endl;
          }  
          m_stream.close();               
        }                          
      }
      timers.close();
      timers_2.close();
    }
    Kokkos::finalize();
}




int main(int argc, char **argv){
    std::string filename;
    int m, n, N, k_min, k_max, d_k, per_tile;
      
    readInput(argc, argv, filename, m, n, N, k_min, k_max, d_k, per_tile);
    if (m<=0)
      return 0;

    InputParser input(argc, argv);
    bool layoutL = input.cmdOptionExists("-l");  

    int algo;

    const std::string &algo_string = input.getCmdOption("-algo");
    if(algo_string.empty())
      algo = 1;
    else
      algo = std::stoi(algo_string);
  
    int e_size;

    const std::string &e_string = input.getCmdOption("-ensemble");
    if(e_string.empty())
      e_size = 1;
    else
      e_size = std::stoi(e_string);

    if (algo == 0)
    {
      constexpr int ensemble_size = 1;

      typedef double scalar;

      if(layoutL)
        test_function_double<Kokkos::LayoutLeft>(argc,argv,ensemble_size,m,n,N,k_min,k_max,d_k,filename,per_tile);
      else
        test_function_double<Kokkos::LayoutRight>(argc,argv,ensemble_size,m,n,N,k_min,k_max,d_k,filename,per_tile);
    }
    else
    {
      if(e_size==8)
      {
        constexpr int ensemble_size = 8;

        typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;

        if(layoutL)
          test_function<scalar,Kokkos::LayoutLeft>(argc,argv,ensemble_size,m,n,N,k_min,k_max,d_k,filename,per_tile);
        else
          test_function<scalar,Kokkos::LayoutRight>(argc,argv,ensemble_size,m,n,N,k_min,k_max,d_k,filename,per_tile);
      }
      if(e_size==16)
      {
        constexpr int ensemble_size = 16;

        typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;

        if(layoutL)
          test_function<scalar,Kokkos::LayoutLeft>(argc,argv,ensemble_size,m,n,N,k_min,k_max,d_k,filename,per_tile);
        else
          test_function<scalar,Kokkos::LayoutRight>(argc,argv,ensemble_size,m,n,N,k_min,k_max,d_k,filename,per_tile);
      }
      if(e_size==24)
      {
        constexpr int ensemble_size = 24;

        typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;

        if(layoutL)
          test_function<scalar,Kokkos::LayoutLeft>(argc,argv,ensemble_size,m,n,N,k_min,k_max,d_k,filename,per_tile);
        else
          test_function<scalar,Kokkos::LayoutRight>(argc,argv,ensemble_size,m,n,N,k_min,k_max,d_k,filename,per_tile);
      }
      if(e_size==32)
      {
        constexpr int ensemble_size = 32;

        typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,ensemble_size,Kokkos::OpenMP>> scalar;

        if(layoutL)
          test_function<scalar,Kokkos::LayoutLeft>(argc,argv,ensemble_size,m,n,N,k_min,k_max,d_k,filename,per_tile);
        else
          test_function<scalar,Kokkos::LayoutRight>(argc,argv,ensemble_size,m,n,N,k_min,k_max,d_k,filename,per_tile);
      }
    }

    return 0;
}
