
namespace recursiveGEMV{



void recursMultiply(const double* const A, const double* const b,
                    double* const c, const long n, const long m, const long lda){
  const long jThreshold = 8192L; assert(n%jThreshold == 0);
  const long iThreshold = 64L; assert(m%iThreshold == 0);
  if ((m<=iThreshold) && (n<=jThreshold)) { // Recursion threshold
    for (long i = 0; i < m; i++)
      #pragma vector aligned
      for (long j = 0; j < n; j++)
        c[i] = A[i*lda+j] * b[j]; // Matrix-vector multiplication
  } else { // Recursive divide-and-conquer
    if (m*jThreshold > n*iThreshold) { // Split i-wise
      double c1[m/2] __attribute__((aligned(64)));
      #pragma omp task
      { recursMultiply(&A[0*lda + 0], &b[0], c1, n, m/2, lda); }
      double c2[m/2] __attribute__((aligned(64)));
      recursMultiply(&A[(m/2)*lda + 0], &b[m/2], c2, n, m/2, lda);
      #pragma omp taskwait
      c[0:m/2] += c1[0:m/2]; c[m/2:m/2] += c2[0:m/2]; // Reduction
  } else { // Split j-wise
    double c1[m] __attribute__((aligned(64)));
    #pragma omp task
    { recursMultiply(&A[0*lda + 0], &b[0], c1, n/2, m, lda); }
    double c2[m] __attribute__((aligned(64)));
    recursMultiply(&A[0*lda + n/2], &b[0], c2, n/2, m, lda);
    #pragma omp taskwait
    c[0:m] += c1[0:m]; c[0:m] += c2[0:m]; // Reduction
} } }

void Multiply(const double* const A, const double* const b,
              double* const c, const long n, const long m){
  #pragma omp parallel
  #pragma omp single
  { recursMultiply(A, b, c, n, m, n); }
}

template<class ViewType1,
         class ViewType2,
         class ViewType3,
         class Scalar,
         class IndexType>
void GEMV(
      const char transA,
      const char transB,
      const double alpha,
      const ViewType1& A,
      const ViewType2& B,
      const double beta,
      const ViewType3& C)
{
    // Assert that A, B, and C are in fact matrices
    static_assert (ViewType1::rank == 2, "GEMM: A must have rank 2 (be a matrix).");
    static_assert (ViewType2::rank == 2, "GEMM: B must have rank 2 (be a matrix).");
    static_assert (ViewType3::rank == 2, "GEMM: C must have rank 2 (be a matrix).");

    // Get the dimensions
    //const IndexType m = C.dimension_0 ();
    //const IndexType n = B.dimension_0 ();
  
    const IndexType m = A.dimension_0 ();
    const IndexType n = A.dimension_1 ();
  
    const IndexType n_rhs = C.dimension_1 ();
    bool trans_bool = (transA == 'N' || transA == 'n');

    typedef typename ViewType1::array_layout array_layout;
    bool layoutL = std::is_same<array_layout, Kokkos::LayoutLeft>::value;   

    int strides[2]; 
    A.stride (strides);
    const int lda = layoutL ? strides[1] : strides[0];     

    double *a = &A(0,0);
    double *x = &B(0,0);
    double *y = &C(0,0);    

    Multiply(a, x, y, n, m);
}

};
