#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"


    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos

    precondList = Teuchos.ParameterList()
    precondSubList = Teuchos.ParameterList()
    precondSubSubList = Teuchos.ParameterList()
    precondSubSubList['fact: iluk level-of-fill'] = 2
    precondSubList['Prec Type'] = "RILUK"
    precondSubList['Overlap'] = 1
    precondSubList['Ifpack2 Settings'] = precondSubSubList
    precondList['Ifpack2'] = precondSubList

    mueluParams = Teuchos.ParameterList()
    #mueluParamsSub1 = Teuchos.ParameterList()
    #mueluParamsSub1['PDE equations'] = 4
    #mueluParams['Matrix'] = mueluParamsSub1
    mueluParams['number of equations'] = 4
    mueluParams['max levels'] = 1
    mueluParams['multigrid algorithm'] = "unsmoothed"
    mueluParams['smoother: type'] = "RELAXATION"

    mueluParamsSub2 = Teuchos.ParameterList()

    mueluParamsSub2['relaxation: type'] = "Symmetric Gauss-Seidel"
    mueluParamsSub2['relaxation: sweeps'] = 4
    mueluParamsSub2['relaxation: damping factor'] = 1.25

    mueluParams['coarse: type'] = "RELAXATION"

    mueluParamsSub3 = Teuchos.ParameterList()

    mueluParamsSub3['relaxation: type'] = "Symmetric Gauss-Seidel"
    mueluParamsSub3['relaxation: sweeps'] = 4
    mueluParamsSub3['relaxation: damping factor'] = 1.25


    mueluParams['smoother: params'] = mueluParamsSub2
    mueluParams['coarse: params'] = mueluParamsSub3

    solverList = Teuchos.ParameterList()
    solverList['type']= "BlockGmres"
    solverList['Convergence Tolerance'] = 10**(-8)
    solverList['Maximum Iterations'] = 300
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 25
    #solverList["Implicit Residual Scaling"] = "None"
    #solverList["Explicit Residual Scaling"] = "None"
    solverList['ifpackParams'] = precondList
    solverList['mueluParams'] = mueluParams

    msh = gmsh.MeshLoader("3bodies.geo",__file__).execute()

    pbl = m.Problem(msh,comm)

    m.Medium(pbl, "Body 1", "TZM",330000,0.36,0.114,0.5E-05)
    m.Medium(pbl, "Body 2", "TZM",66000,0.36,0.114,0.5E-05)
    m.Medium(pbl, "Body 3", "TZM",330000,0.36,0.114,0.5E-05)

    clamped = tbox.Pt(1,1,1)
    values_clamped = tbox.Pt(0,0,0)
    values_clamped2 = tbox.Pt(30,0,0)

    yclamped = tbox.Pt(0,1,0)
    zclamped = tbox.Pt(0,0,1)

    m.uDirichlet(pbl, "Body 3 surface x1", "clamped",1,0.,0,0.,0,0.,1,100.)
    #m.uDirichlet(pbl, "Body 2 point x1", "clamped",1,0.,0,0.,0,0.,1,0.)
    m.uDirichlet(pbl, "Body 1 surface x0","moved",1,50.,0,0.,0,0.,1,0.)
    m.uDirichlet(pbl, "y sym","y symetry bc",0,0.,1,0.,0,0.,0,0.)
    m.uDirichlet(pbl, "z sym","z symetry bc",0,0.,0,0.,1,0.,0,0.)

    norm1 = tbox.Pt(1,0,0)
    cont = m.Contact(pbl,"Body 1 surface x1","contact",norm1)
    cont.setMaster(pbl,"Body 2 surface x0",norm1)

    cont2 = m.Contact(pbl,"Body 2 surface x1","contact",norm1)
    cont2.setMaster(pbl,"Body 3 surface x0",norm1)

    slv = m.Solver(pbl,args.k,solverList,4)

    print slv

    import mrstlnos.viewer as v
    gui = v.MeshViewer(pbl,slv,msh)
    gui.vmin = 0
    gui.vmax = 1
    gui.start()

if __name__ == "__main__":
    main()
