#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"


    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos

    preCondList = Teuchos.ParameterList()
    preCondList['type'] = "RILUK"
    preCondList['fact: iluk level-of-fill'] = 1

    solverList = Teuchos.ParameterList()
    solverList['type']= "BlockGmres"
    solverList['Convergence Tolerance'] = 10**(-10)
    solverList['Maximum Iterations'] = 1000
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 20
    solverList['precondList'] = preCondList

    msh = gmsh.MeshLoader("mirrors01_ansys_partition_4.msh",__file__).myrank_execute(rank)


    pbl = m.Problem(msh,comm)
    m.Medium(pbl, "Volume", "TZM",330000,0.36,0.114,0.5E-05)
    clamped = [1,1,1]
    values_clamped = [0.,0.,0.]
    values_clamped2 = [0,0,100]
    m.uDirichlet(pbl, "M1_S1", "clamped",1,0.,1,0.,1,0.,1,70.)
    m.uDirichlet(pbl, "M1_S2", "clamped",1,0.,1,0.,1,0.,1,70.)
    m.uDirichlet(pbl, "M1_S3", "clamped",1,0.,1,0.,1,0.,1,70.)
    m.uDirichlet(pbl, "HF", "moved",0,0.,0,0.,0,0.,1,270.)

    do_ensemble = False
    slv = m.Solver(pbl,args.k,solverList,4)


    if not args.nogui:
        if rank == 0:
            all_msh = gmsh.MeshLoader("mirrors01_ansys_partition_4.msh",__file__).execute()
            import mrstlnos.viewer as v
            gui = v.MeshViewer(pbl,slv,all_msh)
            gui.vmin = 0
            gui.vmax = 1
            gui.start()
        else:
            slv.start()
    else:
        slv.start()

if __name__ == "__main__":
    main()
