#ifndef WDIRICHLET_LIST_H
#define WDIRICHLET_LIST_H

#include "mrstlnos.h"

#include "Map.h"
#include "wuDirichlet.h"
#include "wProblem.h"
#include "wTag.h"
#include "wNode.h"

#include "Elements_list.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Kokkos_ViewFactory.hpp>

#include "EnsembleTraits.h"

using Teuchos::RCP;

namespace mrstlnos
{
    namespace discrete_problem
    {
        template<typename scalar> class Dirichlet_list
        {
        private:
            Kokkos::View<int**, Kokkos::LayoutRight>  dirichlet_nodes_list;
            Kokkos::View<scalar***, Kokkos::LayoutRight>  dirichlet_values_list;

            Tpetra::Vector<>::local_ordinal_type dirichlet_number;
            Tpetra::Vector<>::local_ordinal_type dirichlet_size;
            Tpetra::Vector<>::local_ordinal_type dirichlet_nodes;

            Tpetra::Vector<>::local_ordinal_type dirichlet_time;
            Tpetra::Vector<>::local_ordinal_type dirichlet_dofs;
            Tpetra::Vector<>::local_ordinal_type dirichlet_values;

        public:
            Dirichlet_list();
            Dirichlet_list(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<Elements_list> elements_list);
            inline Tpetra::Vector<>::local_ordinal_type getDirichletNumber(){return dirichlet_number;};
            inline Tpetra::Vector<>::local_ordinal_type getDirichletSize(local_ordinal_type i)
                    {return dirichlet_nodes_list(i,dirichlet_size);};
            inline Tpetra::Vector<>::local_ordinal_type getDirichletNode(local_ordinal_type i, Tpetra::Vector<>::local_ordinal_type n1)
                    {return dirichlet_nodes_list(i,dirichlet_nodes + n1);};
            inline scalar getDirichletDof(local_ordinal_type i, Tpetra::Vector<>::local_ordinal_type j)
                    {return dirichlet_values_list(i,0,dirichlet_dofs+j);};
            inline scalar getDirichletValue(local_ordinal_type i, Tpetra::Vector<>::local_ordinal_type j)
                    {return dirichlet_values_list(i,0,dirichlet_values+j);};
        };
    };
};

#endif //WDIRICHLET_LIST_H
