#ifndef WDISCRETE_PROBLEM_H
#define WDISCRETE_PROBLEM_H

#include "mrstlnos.h"
#include "Map.h"
#include "Domain.h"
#include "Loads.h"
#include "Algebraic.h"
#include "ElementMatrices.h"
#include "ElementVectors.h"

#include <Xpetra_Map.hpp>
#include <Xpetra_TpetraMap.hpp>
#include <Xpetra_BlockedMap.hpp>
#include <Xpetra_BlockedCrsMatrix.hpp>
#include <Xpetra_BlockedMultiVector.hpp>
#include <Xpetra_TpetraCrsMatrix.hpp>
#include <Xpetra_TpetraMultiVector.hpp>
#include <Xpetra_CrsMatrixWrap.hpp>
//#include <Xpetra_UseShortNames.hpp>

#include "LinearSolver.h"

#include "mortar2.h"

#include <mpi.h>

#include "MaskTraits.h"

namespace mrstlnos
{
    template<typename scalar> class Discrete_problem
    {
    private:
        Kokkos::View<scalar*, Kokkos::LayoutRight> old_old_activity;
        Kokkos::View<scalar*, Kokkos::LayoutRight> old_activity;

    public:
#ifndef SWIG
        Teuchos::RCP<discrete_problem::Domain<scalar> >   domain;
        Teuchos::RCP<discrete_problem::Loads<scalar> >    loads;
        Teuchos::RCP<discrete_problem::Algebraic<scalar> > algebraic;
#endif
        size_t numPrimalDPN;

        Discrete_problem ()
        { };

        Discrete_problem (Problem &pbl, size_t _numPrimalDPN, Teuchos::RCP<Teuchos::ParameterList> randomParams, scalar * m_rv_values)
        {
            numPrimalDPN = _numPrimalDPN;
            using namespace discrete_problem;

            LinearSolver::getTimers()["map"].start();
            Teuchos::RCP<Map> map = Teuchos::rcp (new  Map(pbl,numPrimalDPN));
            LinearSolver::getTimers()["map"].stop();
            LinearSolver::getTimers()["domain"].start();
            domain = Teuchos::rcp (new Domain<scalar>(pbl,map,randomParams,m_rv_values));
            LinearSolver::getTimers()["domain"].stop();
            LinearSolver::getTimers()["loads"].start();
            loads = Teuchos::rcp (new  Loads<scalar> (pbl,map,domain));
            LinearSolver::getTimers()["loads"].stop();
            algebraic = Teuchos::rcp (new  Algebraic<scalar> (map, domain, loads));
        };
#ifndef SWIG
        void createBlockMatrix(scalar gamma, bool write);
        void createBlockMVector(bool write);
        void computeMatrices(void);
        void computeContactMatrices(MPI_Comm comm);
        void computeLoads(void);
        bool updateActivity(int active_set_iteration);
#endif
    };
};

#endif //WDISCRETE_PROBLEM_H
