#ifndef WDOMAIN_H
#define WDOMAIN_H

#include "mrstlnos.h"
#include "Map.h"
#include "Elements_list.h"
#include "Nodes_list.h"
#include "Materials_list.h"
#include "Contacts_list.h"
#include "RandomField.h"

using Teuchos::RCP;

namespace mrstlnos
{
    namespace discrete_problem
    {
        template<typename scalar> class Domain
        {
        public:
            Teuchos::RCP<Elements_list>           elements_list;
            Teuchos::RCP<Nodes_list>              nodes_list;
            Teuchos::RCP<Materials_list<scalar>>  materials_list;
            Teuchos::RCP<Contacts_list>           contacts_list;
            Teuchos::RCP<RandomField<scalar,Kokkos::DefaultExecutionSpace>> random_field;

            Domain(){};
            Domain(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<Teuchos::ParameterList> randomParams, scalar * m_rv_values)
            {
                elements_list = Teuchos::rcp (new Elements_list(pbl,map));
                nodes_list = Teuchos::rcp (new Nodes_list(pbl,map));
                materials_list = Teuchos::rcp (new Materials_list<scalar>(pbl,map));
                contacts_list = Teuchos::rcp (new Contacts_list(pbl,map, elements_list, nodes_list));
                random_field = Teuchos::rcp (new RandomField<scalar,Kokkos::DefaultExecutionSpace>(randomParams,m_rv_values));
            };
        };
    };
};

#endif //WDOMAIN_H
