#ifndef WLINEARSOLVER_H
#define WLINEARSOLVER_H

#include "mrstlnos.h"
#include "wObject.h"
#include "wTimer.h"
#include "wTimers.h"
#include "wTag.h"
#include "wMedium.h"
#include "wNode.h"
#include "wMem.h"
#include "Teuchos_ParameterList.hpp"

#include "wProblem.h"
#include "wContact.h"
#include "wMshData.h"
#include "wDisplayHook.h"
#include "wElement.h"

// Tpetra
#include "Stokhos_Tpetra_MP_Vector.hpp"
#include "Stokhos_Tpetra_Utilities_MP_Vector.hpp"
#include "Tpetra_ConfigDefs.hpp"
#include "Tpetra_DefaultPlatform.hpp"
#include "Tpetra_Map.hpp"
#include "Tpetra_MultiVector.hpp"
#include "Tpetra_Vector.hpp"
#include "Tpetra_CrsGraph.hpp"
#include "Tpetra_CrsMatrix.hpp"
#include "Stokhos_Tpetra_CG.hpp"

#include <Teuchos_Array.hpp>
#include <Teuchos_GlobalMPISession.hpp>
#include <Teuchos_oblackholestream.hpp>
#include <Teuchos_ScalarTraits.hpp>
#include <Teuchos_RCP.hpp>
#include "Teuchos_ParameterList.hpp"

#include "Stokhos_Amesos2_MP_Vector.hpp"
#include "Amesos2_Factory.hpp"

#include "Stokhos_Ifpack2_MP_Vector.hpp"
#include "Ifpack2_Factory.hpp"

#include "Belos_Tpetra_MP_Vector.hpp"
#include "BelosLinearProblem.hpp"
#include "BelosPseudoBlockGmresSolMgr.hpp"
#include "BelosPseudoBlockCGSolMgr.hpp"
#include "BelosBlockGmresSolMgr.hpp"

#include "Kokkos_ViewFactory.hpp"
#include "Kokkos_Vector.hpp"

#include <Epetra_Map.h>
#include "Epetra_MpiComm.h"

#include "BelosConfigDefs.hpp"
#include "BelosThyraAdapter.hpp"

#include "RandomField.h"

#include "wResults.h"
#include "wLazy.h"
#include "Map.h"

#include <map>

#include <limits>

#include "EnsembleTraits.h"

#include "Results_decl.hpp"

namespace mrstlnos {

class MRSTLNOS_API LinearSolver
{
public:
#ifndef SWIG
    Problem &pbl;
#endif
    int nthreads;            ///< nb of threads for the assembly
    Teuchos::RCP<Teuchos::ParameterList> solverList;
    const int numPrimalDPN;
    bool ensemble_propagation;
    const size_t minimum_ensemble_size;
    bool signorini;
#ifndef SWIG
    DisplayHook *dhook;

#endif
public:
    LinearSolver(Problem &_pbl, int _nthreads, Teuchos::RCP<Teuchos::ParameterList> _solverList,int _numPrimalDPN, bool _ensemble_propagation, size_t _minimum_ensemble_size, bool _signorini) : pbl(_pbl), nthreads(_nthreads), solverList(_solverList), numPrimalDPN(_numPrimalDPN), ensemble_propagation(_ensemble_propagation),minimum_ensemble_size(_minimum_ensemble_size), signorini(_signorini)
    {
        dhook = NULL;
    };
    static fwk::Timers &getTimers()
    {
        static fwk::Timers tms;
        return tms;
    };
    void setGUI(DisplayHook &hook) { dhook = &hook; }
    virtual ~LinearSolver() { }
protected:
    template<typename scalar> void writeResults (
        Teuchos::RCP<Tpetra::Vector<scalar,Tpetra::Vector<>::local_ordinal_type,Tpetra::Vector<>::global_ordinal_type> > xPost,
        size_t numNodes,
        std::string name,
        size_t ast,
        bool final_result,
        bool write_txt,
        Teuchos::RCP<discrete_problem::Map> map,
        Teuchos::RCP<RandomField<scalar,Kokkos::DefaultExecutionSpace>> random_field
        )
    {
      LinearSolver::getTimers()["write results"].start();
      
      typedef Tpetra::Vector<>::local_ordinal_type LocalOrdinal;
      typedef Tpetra::Vector<>::global_ordinal_type GlobalOrdinal;

      Write_Results<scalar,LocalOrdinal,GlobalOrdinal> (pbl,xPost,map,
              random_field,
              name, ast, final_result, write_txt);
      
      LinearSolver::getTimers()["write results"].stop();
    };
};

}

#endif //WSOLVER_H
