#ifndef WMATERIALS_LIST_H
#define WMATERIALS_LIST_H

#include "Stokhos_Sacado_Kokkos_MP_Vector.hpp"

#include "mrstlnos.h"

#include "Map.h"
#include "wProblem.h"
#include "wTag.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"
#include "wMedium.h"

#include "Material.h"

#include <Kokkos_Core.hpp>
//#include "Sacado.hpp"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Kokkos_ViewFactory.hpp>
#include <vector>

#include "EnsembleTraits.h"

using Teuchos::RCP;

namespace mrstlnos
{
    namespace discrete_problem
    {
        template<typename scalar> class Materials_list
        {
        private:
            Kokkos::View<scalar***, Kokkos::LayoutRight>  materials;

            //Kokkos::View<scalar**, Kokkos::LayoutRight> material_properties;

            Tpetra::Vector<>::local_ordinal_type materials_number;


        public:

            Tpetra::Vector<>::local_ordinal_type materials_T;
            Tpetra::Vector<>::local_ordinal_type materials_E;
            Tpetra::Vector<>::local_ordinal_type materials_nu;
            Tpetra::Vector<>::local_ordinal_type materials_k;
            Tpetra::Vector<>::local_ordinal_type materials_beta;

            Materials_list();
            Materials_list(Problem &pbl, Teuchos::RCP<Map> map);
            /*
            Material<scalar> getMaterial(Sacado::Fad::SFad<scalar,32>  T, int curent_material)
            {
                size_t materials_T = 0;

                typedef EnsembleTraits<scalar> ET;
                const int ensemble_size = ET::size;

                Kokkos::View<scalar*, Kokkos::LayoutRight> material_properties("A",materials.dimension_2());
                auto material_properties_size = 5;//materials.dimension_0();

                for (int j=0; j<ensemble_size; ++j)
                {
                    typename ET::value_type Tj = ET::coeff(T.val(),j);

                    size_t index_0 = 0;
                    size_t index_1 = materials.dimension_1()-1;

                    if (Tj <= ET::coeff(materials(curent_material,index_0,materials_T),j))
                        for(auto i=0; i< material_properties_size; ++i)
                            ET::coeff(material_properties(i),j) = ET::coeff(materials(curent_material,index_0,i),j);
                    else if (Tj > ET::coeff(materials(curent_material,index_1,materials_T),j))
                        for(auto i=0; i< material_properties_size; ++i)
                            ET::coeff(material_properties(i),j) = ET::coeff(materials(curent_material,index_1,i),j);
                    else
                    {
                        while(index_1!=index_0+1)
                        {
                            size_t index_c = index_0 + (index_1-index_0)/2;

                            if(Tj == ET::coeff(materials(curent_material,index_c,materials_T),j))
                            {
                                index_1 = index_c;
                                index_0 = index_1-1;
                            }
                            else if (Tj < ET::coeff(materials(curent_material,index_c,materials_T),j))
                                index_1 = index_c;
                            else
                                index_0 = index_c;
                        }

                        double coeff = (Tj-ET::coeff(materials(curent_material,index_0,materials_T),j)) /(ET::coeff(materials(curent_material,index_1,materials_T),j)-ET::coeff(materials(curent_material,index_0,materials_T),j));
                        for(auto i=0; i< material_properties_size; ++i)
                            ET::coeff(material_properties(i),j) = ET::coeff(materials(curent_material,index_0,i),j)
                                                     + coeff * (ET::coeff(materials(curent_material,index_1,i),j)-ET::coeff(materials(curent_material,index_0,i),j));
                    }
                }
                Material<scalar> material = Material<scalar> (material_properties);
                return material;
            };
            */
            Material<scalar> getMaterial(int curent_material,int i_thread)
            {
                scalar E;
                scalar nu;
                scalar k;
                scalar beta;

                size_t materials_T = 0;

                typedef EnsembleTraits<scalar> ET;
                const int ensemble_size = ET::size;

                for (int j=0; j<ensemble_size; ++j)
                {
                    typename ET::value_type Tj = 0.;

                    size_t index_0 = 0;
                    size_t index_1 = materials.dimension_1()-1;

                    if (Tj <= ET::coeff(materials(curent_material,index_0,materials_T),j))
                    {
                        ET::coeff(E,j) = ET::coeff(materials(curent_material,index_0,1),j);
                        ET::coeff(nu,j) = ET::coeff(materials(curent_material,index_0,2),j);
                        ET::coeff(k,j) = ET::coeff(materials(curent_material,index_0,3),j);
                        ET::coeff(beta,j) = ET::coeff(materials(curent_material,index_0,4),j);
                    }
                    else if (Tj > ET::coeff(materials(curent_material,index_1,materials_T),j))
                    {
                        ET::coeff(E,j) = ET::coeff(materials(curent_material,index_1,1),j);
                        ET::coeff(nu,j) = ET::coeff(materials(curent_material,index_1,2),j);
                        ET::coeff(k,j) = ET::coeff(materials(curent_material,index_1,3),j);
                        ET::coeff(beta,j) = ET::coeff(materials(curent_material,index_1,4),j);
                    }
                    else
                    {
                        while(index_1!=index_0+1)
                        {
                            size_t index_c = index_0 + (index_1-index_0)/2;

                            if(Tj == ET::coeff(materials(curent_material,index_c,materials_T),j))
                            {
                                index_1 = index_c;
                                index_0 = index_1-1;
                            }
                            else if (Tj < ET::coeff(materials(curent_material,index_c,materials_T),j))
                                index_1 = index_c;
                            else
                                index_0 = index_c;
                        }

                        double coeff = (Tj-ET::coeff(materials(curent_material,index_0,materials_T),j)) /(ET::coeff(materials(curent_material,index_1,materials_T),j)-ET::coeff(materials(curent_material,index_0,materials_T),j));


                        ET::coeff(E,j) = ET::coeff(materials(curent_material,index_0,1),j)
                                                     + coeff * (ET::coeff(materials(curent_material,index_1,1),j)-ET::coeff(materials(curent_material,index_0,1),j));
                        ET::coeff(nu,j) = ET::coeff(materials(curent_material,index_0,2),j)
                                                     + coeff * (ET::coeff(materials(curent_material,index_1,2),j)-ET::coeff(materials(curent_material,index_0,2),j));
                        ET::coeff(k,j) = ET::coeff(materials(curent_material,index_0,3),j)
                                                     + coeff * (ET::coeff(materials(curent_material,index_1,3),j)-ET::coeff(materials(curent_material,index_0,3),j));
                        ET::coeff(beta,j) = ET::coeff(materials(curent_material,index_0,4),j)
                                                     + coeff * (ET::coeff(materials(curent_material,index_1,4),j)-ET::coeff(materials(curent_material,index_0,4),j));
                    }
                }

                Material<scalar> material(E,nu,k,beta);
                return material;
            };
        };

        template<typename scalar> Materials_list<scalar>::Materials_list()
        {
            materials_number =0;
            materials_T = 0;
            materials_E = 0;
            materials_nu = 0;
            materials_k = 0;
            materials_beta = 0;

            materials = Kokkos::View<scalar***, Kokkos::LayoutRight> ("R",materials_number,0);

            //material_properties = Kokkos::View<scalar**, Kokkos::LayoutRight> ("R",0,0);
        }

        template<typename scalar> Materials_list<scalar>::Materials_list(Problem &pbl, Teuchos::RCP<Map> map)
        {
            materials_number = pbl.media.size();

            materials_T = 0;
            materials_E = 1;
            materials_nu = 2;
            materials_k = 3;
            materials_beta = 4;

            typedef EnsembleTraits<scalar> ET;
            const int ensemble_size = ET::size;

            materials = Kokkos::View<scalar***, Kokkos::LayoutRight> ("R",materials_number,1,5);
            for (auto i = 0; i < materials_number; ++i)
            {
                materials(i,0,materials_T) = (scalar) 0; // !!! needed to be changed latter on !!!
                for (int j=0; j<ensemble_size; ++j)
                {
                    ET::coeff(materials(i,0,materials_E),j) = (double) pbl.media[i]->E_vector[j];
                    ET::coeff(materials(i,0,materials_nu),j) = (double) pbl.media[i]->nu_vector[j];
                }
                materials(i,0,materials_k) = (scalar) pbl.media[i]->k;
                materials(i,0,materials_beta) = (scalar) pbl.media[i]->beta;

                //std::cout << materials(i,0,materials_E) << " " << materials(i,0,materials_nu) << std::endl;
            }
            //int pool_size = 4;
            //material_properties = Kokkos::View<scalar**, Kokkos::LayoutRight> ("R",pool_size,5);
        }
    };
};

#endif //WMATERILAS_LIST_H
