#ifndef WPRELOAD_LIST_H
#define WPRELOAD_LIST_H

#include "mrstlnos.h"

#include "Map.h"
#include "wContact.h"
#include "wTag.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Kokkos_ViewFactory.hpp>

#include "EnsembleTraits.h"

using Teuchos::RCP;

namespace mrstlnos
{
    namespace discrete_problem
    {
        template<typename scalar> class Preload_list
        {
        private:
            Kokkos::View<scalar*, Kokkos::LayoutRight>  contacts_dg;

        public:
            Preload_list();
            Preload_list(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<Elements_list> elements_list);
            inline scalar getPreloadValue(local_ordinal_type i)
                    {return contacts_dg(i);};
        };

        template<typename scalar> Preload_list<scalar>::Preload_list()
        {
            contacts_dg = Kokkos::View<scalar*, Kokkos::LayoutRight> ("R",0);

        }

        template<typename scalar> Preload_list<scalar>::Preload_list(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<Elements_list> elements_list)
        {
            int contacts_number = pbl.Contacts.size();

            typedef EnsembleTraits<scalar> ET;
            const int ensemble_size = ET::size;

            contacts_dg = Kokkos::View<scalar*, Kokkos::LayoutRight> ("R",contacts_number);
            for (auto i = 0; i < contacts_number; ++i)
                for (int j=0; j<ensemble_size; ++j)
                    ET::coeff(contacts_dg(i),j) = pbl.Contacts[i]->dg_vector[j];
        }
    };
};

#endif //WPRELOAD_LIST_H
