#ifndef WVECTORS_H
#define WVECTORS_H

#include "mrstlnos.h"

#include "Map.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Xpetra_BlockedMultiVector.hpp>
#include <Xpetra_MultiVector.hpp>

using Teuchos::RCP;

typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;
typedef Xpetra::BlockedMap<>::node_type node_type;

namespace mrstlnos
{
    namespace discrete_problem
    {
        template<typename scalar> class Vectors
        {
        public:
            RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > bWO;
            RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > b;

            RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > x;
            RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > lagrange;

            RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > initialGap;

            RCP<Xpetra::BlockedMultiVector<scalar,local_ordinal_type,global_ordinal_type,node_type> > rhsBlockedMultiVector;
            RCP<Xpetra::BlockedMultiVector<scalar,local_ordinal_type,global_ordinal_type,node_type> > solutionBlockedMultiVector;

            RCP<Xpetra::MultiVector<scalar,local_ordinal_type,global_ordinal_type,node_type> > rhsMultiVector;
            RCP<Xpetra::MultiVector<scalar,local_ordinal_type,global_ordinal_type,node_type> > solutionMultiVector;

            Vectors(){};
            Vectors(RCP<Map> map)
            {
                bWO = rcp (new Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> (map->mapDofsWO, true));
                b = rcp (new Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> (map->mapDofs, true));
                x = rcp (new Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> (map->mapDofs, true));
            };
        };
    };
};

#endif //WVECTORS_H
