#include "Tpetra_Vector.hpp"
#include "Kokkos_ViewFactory.hpp"
#include "Tpetra_Map.hpp"
#include <mpi.h>
#include "Tpetra_CrsGraph.hpp"
#include <Epetra_Map.h>

#include "Tpetra_ConfigDefs.hpp"
#include "Tpetra_DefaultPlatform.hpp"
#include "Tpetra_Map.hpp"
#include "Tpetra_MultiVector.hpp"
#include "Tpetra_Vector.hpp"
#include "Tpetra_CrsGraph.hpp"
#include "Tpetra_CrsMatrix.hpp"

#include "Elements_list.h"
#include "Nodes_list.h"
#include "Contacts_list.h"
#include "Map.h"
#include "Matrices.h"


Teuchos::RCP<Tpetra::CrsMatrix<>> compute_B(Teuchos::RCP<mrstlnos::discrete_problem::Elements_list> elements_list,
                                            Teuchos::RCP<mrstlnos::discrete_problem::Nodes_list> nodes_list,
                                            Teuchos::RCP<mrstlnos::discrete_problem::Contacts_list> contacts_list,
                                            Teuchos::RCP<mrstlnos::discrete_problem::Map> map, MPI_Comm comm, bool SignoriniCheck);


template<typename scalar> void compute_B(Teuchos::RCP<mrstlnos::discrete_problem::Matrices<scalar>> matrices,
                                         Teuchos::RCP<mrstlnos::discrete_problem::Elements_list> elements_list,
                                         Teuchos::RCP<mrstlnos::discrete_problem::Nodes_list> nodes_list,
                                         Teuchos::RCP<mrstlnos::discrete_problem::Contacts_list> contacts_list,
                                         Teuchos::RCP<mrstlnos::discrete_problem::Map> map, MPI_Comm comm)
{
    std::cout << " compute_B : a" << std::endl;
    Teuchos::RCP<Tpetra::CrsMatrix<>> B_double = compute_B(elements_list, nodes_list, contacts_list, map, comm, true);
    Teuchos::RCP<Tpetra::CrsMatrix<>> B_G_double;
    if(contacts_list->isSignorini())
        B_G_double = compute_B(elements_list, nodes_list, contacts_list, map, comm, false);
    else
        B_G_double = B_double;

    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    std::cout << " compute_B : b" << std::endl;
    /*
    matrices->B = rcp(new Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> (map->mapLagrangeDofs,map->mapDofs,0));
    matrices->B_G = rcp(new Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> (map->mapLagrangeDofs,map->mapDofs,0));
    matrices->B_T = rcp(new Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> (map->mapDofs,map->mapLagrangeDofs,0));
    matrices->C = rcp(new Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> (map->mapLagrangeDofs,map->mapLagrangeDofs,1));
    */

    std::cout << " compute_B : c" << std::endl;
    matrices->B = rcp(new Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> (map->mapLagrangeDofs,0));
    matrices->B_2 = rcp(new Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> (map->mapLagrangeDofs,0));
    matrices->B_G = rcp(new Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> (map->mapLagrangeDofs,0));
    matrices->B_T = rcp(new Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> (map->mapDofs,0));
    matrices->C = rcp(new Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> (map->mapLagrangeDofs,1));
    matrices->Cb = rcp(new Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> (map->mapLagrangeDofs,1));

    std::cout << " compute_B : d" << std::endl;
    for (auto i=0; i< map->mapLagrangeDofs->getNodeNumElements(); ++i)
    {
        if(map->mapLagrangeDofs->isNodeLocalElement (i))
        {
            local_ordinal_type local_index = i;
            global_ordinal_type global_index = map->mapLagrangeDofs->getGlobalElement(local_index);
            /*
            Teuchos::ArrayView<const local_ordinal_type> local_indices;
            Teuchos::ArrayView<const double> values ;

            B_double->getLocalRowView (local_index, local_indices, values);

            for ( auto k = 0; k < local_indices.size() ; ++k)
            {
                matrices->B->insertGlobalValues(global_index,
                    Teuchos::tuple<global_ordinal_type> (map->mapDofs->getGlobalElement(local_indices[k])),
                    Teuchos::tuple<scalar> ((scalar) values[k]));

                matrices->B_T->insertGlobalValues(map->mapDofs->getGlobalElement(local_indices[k]),
                    Teuchos::tuple<global_ordinal_type> (global_index),
                    Teuchos::tuple<scalar> ((scalar) values[k]));
            }

            B_G_double->getLocalRowView (local_index, local_indices, values);

            for ( auto k = 0; k < local_indices.size() ; ++k)
            {
                matrices->B_G->insertGlobalValues(global_index,
                    Teuchos::tuple<global_ordinal_type> (map->mapDofs->getGlobalElement(local_indices[k])),
                    Teuchos::tuple<scalar> ((scalar) values[k]));
            }
            */

            Teuchos::ArrayView<const local_ordinal_type> local_indices;
            Teuchos::ArrayView<const double> values ;

            B_double->getLocalRowView (local_index, local_indices, values);

            for ( auto k = 0; k < local_indices.size() ; ++k)
            {
                global_ordinal_type column_index = B_double->getColMap()->getGlobalElement(local_indices[k]);

                matrices->B->insertGlobalValues(global_index,
                    Teuchos::tuple<global_ordinal_type> (column_index),
                    Teuchos::tuple<scalar> ((scalar) values[k]));

                matrices->B_2->insertGlobalValues(global_index,
                    Teuchos::tuple<global_ordinal_type> (column_index),
                    Teuchos::tuple<scalar> ((scalar) values[k]));

                matrices->B_T->insertGlobalValues(column_index,
                    Teuchos::tuple<global_ordinal_type> (global_index),
                    Teuchos::tuple<scalar> ((scalar) values[k]));
            }

            B_G_double->getLocalRowView (local_index, local_indices, values);

            for ( auto k = 0; k < local_indices.size() ; ++k)
            {
                global_ordinal_type column_index = B_G_double->getColMap()->getGlobalElement(local_indices[k]);

                matrices->B_G->insertGlobalValues(global_index,
                    Teuchos::tuple<global_ordinal_type> (column_index),
                    Teuchos::tuple<scalar> ((scalar) values[k]));
            }

            /*
            Teuchos::ArrayView<global_ordinal_type> global_indices;
            Teuchos::ArrayView<double> values ;
            size_t numOfEntries;

            B_double->getGlobalRowCopy (global_index, global_indices, values, numOfEntries);

            for ( auto k = 0; k < numOfEntries ; ++k)
            {
                matrices->B->insertGlobalValues(global_index,
                    Teuchos::tuple<global_ordinal_type> (global_indices[k]),
                    Teuchos::tuple<scalar> ((scalar) values[k]));

                matrices->B_T->insertGlobalValues(global_indices[k],
                    Teuchos::tuple<global_ordinal_type> (global_index),
                    Teuchos::tuple<scalar> ((scalar) values[k]));
            }

            B_G_double->getGlobalRowCopy (global_index, global_indices, values, numOfEntries);

            for ( auto k = 0; k < numOfEntries ; ++k)
            {
                matrices->B_G->insertGlobalValues(global_index,
                    Teuchos::tuple<global_ordinal_type> (global_indices[k]),
                    Teuchos::tuple<scalar> ((scalar) values[k]));
            }
            */

            matrices->C->insertGlobalValues(global_index, Teuchos::tuple<global_ordinal_type> (global_index), Teuchos::tuple<scalar> ((scalar) 0.));
            matrices->Cb->insertGlobalValues(global_index, Teuchos::tuple<global_ordinal_type> (global_index), Teuchos::tuple<scalar> ((scalar) 1.));
        }
    }
    std::cout << " compute_B : e" << std::endl;

    matrices->B->fillComplete(map->mapDofs,map->mapLagrangeDofs);
    matrices->B_2->fillComplete(map->mapDofs,map->mapLagrangeDofs);
    matrices->B_G->fillComplete(map->mapDofs,map->mapLagrangeDofs);
    matrices->B_T->fillComplete(map->mapLagrangeDofs,map->mapDofs);
    matrices->C->fillComplete();
    matrices->Cb->fillComplete();

    std::cout << " compute_B : f" << std::endl;

    //std::cout << *(matrices->B) << std::endl;
    //std::cout << *(matrices->B_T) << std::endl;
};
