#ifndef WCONTACT_H
#define WCONTACT_H

#include "mrstlnos.h"
#include "wGroup.h"
#include "wPt.h"
#include <vector>
#include <string>
using namespace tbox;

namespace mrstlnos {

/**
 * @brief manage mechanical Dirichlet BC
 */

class MRSTLNOS_API Contact : public Group
{
public:
    std::string Contact_name;
    std::vector<Element *> slave_elems;
    std::vector<Element *> master_elems;

    std::vector<int> slave_boundary_nodes;
    std::vector<int> master_boundary_nodes;

    std::vector<int> slave_dirichlet_nodes;
    std::vector<int> master_dirichlet_nodes;

    std::vector<int> initially_closed_nodes={};

    bool initial_closed = true; 
    bool standard_shape_functions = true; 
    bool is_Signorini = false;
    bool no_Update = false;
    bool sticking = false;
    bool mesh_tying = false;
  
    int nb_Lagrange_multipliers = 1;
  
    double dg;
  
    double scale = 1.;

    std::vector<double> dg_vector;
  
  

    Pt slave_normal;
    Pt master_normal;

    Contact(Problem &pbl, int no, std::string _Contact_name, Pt _slave_normal, double dg = 0.);
    Contact(Problem &pbl, std::string const &name, std::string _Contact_name, Pt _slave_normal, double dg = 0.);

    Contact(Problem &pbl, int no, std::string _Contact_name, Pt _slave_normal, std::vector<double> _dg);
    Contact(Problem &pbl, std::string const &name, std::string _Contact_name, Pt _slave_normal, std::vector<double> _dg);    

    void setMaster(mrstlnos::Problem &pbl, int no, Pt _master_normal);
    void setMaster(mrstlnos::Problem &pbl, std::string const &name, Pt _master_normal);

    void setSlaveBoundary(std::string const &file_name, std::string const &name);
    void setMasterBoundary(std::string const &file_name, std::string const &name);

    void setSlaveDirichlet(std::string const &file_name, std::string const &name);
    void setMasterDirichlet(std::string const &file_name, std::string const &name);

    void setInitialyClosed()
    {
        initial_closed = true;
    };
    void setInitialyOpen()
    {
        initial_closed = false;
    };

    void setNoUpdate()
    {
        no_Update = true;
    };
    void setUpdate()
    {
        no_Update = false;
    };
  
    void setMeshTying()
    {
        mesh_tying = true;
        sticking = true;
        no_Update = true;
    };
  
    void setSticking()
    {
        sticking = true;
    };
    void setNoSticking()
    {
        sticking = false;
    };

    void setStandardShapeFunctions()
    {
        standard_shape_functions = true;
    };
    void setDualShapeFunctions()
    {
        standard_shape_functions = false;
    };

    void setSignoriniContact()
    {
        is_Signorini = true;
    };
    void setTwoBodyContact()
    {
        is_Signorini = false;
    };   
    
    void setInitiallyClosedNodes(std::vector<double> _initially_closed_nodes)
    {
        for (auto i=0; i<_initially_closed_nodes.size(); ++i)
            initially_closed_nodes.push_back(((int) _initially_closed_nodes[i] ));
    };       

    void setScale(double a)
    {
        scale = a;
    };

#ifndef SWIG
    virtual void write(std::ostream &out) const;
#endif
};

}

#endif //WCONTACT_H
