#include "wExample2.h"

#include "wTimer.h"
#include "wTimers.h"

// Tpetra
#include "Stokhos_Tpetra_MP_Vector.hpp"
#include "Stokhos_Tpetra_Utilities_MP_Vector.hpp"
#include "Tpetra_ConfigDefs.hpp"
#include "Tpetra_DefaultPlatform.hpp"
#include "Tpetra_Map.hpp"
#include "Tpetra_MultiVector.hpp"
#include "Tpetra_Vector.hpp"
#include "Tpetra_CrsGraph.hpp"
#include "Tpetra_CrsMatrix.hpp"
#include "Stokhos_Tpetra_CG.hpp"


#include "TpetraExt_MatrixMatrix.hpp"
#include "Tpetra_MatrixIO.hpp"
#include "Tpetra_DefaultPlatform.hpp"
#include "Tpetra_Vector.hpp"
#include "Tpetra_CrsMatrixMultiplyOp.hpp"
#include "Tpetra_Import.hpp"
#include "Tpetra_Export.hpp"
#include "Teuchos_DefaultComm.hpp"
#include "Teuchos_CommandLineProcessor.hpp"
#include "Teuchos_XMLParameterListHelpers.hpp"
#include "Teuchos_UnitTestHarness.hpp"
#include "MatrixMarket_Tpetra.hpp"
#include "Tpetra_RowMatrixTransposer.hpp"


#include <Teuchos_Array.hpp>
#include <Teuchos_GlobalMPISession.hpp>
#include <Teuchos_oblackholestream.hpp>
#include <Teuchos_ScalarTraits.hpp>
#include <Teuchos_RCP.hpp>
#include "Teuchos_ParameterList.hpp"

#include "Stokhos_Amesos2_MP_Vector.hpp"
#include "Amesos2_Factory.hpp"

#include "Stokhos_Ifpack2_MP_Vector.hpp"
#include "Ifpack2_Factory.hpp"

#include "Belos_Tpetra_MP_Vector.hpp"
#include "BelosLinearProblem.hpp"
#include "BelosPseudoBlockGmresSolMgr.hpp"
#include "BelosPseudoBlockCGSolMgr.hpp"
#include "BelosBlockGmresSolMgr.hpp"

#include "EnsembleTraits.h"
#include "Kokkos_InnerProductSpaceTraits_MP_Vector.hpp"

#include <math.h>

using namespace fwk;

using namespace mrstlnos;

typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

template<typename scalar> void writeLocalCrsMatrixToFile(std::string &name, size_t myRank, Teuchos::RCP<Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> > A)
{
    typedef EnsembleTraits<scalar> ET;
    std::ofstream outputfile;
    outputfile.open (name + "_" + std::to_string(myRank) + ".txt");

    typedef std::numeric_limits< double > dbl;
    outputfile.precision(dbl::max_digits10+2);

    const int ensemble_size = ET::size;

    auto line_size =  A->getRowMap()->getNodeNumElements ();

    for (local_ordinal_type local_index = 0; local_index < line_size; ++local_index)
    {
        global_ordinal_type global_index =   A->getRowMap()->getGlobalElement(local_index);

        Teuchos::ArrayView<const local_ordinal_type> local_indices ;
        Teuchos::ArrayView<const scalar> local_values ;
        A->getLocalRowView (local_index, local_indices, local_values);

        for ( auto k = 0; k < local_indices.size() ; ++k)
        {
           outputfile << global_index << "   " << A->getColMap()->getGlobalElement(local_indices[k]) << "   ";
           for (int sample=0; sample<ensemble_size; ++sample)
                outputfile << (double) ET::coeff(local_values[k], sample) << " ";
           outputfile << std::endl; 
        }

    }
    outputfile.close();

}

template<typename scalar> void writeVectorToFile(std::string &name, size_t myRank, Teuchos::RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > b)
{
    using Teuchos::Array;
    using Teuchos::ArrayView;
    using Teuchos::ArrayRCP;
    using Teuchos::arcp;
    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;
    
    std::ofstream outputfile;
    if (myRank ==0)
        outputfile.open (name);

    typedef std::numeric_limits< double > dbl;
    outputfile.precision(dbl::max_digits10+2);

    if (myRank ==0)
    {
        b->template sync<Kokkos::HostSpace> ();
        auto b_2d = b->template getLocalView<Kokkos::HostSpace> ();
        auto B = Kokkos::subview (b_2d, Kokkos::ALL (), 0);

        const int ensemble_size = Kokkos::dimension_scalar(B) == 0 ? 1 : Kokkos::dimension_scalar(B);
        typedef decltype(B) B_View;
        typename Kokkos::FlatArrayType<B_View>::type B_flat = B;

        auto line_size =  b->getMap()->getGlobalNumElements ();
        for (global_ordinal_type global_index = 0; global_index < line_size; ++global_index)
        {
            for(auto j=0; j<ensemble_size; ++j)
                outputfile << B_flat[ensemble_size*global_index+j] << " ";
            outputfile << std::endl;
        }
        outputfile.close();
    }
}

Example2::Example2()
{
}

template<typename scalar> void vCycle(int currentLvl, int maxLvl, int nu1, int nu2, Teuchos::RCP<Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> > A, Teuchos::RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > b, Teuchos::RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > x, std::string &name, size_t myRank, std::string &name2, int nx,Teuchos::RCP<const Teuchos::Comm<int> >  comm, bool verbose) 
{
    using Teuchos::Array;
    using Teuchos::ArrayView;
    using Teuchos::ArrayRCP;
    using Teuchos::arcp;
    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;
    using std::cerr;
    using std::cout;
    using std::endl;

    using Kokkos::subview;
    using Kokkos::ALL;
    
    typedef Tpetra::Map<> map_type;
    typedef Tpetra::Vector<>::scalar_type scalar_type;
    typedef Tpetra::Vector<>::mag_type magnitude_type;
    typedef Tpetra::MultiVector<scalar,local_ordinal_type,global_ordinal_type> multivector_type;
    typedef Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> crs_matrix_type;
    typedef Tpetra::CrsGraph<local_ordinal_type,global_ordinal_type> crs_graph_type;

    typedef Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> vector_type;
    typedef Tpetra::Export<> export_type;
    //typedef typename vector_type::mag_type magnitude_type;
    typedef Amesos2::Solver<crs_matrix_type,multivector_type> solver_type;
    using Teuchos::RCP;
    using Teuchos::rcp;
    std::ofstream outputfile;
    if (verbose)
    if (myRank ==0)
    {
        std::string tmp = name + "_lvl_" + std::to_string(currentLvl) + ".txt";
        outputfile.open (tmp);
    }
        
      
    int indexBase = 0;
    int nx_coarsed = (nx-1)/2;
    int N_coarsed = nx_coarsed*nx_coarsed;

    RCP<const map_type> map_coarsed_y_lines = rcp (new map_type (nx_coarsed, indexBase, comm));

    const size_t num_my_coarsed_y_lines = map_coarsed_y_lines->getNodeNumElements ();
    Teuchos::ArrayView<const global_ordinal_type> my_coarsed_y_lines = map_coarsed_y_lines->getNodeElementList ();    
    Array<global_ordinal_type> my_dofs (num_my_coarsed_y_lines*nx_coarsed);
    for (auto i = 0; i < num_my_coarsed_y_lines; ++i)
        for (auto j = 0; j < nx_coarsed; ++j)
        my_dofs[nx_coarsed*i+j] = nx_coarsed*my_coarsed_y_lines[i]+j;

    RCP<const map_type> map_coarsed_dofs = rcp (new map_type (N_coarsed, my_dofs, indexBase, comm));
    RCP<const map_type> map_dofs = A->getRowMap();
    // Prolongation graph
    RCP<crs_graph_type> graph_P (new crs_graph_type (map_dofs,map_coarsed_dofs, 0));

    for (auto i = 0; i < nx; ++i)
    for (auto j = 0; j < nx_coarsed; ++j)
    for (auto k = 0; k < nx; ++k)
    for (auto l = 0; l < nx_coarsed; ++l)
    {
        global_ordinal_type global_index_1 = i*nx+k;
        global_ordinal_type global_index_2 = j*nx_coarsed +l;
        if(i==j*2 || i==j*2+1 || i==j*2+2)
            if(k==l*2 || k==l*2+1 || k==l*2+2)
                graph_P->insertGlobalIndices(global_index_1, tuple<global_ordinal_type> (global_index_2));  
    }

    graph_P->fillComplete (map_coarsed_dofs,map_dofs);
    // Prolongation matrix
    RCP<crs_matrix_type> P (new crs_matrix_type (graph_P));
    for (auto i = 0; i < nx; ++i)
    for (auto j = 0; j < nx_coarsed; ++j)
    for (auto k = 0; k < nx; ++k)
    for (auto l = 0; l < nx_coarsed; ++l)
    {
        global_ordinal_type global_index_1 = i*nx+k;
        global_ordinal_type global_index_2 = j*nx_coarsed +l;
        if(i==j*2 || i==j*2+2)
        {
            if(k==l*2 || k==l*2+2)
                P->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 0.25));
            else if (k==l*2+1)
                P->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 0.5));       
        }
        else if(i==j*2+1)
        {
            if(k==l*2 || k==l*2+2)
                P->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 0.5));
            else if (k==l*2+1)
                P->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 1.));       
        }
    }
    
    P->fillComplete (map_coarsed_dofs,map_dofs);
    if (verbose)
    {
        std::string name_P = "prolongation_matrix_lvl_" + std::to_string(currentLvl);
        writeLocalCrsMatrixToFile<scalar>(name_P, myRank, P);
    }
    

    // Restriction graph
    RCP<crs_graph_type> graph_R (new crs_graph_type (map_coarsed_dofs,map_dofs, 0));

    for (auto i = 0; i < nx_coarsed; ++i)
    for (auto j = 0; j < nx; ++j)
    for (auto k = 0; k < nx_coarsed; ++k)
    for (auto l = 0; l < nx; ++l)
    {
        global_ordinal_type global_index_1 = i*nx_coarsed +k;
        global_ordinal_type global_index_2 = j*nx +l;
        if(j==i*2 || j==i*2+1 || j==i*2+2)
            if(l==k*2 || l==k*2+1 || l==k*2+2)
                graph_R->insertGlobalIndices(global_index_1, tuple<global_ordinal_type> (global_index_2));  
    }

    graph_R->fillComplete (map_dofs,map_coarsed_dofs);
    // Restrictionmatrix
    RCP<crs_matrix_type> R (new crs_matrix_type (graph_R));
    for (auto i = 0; i < nx_coarsed; ++i)
    for (auto j = 0; j < nx; ++j)
    for (auto k = 0; k < nx_coarsed; ++k)
    for (auto l = 0; l < nx; ++l)
    {
        global_ordinal_type global_index_1 = i*nx_coarsed+k;
        global_ordinal_type global_index_2 = j*nx +l;
        if(j==i*2 || j==i*2+2)
        {
            if(l==k*2 || l==k*2+2)
                R->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 0.25/4.));
            else if (l==k*2+1)
                R->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 0.5/4.));       
        }
        else if(j==i*2+1)
        {
            if(l==k*2 || l==k*2+2)
                R->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 0.5/4.));
            else if (l==k*2+1)
                R->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 1./4.));       
        }
    }
    
    R->fillComplete (map_dofs,map_coarsed_dofs);
    if (verbose)
    {
        std::string name_R = "restriction_matrix_lvl_" + std::to_string(currentLvl);
        writeLocalCrsMatrixToFile<scalar>(name_R, myRank, R);   
    }
    

    //Coarsed matrix
    RCP<crs_matrix_type> AP (new crs_matrix_type (map_dofs,map_coarsed_dofs,1));

    Tpetra::MatrixMatrix::Multiply(*A, false, *P, false, *AP, false); // AP = A*P
    AP->fillComplete (map_coarsed_dofs,map_dofs);
    
    RCP<crs_matrix_type> AH (new crs_matrix_type (map_coarsed_dofs,map_coarsed_dofs,1)); 
    Tpetra::MatrixMatrix::Multiply(*R, false, *AP, false, *AH, false); // AH = R*AP
    AH->fillComplete (map_coarsed_dofs,map_coarsed_dofs);
    //std::cout << *AH << std::endl;
    if (verbose)
    {
        std::string name_AH = "coarsed_matrix_lvl_" + std::to_string(currentLvl);
        writeLocalCrsMatrixToFile<scalar>(name_AH, myRank, AH); 
    }
    const scalar one  = (scalar) 1;
    const scalar zero = (scalar) 0;

    vector_type z (A->getRangeMap ());
    vector_type resid (A->getRangeMap ());

    RCP<vector_type> residH (new vector_type(AH->getRangeMap ()));
    RCP<vector_type> deltaH (new vector_type(AH->getRangeMap ()),true);

    vector_type delta (A->getRangeMap ());
    if (currentLvl ==0)
        x->scale (zero, z); 

    // Presmoothing
    double omega = 1./6.;
    if (verbose)
    if (myRank ==0)
        {
            std::string name3 = name2 + "_preS_0_"+ std::to_string(nu1) +"_lvl_" + std::to_string(currentLvl)+".txt"; 
            writeVectorToFile<scalar>(name3, myRank, x);
        }

    for (int iter = 1; iter <= nu1; ++iter) 
    {
        A->apply(*x, z);                            // z := A * x
        x->update(-omega ,z, omega, *b, one);       // x := x - omega*z + omega*b
        if (verbose)
        if (myRank ==0)
        {
            std::string name3 = name2 + "_preS_" + std::to_string(iter) + "_"+ std::to_string(nu1) +"_lvl_" + std::to_string(currentLvl)+".txt"; 
            writeVectorToFile<scalar>(name3, myRank, x);
        }
    }

    // Coarse correction
    A->apply(*x, z);                                // z := A * x
    resid.update (one, *b, -one, z, zero);          // resid := b - z
    R->apply(resid, *residH);                      // residH := R*resid

    if (verbose)
    if (myRank ==0)
    {
        std::string name3 = name2 + "_residH_lvl_" + std::to_string(currentLvl)+".txt"; 
        writeVectorToFile<scalar>(name3, myRank, residH);
    }

    if (currentLvl+1 == maxLvl)
    {
        RCP<solver_type> solverA;
        solverA = Amesos2::create<crs_matrix_type,multivector_type>("KLU2", AH, deltaH, residH);
        solverA->solve();
    }
    else
    {
        int nxH = (nx-1)/2;
        vCycle(currentLvl+1, maxLvl, nu1, nu2, AH, residH, deltaH, name, myRank, name2, nxH, comm, verbose); 
    }
    
    if (verbose)
    if (myRank ==0)
    {
        std::string name3 = name2 + "_deltaH_lvl_" + std::to_string(currentLvl)+".txt"; 
        writeVectorToFile<scalar>(name3, myRank, deltaH);
    }

    P->apply(*deltaH, delta);         // delta := P*deltaH
    x->update(one ,delta, one);       // x := x + delta

    // Postsmoothing
    if (verbose)
    if (myRank ==0)
    {
        std::string name3 = name2 + "_postS_0_"+ std::to_string(nu2) +"_lvl_" + std::to_string(currentLvl)+".txt"; 
        writeVectorToFile<scalar>(name3, myRank, x);
    }

    for (int iter = 1; iter <= nu2; ++iter) 
    {
        A->apply(*x, z);                            // z := A * x
        x->update(-omega ,z, omega, *b, one);       // x := x - omega*z + omega*b
        if (verbose)
        if (myRank ==0)
        {
            std::string name3 = name2 + "_postS_" + std::to_string(iter) + "_"+ std::to_string(nu2) +"_lvl_" + std::to_string(currentLvl)+".txt"; 
            writeVectorToFile<scalar>(name3, myRank, x);
        }
    }

    if (verbose)
        outputfile.close();
}

template<typename scalar> void twoGridCycle(int nu1, int nu2, Teuchos::RCP<Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> > A, Teuchos::RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > b, Teuchos::RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > x, std::string &name, size_t myRank, std::string &name2, int nx,Teuchos::RCP<const Teuchos::Comm<int> >  comm, bool verbose) 
{
    using Teuchos::Array;
    using Teuchos::ArrayView;
    using Teuchos::ArrayRCP;
    using Teuchos::arcp;
    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;
    using std::cerr;
    using std::cout;
    using std::endl;

    using Kokkos::subview;
    using Kokkos::ALL;
    
    typedef Tpetra::Map<> map_type;
    typedef Tpetra::Vector<>::scalar_type scalar_type;
    typedef Tpetra::Vector<>::mag_type magnitude_type;
    typedef Tpetra::MultiVector<scalar,local_ordinal_type,global_ordinal_type> multivector_type;
    typedef Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> crs_matrix_type;
    typedef Tpetra::CrsGraph<local_ordinal_type,global_ordinal_type> crs_graph_type;

    typedef Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> vector_type;
    typedef Tpetra::Export<> export_type;
    //typedef typename vector_type::mag_type magnitude_type;
    typedef Amesos2::Solver<crs_matrix_type,multivector_type> solver_type;
    using Teuchos::RCP;
    using Teuchos::rcp;
    std::ofstream outputfile;
    if (verbose)
    if (myRank ==0)
        outputfile.open (name);
      
    int indexBase = 0;
    int nx_coarsed = (nx-1)/2;
    int N_coarsed = nx_coarsed*nx_coarsed;

    RCP<const map_type> map_coarsed_y_lines = rcp (new map_type (nx_coarsed, indexBase, comm));

    const size_t num_my_coarsed_y_lines = map_coarsed_y_lines->getNodeNumElements ();
    Teuchos::ArrayView<const global_ordinal_type> my_coarsed_y_lines = map_coarsed_y_lines->getNodeElementList ();    
    Array<global_ordinal_type> my_dofs (num_my_coarsed_y_lines*nx_coarsed);
    for (auto i = 0; i < num_my_coarsed_y_lines; ++i)
        for (auto j = 0; j < nx_coarsed; ++j)
        my_dofs[nx_coarsed*i+j] = nx_coarsed*my_coarsed_y_lines[i]+j;

    RCP<const map_type> map_coarsed_dofs = rcp (new map_type (N_coarsed, my_dofs, indexBase, comm));
    RCP<const map_type> map_dofs = A->getRowMap();
    // Prolongation graph
    RCP<crs_graph_type> graph_P (new crs_graph_type (map_dofs,map_coarsed_dofs, 0));

    for (auto i = 0; i < nx; ++i)
    for (auto j = 0; j < nx_coarsed; ++j)
    for (auto k = 0; k < nx; ++k)
    for (auto l = 0; l < nx_coarsed; ++l)
    {
        global_ordinal_type global_index_1 = i*nx+k;
        global_ordinal_type global_index_2 = j*nx_coarsed +l;
        if(i==j*2 || i==j*2+1 || i==j*2+2)
            if(k==l*2 || k==l*2+1 || k==l*2+2)
                graph_P->insertGlobalIndices(global_index_1, tuple<global_ordinal_type> (global_index_2));  
    }

    graph_P->fillComplete (map_coarsed_dofs,map_dofs);
    // Prolongation matrix
    RCP<crs_matrix_type> P (new crs_matrix_type (graph_P));
    for (auto i = 0; i < nx; ++i)
    for (auto j = 0; j < nx_coarsed; ++j)
    for (auto k = 0; k < nx; ++k)
    for (auto l = 0; l < nx_coarsed; ++l)
    {
        global_ordinal_type global_index_1 = i*nx+k;
        global_ordinal_type global_index_2 = j*nx_coarsed +l;
        if(i==j*2 || i==j*2+2)
        {
            if(k==l*2 || k==l*2+2)
                P->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 0.25));
            else if (k==l*2+1)
                P->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 0.5));       
        }
        else if(i==j*2+1)
        {
            if(k==l*2 || k==l*2+2)
                P->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 0.5));
            else if (k==l*2+1)
                P->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 1.));       
        }
    }
    
    P->fillComplete (map_coarsed_dofs,map_dofs);

    if (verbose)
    {
        std::string name_P = "prolongation_matrix";
        writeLocalCrsMatrixToFile<scalar>(name_P, myRank, P);
    }
    

    // Restriction graph
    RCP<crs_graph_type> graph_R (new crs_graph_type (map_coarsed_dofs,map_dofs, 0));

    for (auto i = 0; i < nx_coarsed; ++i)
    for (auto j = 0; j < nx; ++j)
    for (auto k = 0; k < nx_coarsed; ++k)
    for (auto l = 0; l < nx; ++l)
    {
        global_ordinal_type global_index_1 = i*nx_coarsed +k;
        global_ordinal_type global_index_2 = j*nx +l;
        if(j==i*2 || j==i*2+1 || j==i*2+2)
            if(l==k*2 || l==k*2+1 || l==k*2+2)
                graph_R->insertGlobalIndices(global_index_1, tuple<global_ordinal_type> (global_index_2));  
    }

    graph_R->fillComplete (map_dofs,map_coarsed_dofs);
    // Restrictionmatrix
    RCP<crs_matrix_type> R (new crs_matrix_type (graph_R));
    for (auto i = 0; i < nx_coarsed; ++i)
    for (auto j = 0; j < nx; ++j)
    for (auto k = 0; k < nx_coarsed; ++k)
    for (auto l = 0; l < nx; ++l)
    {
        global_ordinal_type global_index_1 = i*nx_coarsed+k;
        global_ordinal_type global_index_2 = j*nx +l;
        if(j==i*2 || j==i*2+2)
        {
            if(l==k*2 || l==k*2+2)
                R->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 0.25/4.));
            else if (l==k*2+1)
                R->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 0.5/4.));       
        }
        else if(j==i*2+1)
        {
            if(l==k*2 || l==k*2+2)
                R->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 0.5/4.));
            else if (l==k*2+1)
                R->replaceGlobalValues(global_index_1, tuple<global_ordinal_type> (global_index_2), tuple<scalar> ((scalar) 1./4.));       
        }
    }
    
    R->fillComplete (map_dofs,map_coarsed_dofs);

    if (verbose)
    {
        std::string name_R = "restriction_matrix";
        writeLocalCrsMatrixToFile<scalar>(name_R, myRank, R); 
    }
      

    //Coarsed matrix
    RCP<crs_matrix_type> AP (new crs_matrix_type (map_dofs,map_coarsed_dofs,1));

    Tpetra::MatrixMatrix::Multiply(*A, false, *P, false, *AP, false); // AP = A*P
    AP->fillComplete (map_coarsed_dofs,map_dofs);

    RCP<crs_matrix_type> AH (new crs_matrix_type (map_coarsed_dofs,map_coarsed_dofs,1)); 
    Tpetra::MatrixMatrix::Multiply(*R, false, *AP, false, *AH, false); // AH = R*AP
    AH->fillComplete (map_coarsed_dofs,map_coarsed_dofs);
    //std::cout << *AH << std::endl;
    if (verbose)
    {
        std::string name_AH = "coarsed_matrix";
        writeLocalCrsMatrixToFile<scalar>(name_AH, myRank, AH); 
    }
    

    const scalar one  = (scalar) 1;
    const scalar zero = (scalar) 0;

    vector_type z (A->getRangeMap ());
    vector_type resid (A->getRangeMap ());

    RCP<vector_type> residH (new vector_type(AH->getRangeMap ()));
    RCP<vector_type> deltaH (new vector_type(AH->getRangeMap ()));

    vector_type delta (A->getRangeMap ());
    x->scale (zero, z); 

    double omega = 1./6.;
    if (verbose)
    if (myRank ==0)
        {
            std::string name3 = name2 + "_preS_0_"+ std::to_string(nu1) +".txt"; 
            writeVectorToFile<scalar>(name3, myRank, x);
        }

    for (int iter = 1; iter <= nu1; ++iter) 
    {
        A->apply(*x, z);                            // z := A * x
        x->update(-omega ,z, omega, *b, one);       // x := x - omega*z + omega*b
        if (verbose)
        if (myRank ==0)
        {
            std::string name3 = name2 + "_preS_" + std::to_string(iter) + "_"+ std::to_string(nu1) +".txt"; 
            writeVectorToFile<scalar>(name3, myRank, x);
        }
    }
    A->apply(*x, z);                                // z := A * x
    resid.update (one, *b, -one, z, zero);          // resid := b - z
    R->apply(resid, *residH);                      // residH := R*resid

    if (verbose)
    if (myRank ==0)
    {
        std::string name3 = name2 + "_residH.txt"; 
        writeVectorToFile<scalar>(name3, myRank, residH);
    }

    RCP<solver_type> solverA;
    solverA = Amesos2::create<crs_matrix_type,multivector_type>("KLU2", AH, deltaH, residH);
    solverA->solve();

    if (verbose)
    if (myRank ==0)
    {
        std::string name3 = name2 + "_deltaH.txt"; 
        writeVectorToFile<scalar>(name3, myRank, deltaH);
    }

    P->apply(*deltaH, delta);                      // delta := P*deltaH
    x->update(one ,delta, one);       // x := x + delta

    if (verbose)
    if (myRank ==0)
    {
        std::string name3 = name2 + "_postS_0_"+ std::to_string(nu2) +".txt"; 
        writeVectorToFile<scalar>(name3, myRank, x);
    }

    for (int iter = 1; iter <= nu2; ++iter) 
    {
        A->apply(*x, z);                            // z := A * x
        x->update(-omega ,z, omega, *b, one);       // x := x - omega*z + omega*b
        if (verbose)
        if (myRank ==0)
        {
            std::string name3 = name2 + "_postS_" + std::to_string(iter) + "_"+ std::to_string(nu2) +".txt"; 
            writeVectorToFile<scalar>(name3, myRank, x);
        }
    }

    outputfile.close();
}


template<typename scalar> scalar dotProductEnsemble(Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type>  x1, Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type>  x2)
{
    typedef Kokkos::Details::InnerProductSpaceTraits<scalar> IPT;
    using Teuchos::ArrayRCP;
    scalar solution = (scalar) 0;

    ArrayRCP<const scalar> x1_data= x1.getData();
    ArrayRCP<const scalar> x2_data= x2.getData();

    auto n = x1_data.size ();
    for (auto i = 0; i<n; ++i)
        solution = solution + IPT::dot(x1_data[i],x2_data[i]);

    return solution;
}

template<typename scalar> scalar normEnsemble(Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type>  x)
{
    typedef Kokkos::Details::InnerProductSpaceTraits<scalar> IPT;
    using Teuchos::ArrayRCP;
    scalar solution = (scalar) 0;

    ArrayRCP<const scalar> x_data= x.getData();

    auto n = x_data.size ();
    for (auto i = 0; i<n; ++i)
        solution = solution + IPT::dot(x_data[i],x_data[i]);

    solution = sqrt(solution);

    return solution;
}

template<typename scalar> double dotProduct(Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type>  x1, Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type>  x2)
{
    scalar tmp = dotProductEnsemble<scalar> (x1,x2);

    double solution = 0.;

    typedef EnsembleTraits<scalar> ET;
    const int ensemble_size = ET::size;
    std::cout << "ensemble_size " << ensemble_size << std::endl;
    for (int sample=0; sample<ensemble_size; ++sample)
        solution += ET::coeff(tmp,sample);
    
    std::cout << tmp << " " << solution << std::endl;
    return solution;
}

template<typename scalar> double norm(Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type>  x)
{
    scalar tmp1 = normEnsemble<scalar> (x);
    scalar tmp = pow(tmp1,2);

    double solution = 0.;

    typedef EnsembleTraits<scalar> ET;
    const int ensemble_size = ET::size;
    std::cout << "ensemble_size " << ensemble_size << std::endl; 
    for (int sample=0; sample<ensemble_size; ++sample)
        solution += ET::coeff(tmp,sample);

    solution = sqrt(solution);
    std::cout << tmp1 << " " << solution << std::endl;
    return solution;
}

template<typename scalar> int conjugateGradient(int niters, Teuchos::RCP<Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> > A, Teuchos::RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > b, Teuchos::RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > x, std::string &name, size_t myRank, std::string &name2, bool verbose) 
{
    typedef Tpetra::Map<> map_type;
    typedef Tpetra::Vector<>::scalar_type scalar_type;
    typedef Tpetra::Vector<>::mag_type magnitude_type;
    typedef Tpetra::MultiVector<scalar,local_ordinal_type,global_ordinal_type> multivector_type;
    typedef Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> crs_matrix_type;
    typedef Tpetra::CrsGraph<local_ordinal_type,global_ordinal_type> crs_graph_type;

    typedef Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> vector_type;
    typedef Tpetra::Export<> export_type;
    //typedef typename vector_type::mag_type magnitude_type;
    using Teuchos::RCP;
    using Teuchos::rcp;
    std::ofstream outputfile;
    typedef EnsembleTraits<scalar> ET;

    if (myRank ==0)
        outputfile.open (name);
    
    const scalar one  = (scalar) 1;
    const scalar zero = (scalar) 0;
    vector_type q (A->getDomainMap ());
    vector_type z (A->getRangeMap ());    
    vector_type resid_k (A->getRangeMap ());
    vector_type p_k (A->getRangeMap ());

    magnitude_type tolerance = (magnitude_type) 0.00000000001; //0.00000001;

    scalar alpha_k;
    scalar beta_k;
    scalar normz;
    scalar residual_0;
    scalar residual_k;
    scalar residual_kp1;
    magnitude_type alpha_k_total;
    magnitude_type beta_k_total;
    magnitude_type residual_0_total;
    magnitude_type residual_k_total;
    magnitude_type residual_kp1_total;

    const int reportFrequency = 1;
    const int outputFrequency = 1;

    x->scale (zero, z); //x->randomize();
    A->apply(*x, z);                            // z := A * x
    resid_k.update (one, *b, -one, z, zero);    // resid_k := b - z
    p_k.scale (one, resid_k);                   // p_k := resid_k
    residual_k = normEnsemble<scalar> (resid_k);
    residual_0 = residual_k;
    residual_k_total = norm<scalar> (resid_k);
    residual_0_total = residual_k_total;

    std::string name3 = name2 + "_0.txt"; 
    writeVectorToFile<scalar>(name3, myRank, x);
    
    if (myRank ==0)
        outputfile << "Initial residual norm: " << residual_0  << " "  << residual_0_total << endl;  

    for (int iter = 1; iter <= niters; ++iter) 
    {
        A->apply(p_k, z);                            // z := A * p_k
        alpha_k = pow(residual_k,2)/dotProductEnsemble<scalar> (p_k,z);
        alpha_k_total = pow(residual_k_total,2)/dotProduct<scalar> (p_k,z);
        //std::cout << alpha_k << std::endl << alpha_k_total << std::endl;
        x->update(alpha_k_total ,p_k,one);                 // x := x + alpha_k*p_k
        resid_k.update(-alpha_k_total,z,one);
        residual_kp1 = normEnsemble<scalar> (resid_k);
        residual_kp1_total = norm<scalar> (resid_k);
        if (iter % reportFrequency == 0 || iter + 1 == niters) 
        {
            if (myRank ==0)
                outputfile << "Iteration " << iter << ": " << residual_kp1 << " " << residual_kp1/residual_0  << " " <<  residual_kp1_total << " "  << residual_kp1_total/residual_0_total << endl;    
        }
        if (iter % outputFrequency == 0 || iter + 1 == niters) 
        {
            if (verbose)
            if (myRank ==0)
            {
                std::string name3 = name2 + "_" + std::to_string(iter) + ".txt"; 
                writeVectorToFile<scalar>(name3, myRank, x);
            }
                
        }
        if (residual_k_total/residual_0_total < tolerance) 
        {
            if (myRank ==0)
                outputfile << "Converged after " << iter << " iterations" << endl;
            niters = iter;
            break;
        }
        else if (iter+1 == niters) 
        {
            if (myRank ==0)
                outputfile << "Failed to converge after " << niters << " iterations" << endl;
            break;
        }

        beta_k = pow(residual_kp1,2)/pow(residual_k,2);    
        beta_k_total = pow(residual_kp1_total,2)/pow(residual_k_total,2); 
        //std::cout << beta_k << std::endl << beta_k_total << std::endl;
        p_k.update(one,resid_k,beta_k_total);

        residual_k = residual_kp1;
        residual_k_total = residual_kp1_total;
    }
  outputfile.close();
  return niters;
}


template<typename scalar> int conjugateGradientV2(int niters, Teuchos::RCP<Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> > A, Teuchos::RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > b, Teuchos::RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > x, std::string &name, size_t myRank, std::string &name2, bool verbose) 
{
    typedef Tpetra::Map<> map_type;
    typedef Tpetra::Vector<>::scalar_type scalar_type;
    typedef Tpetra::Vector<>::mag_type magnitude_type;
    typedef Tpetra::MultiVector<scalar,local_ordinal_type,global_ordinal_type> multivector_type;
    typedef Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> crs_matrix_type;
    typedef Tpetra::CrsGraph<local_ordinal_type,global_ordinal_type> crs_graph_type;

    typedef Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> vector_type;
    typedef Tpetra::Export<> export_type;
    //typedef typename vector_type::mag_type magnitude_type;
    using Teuchos::RCP;
    using Teuchos::rcp;
    std::ofstream outputfile;
    typedef EnsembleTraits<scalar> ET;

    if (myRank ==0)
        outputfile.open (name);
    
    const scalar one  = (scalar) 1;
    const scalar zero = (scalar) 0;
    vector_type q (A->getDomainMap ());
    vector_type z (A->getRangeMap ());    
    vector_type resid_k (A->getRangeMap ());
    vector_type p_k (A->getRangeMap ());

    magnitude_type tolerance = (magnitude_type) 0.00000000001; //0.00000001;

    scalar alpha_k;
    scalar beta_k;
    scalar normz;
    scalar residual_0;
    scalar residual_k;
    scalar residual_kp1;
    magnitude_type alpha_k_total;
    magnitude_type beta_k_total;
    magnitude_type residual_0_total;
    magnitude_type residual_k_total;
    magnitude_type residual_kp1_total;

    const int reportFrequency = 1;
    const int outputFrequency = 1;

    x->scale (zero, z); //x->randomize();
    A->apply(*x, z);                            // z := A * x
    resid_k.update (one, *b, -one, z, zero);    // resid_k := b - z
    p_k.scale (one, resid_k);                   // p_k := resid_k
    residual_k = normEnsemble<scalar> (resid_k);
    residual_0 = residual_k;
    residual_k_total = norm<scalar> (resid_k);
    residual_0_total = residual_k_total;

    std::string name3 = name2 + "_0.txt"; 
    writeVectorToFile<scalar>(name3, myRank, x);
    
    if (myRank ==0)
        outputfile << "Initial residual norm: " << residual_0  << " "  << residual_0_total << endl;  

    for (int iter = 1; iter <= niters; ++iter) 
    {
        A->apply(p_k, z);                            // z := A * p_k
        alpha_k = pow(residual_k,2)/dotProductEnsemble<scalar> (p_k,z);
        alpha_k_total = pow(residual_k_total,2)/dotProduct<scalar> (p_k,z);
        //std::cout << alpha_k << std::endl << alpha_k_total << std::endl;
        x->update(alpha_k ,p_k,one);                 // x := x + alpha_k*p_k
        resid_k.update(-alpha_k,z,one);
        residual_kp1 = normEnsemble<scalar> (resid_k);
        residual_kp1_total = norm<scalar> (resid_k);
        if (iter % reportFrequency == 0 || iter + 1 == niters) 
        {
            if (myRank ==0)
                outputfile << "Iteration " << iter << ": " << residual_kp1 << " " << residual_kp1/residual_0  << " " <<  residual_kp1_total << " "  << residual_kp1_total/residual_0_total << endl;    
        }
        if (iter % outputFrequency == 0 || iter + 1 == niters) 
        {
            if (verbose)
            if (myRank ==0)
            {
                std::string name3 = name2 + "_" + std::to_string(iter) + ".txt"; 
                writeVectorToFile<scalar>(name3, myRank, x);
            }
                
        }
        if (residual_k_total/residual_0_total < tolerance) 
        {
            if (myRank ==0)
                outputfile << "Converged after " << iter << " iterations" << endl;
            niters = iter;
            break;
        }
        else if (iter+1 == niters) 
        {
            if (myRank ==0)
                outputfile << "Failed to converge after " << niters << " iterations" << endl;
            break;
        }

        beta_k = pow(residual_kp1,2)/pow(residual_k,2);    
        beta_k_total = pow(residual_kp1_total,2)/pow(residual_k_total,2); 
        //std::cout << beta_k << std::endl << beta_k_total << std::endl;
        p_k.update(one,resid_k,beta_k);

        residual_k = residual_kp1;
        residual_k_total = residual_kp1_total;
    }
  outputfile.close();
  return niters;
}

template<typename scalar> int preconditionedConjugateGradient(int nu1, int nu2, int nx, int niters, Teuchos::RCP<Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> > A, Teuchos::RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > b, Teuchos::RCP<Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> > x, std::string &name, size_t myRank, std::string &name2) 
{
    typedef Tpetra::Map<> map_type;
    typedef Tpetra::Vector<>::scalar_type scalar_type;
    typedef Tpetra::Vector<>::mag_type magnitude_type;
    typedef Tpetra::MultiVector<scalar,local_ordinal_type,global_ordinal_type> multivector_type;
    typedef Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> crs_matrix_type;
    typedef Tpetra::CrsGraph<local_ordinal_type,global_ordinal_type> crs_graph_type;

    typedef Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> vector_type;
    typedef Tpetra::Export<> export_type;
    //typedef typename vector_type::mag_type magnitude_type;
    using Teuchos::RCP;
    using Teuchos::rcp;
    std::ofstream outputfile;
    if (myRank ==0)
        outputfile.open (name);

       

    const scalar one  = (scalar) 1;
    const scalar zero = (scalar) 0;
    vector_type q (A->getDomainMap ());
    vector_type z (A->getRangeMap ());    
    vector_type resid_k (A->getRangeMap ());
    vector_type p_k (A->getRangeMap ());

    magnitude_type tolerance = (magnitude_type) 0.00000000001; //0.00000001;

    magnitude_type alpha_k;
    magnitude_type beta_k;
    magnitude_type normz;
    magnitude_type residual_0;
    magnitude_type residual_k;
    magnitude_type residual_kp1;

    const int reportFrequency = 1;
    const int outputFrequency = 1;

    x->scale (zero, z); //x->randomize();
    A->apply(*x, z);                            // z := A * x
    resid_k.update (one, *b, -one, z, zero);    // resid_k := b - z
    p_k.scale (one, resid_k);                   // p_k := resid_k
    residual_k = resid_k.norm2();
    residual_0 = residual_k;

    std::string name3 = name2 + "_0.txt"; 
    writeVectorToFile<scalar>(name3, myRank, x);
    // Do the power method, until the method has converged or the
    // maximum iteration count has been reached.
    for (int iter = 1; iter <= niters; ++iter) 
    {
        A->apply(p_k, z);                            // z := A * p_k
        alpha_k = residual_k*residual_k/(p_k.dot (z));          // alpha_k := resid_k^T * resid_k /(p_k^T*A*p_k)
        x->update(alpha_k ,p_k,one);                 // x := x + alpha_k*p_k
        resid_k.update(-alpha_k,z,one);
        residual_kp1 = resid_k.norm2();
        // Compute and report the residual norm every reportFrequency
        // iterations, or if we've reached the maximum iteration count.
        if (iter % reportFrequency == 0 || iter + 1 == niters) 
        {
            if (myRank ==0)
                outputfile << "Iteration " << iter << ": " << residual_k/residual_0  << " "  << residual_kp1/residual_0 << endl;    
        }
        if (iter % outputFrequency == 0 || iter + 1 == niters) 
        {
            if (myRank ==0)
            {
                std::string name3 = name2 + "_" + std::to_string(iter) + ".txt"; 
                writeVectorToFile<scalar>(name3, myRank, x);
            }
                
        }
        if (residual_kp1/residual_0 < tolerance) 
        {
            if (myRank ==0)
                outputfile << "Converged after " << iter << " iterations" << endl;
            niters = iter;
            break;
        }
        else if (iter+1 == niters) 
        {
            if (myRank ==0)
                outputfile << "Failed to converge after " << niters << " iterations" << endl;
            break;
        }
        beta_k = residual_kp1*residual_kp1/(residual_k*residual_k);
        p_k.update(one,resid_k,beta_k);

        residual_k = residual_kp1;
    }
  outputfile.close();
  return niters;
}

template<typename scalar> int tstart(int nx, int ny, int ensemble_size, MPI_Comm  _comm)
{
  using Teuchos::Array;
  using Teuchos::ArrayView;
  using Teuchos::ArrayRCP;
  using Teuchos::arcp;
  using Teuchos::RCP;
  using Teuchos::rcp;
  using Teuchos::tuple;
  using std::cerr;
  using std::cout;
  using std::endl;

  using Kokkos::subview;
  using Kokkos::ALL;

  typedef EnsembleTraits<scalar> ET;
  const int real_ensemble_size = ET::size;

  typedef Tpetra::Map<> map_type;
  typedef Tpetra::Vector<>::scalar_type scalar_type;
  typedef Tpetra::Vector<>::mag_type magnitude_type;
  typedef Tpetra::MultiVector<scalar,local_ordinal_type,global_ordinal_type> multivector_type;
  typedef Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> crs_matrix_type;
  typedef Tpetra::CrsGraph<local_ordinal_type,global_ordinal_type> crs_graph_type;

  typedef Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> vector_type;
  typedef Tpetra::Export<> export_type;
  typedef Amesos2::Solver<crs_matrix_type,multivector_type> solver_type;
  //typedef Kokkos::CrsMatrix<scalar,local_ordinal_type,Tpetra::CrsMatrix<>::execution_space> local_matrix_type;
  typedef KokkosSparse::CrsMatrix<scalar,local_ordinal_type,Tpetra::CrsMatrix<>::execution_space> local_matrix_type;
    

  // MPI comm
  RCP<const Teuchos::Comm<int> > comm = rcp (new Teuchos::MpiComm<int> (_comm));
  const size_t myRank = comm->getRank();

  // Make an output stream (for verbose output) that only prints on
  // Proc 0 of the communicator.
  Teuchos::oblackholestream blackHole;
  std::ostream& out = (myRank == 0) ? std::cout : blackHole;

  // Initiate Kokkos with the given number  of threads
  Kokkos::InitArguments args;
  args.num_threads = 1;
  Kokkos::initialize (args);
  {
    int numiters = ny;
    //
    ny = nx;
    double L = 1;
    double h = L/(nx+1);
    //
    int N = nx*ny;

    Timers tms;
    tms["total"].start();
    
    int indexBase = 0;
    tms["assembling"].start();
    // Map
    RCP<const map_type> map_y_lines = rcp (new map_type (ny, indexBase, comm));
    const size_t num_my_y_lines = map_y_lines->getNodeNumElements ();
    Teuchos::ArrayView<const global_ordinal_type> my_y_lines = map_y_lines->getNodeElementList ();

    std::vector<global_ordinal_type> my_nodes ={};
    Array<global_ordinal_type> my_dofs (num_my_y_lines*nx);
    for (auto i = 0; i < num_my_y_lines; ++i)
        for (auto j = 0; j < nx; ++j)
        my_dofs[nx*i+j] = nx*my_y_lines[i]+j;

    RCP<const map_type> map_dofs = rcp (new map_type (N, my_dofs, indexBase, comm));

    const size_t num_lcl_indices = (myRank == 0) ? N : 0;
    RCP<const Tpetra::Map<>> map_out = rcp (new Tpetra::Map<> (N, num_lcl_indices, indexBase, comm));

    // Graph
    RCP<crs_graph_type> graph_A (new crs_graph_type (map_dofs, 0));

    for (auto i = 0; i < num_my_y_lines; ++i)
    for (auto j = 0; j < nx; ++j)
    {
        global_ordinal_type global_index = map_dofs->getGlobalElement(nx*i+j);
        graph_A->insertGlobalIndices(global_index, tuple<global_ordinal_type> (global_index));
        if(global_index-nx >= 0 )
            graph_A->insertGlobalIndices(global_index, tuple<global_ordinal_type> (global_index-nx));
        if(global_index+nx < N )
            graph_A->insertGlobalIndices(global_index, tuple<global_ordinal_type> (global_index+nx));
        if(global_index-1 >= 0 && j != 0)
            graph_A->insertGlobalIndices(global_index, tuple<global_ordinal_type> (global_index-1));
        if(global_index+1 < N && j != nx-1)
            graph_A->insertGlobalIndices(global_index, tuple<global_ordinal_type> (global_index+1));    
    }

    graph_A->fillComplete ();

    // Matrix
    RCP<crs_matrix_type> A (new crs_matrix_type (graph_A));
    for (auto i = 0; i < num_my_y_lines; ++i)
    for (auto j = 0; j < nx; ++j)
    {
        global_ordinal_type global_index = map_dofs->getGlobalElement(nx*i+j);
        A->replaceGlobalValues(global_index, tuple<global_ordinal_type> (global_index), tuple<scalar> ((scalar) 4));
        if(global_index-nx >= 0 )
            A->replaceGlobalValues(global_index, tuple<global_ordinal_type> (global_index-nx), tuple<scalar> ((scalar) -1));
        if(global_index+nx < N )
            A->replaceGlobalValues(global_index, tuple<global_ordinal_type> (global_index+nx), tuple<scalar> ((scalar) -1));
        if(global_index-1 >= 0 && j != 0)
            A->replaceGlobalValues(global_index, tuple<global_ordinal_type> (global_index-1), tuple<scalar> ((scalar) -1));
        if(global_index+1 < N && j != nx-1)
            A->replaceGlobalValues(global_index, tuple<global_ordinal_type> (global_index+1), tuple<scalar> ((scalar) -1));    
    }
    A->fillComplete ();

    // b
    int x_mode = 3;
    int y_mode = 2;

    RCP<vector_type> b (new vector_type(map_dofs, true));
    RCP<vector_type> b_out (new vector_type(map_out, true));
    for (auto i = 0; i < num_my_y_lines; ++i)
    for (auto j = 0; j < nx; ++j)
    {
        scalar value;
        if (real_ensemble_size >= 1)
            for (int sample=0; sample<ensemble_size; ++sample)
            {
                ET::coeff(value, sample) = (sin((i+1)*(y_mode+sample)*M_PI/(ny+1))* sin((j+1)*(x_mode+sample)*M_PI/(nx+1)))*pow(h,2);
                //ET::coeff(value, sample) = 1./(h*h);
            }
        else
            value = sin((i+1)*y_mode*M_PI/(ny+1))* sin((j+1)*x_mode*M_PI/(nx+1))/(h*h);
        global_ordinal_type global_index = map_dofs->getGlobalElement(nx*i+j);
        b->replaceGlobalValue(global_index, value); 
    }    
    RCP<vector_type> x (new vector_type(map_dofs, true));
    RCP<vector_type> x_out (new vector_type(map_out, true));
    tms["assembling"].stop();

    tms["KLU2"].start();
    RCP<solver_type> solverA;
    solverA = Amesos2::create<crs_matrix_type,multivector_type>("KLU2", A, x, b);
    solverA->solve();
    tms["KLU2"].stop();

    scalar zero = (scalar) 0;
    scalar oneInf = (scalar) 1;
    ET::coeff(oneInf,0) = (double) 1./0.;

    out << "zero: " << zero << std::endl;
    out << "oneInf: " << oneInf << std::endl;
    out << "zero*oneInf: " << zero*oneInf << std::endl;


    //scalar value_tmp = dotProduct(x,x);
    //out << endl << value_tmp << endl;
    //value_tmp = norm(x);
    //out << endl << value_tmp << endl;


    const export_type export_out(map_dofs,map_out);
    b_out->doExport(*b, export_out, Tpetra::INSERT);
    x_out->doExport(*x, export_out, Tpetra::INSERT);

    std::string name_A = "A_E" + std::to_string(ensemble_size);
    std::string name_b = "b_E" + std::to_string(ensemble_size) + ".txt";
    std::string name_x = "x_KLU2_E" + std::to_string(ensemble_size) + ".txt";
    std::string name_x2 = "x_CG_E" + std::to_string(ensemble_size) + ".txt";
    std::string name_x2_V2 = "x_CG_V2_E" + std::to_string(ensemble_size) + ".txt";
    std::string name_x3 = "x_tgc_E" + std::to_string(ensemble_size) + ".txt";
    std::string name_x4 = "x_vc_E" + std::to_string(ensemble_size) + ".txt";
    std::string name_CG = "CG_E" + std::to_string(ensemble_size) + ".txt";
    std::string name_CG2 = "CG_E" + std::to_string(ensemble_size) + "_iter";
    std::string name_CG_V2 = "CG_V2_E" + std::to_string(ensemble_size) + ".txt";
    std::string name_CG2_V2 = "CG_V2_E" + std::to_string(ensemble_size) + "_iter";
    std::string name_tgc = "twoGridCycle_E" + std::to_string(ensemble_size) + ".txt";
    std::string name_tgc2 = "twoGridCycle_E" + std::to_string(ensemble_size) + "_iter";
    std::string name_vc = "vCycle_E" + std::to_string(ensemble_size);
    std::string name_vc2 = "vCycle_E" + std::to_string(ensemble_size) + "_iter";    
    std::string name_timer = "timers_E" + std::to_string(ensemble_size) + ".txt";
    writeLocalCrsMatrixToFile<scalar>(name_A, myRank, A);

    writeVectorToFile<scalar>(name_b, myRank, b);
    writeVectorToFile<scalar>(name_x, myRank, x);
    tms["CG"].start();
    conjugateGradient<scalar> (numiters,A,b,x,name_CG,myRank,name_CG2,false);
    tms["CG"].stop();
    writeVectorToFile<scalar>(name_x2, myRank, x);
    tms["CG_V2"].start();
    conjugateGradientV2<scalar> (numiters,A,b,x,name_CG_V2,myRank,name_CG2_V2,false);
    tms["CG_V2"].stop();
    writeVectorToFile<scalar>(name_x2_V2, myRank, x);
/*
    tms["twoGridCycle"].start();
    twoGridCycle(5,5,A,b,x,name_tgc,myRank,name_tgc2,nx,comm,true);
    writeVectorToFile<scalar>(name_x3, myRank, x);
    tms["twoGridCycle"].stop();

    tms["vCycle"].start();
    int maxLvl = 5;
    int nu1 =5;
    int nu2 =5;
    vCycle(0, maxLvl, nu1, nu2, A, b, x, name_vc, myRank, name_vc2, nx, comm,true); 
    writeVectorToFile<scalar>(name_x4, myRank, x);
    tms["vCycle"].stop();
*/
    vector_type z (A->getRangeMap ());
    tms["apply1000"].start();
    for (auto i=0; i < 1000; ++i)
        A->apply(*x, z); 
    tms["apply1000"].stop();

    tms["total"].stop();
    std::ofstream outputfile;
    if (myRank ==0)
        outputfile.open (name_timer);

    outputfile << "\n---CPU statistics of proc 0 for ensemble of size: "<< ensemble_size<< " ---\n";
    outputfile << tms;
    outputfile << "---------------------------\n";
    outputfile.close();
  }
  Kokkos::finalize ();
  comm = Teuchos::null;
  return 0;
  
}

int Example2::execute(int nx, int ny, int ensemble_size, MPI_Comm  _comm)
{
    //if (ensemble_size > 0)
    //{
        /*
        if (ensemble_size <= 1)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,1,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;
        }
        else if (ensemble_size <= 2)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,2,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;
        }
        else
        */
        /*
        if (ensemble_size <= 4)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,4,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        }
        */
        /*
        else if (ensemble_size <= 6)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,6,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        }
        */
        /*
        else if (ensemble_size <= 8)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,8,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        } 
        */
        /*
        else if (ensemble_size <= 10)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,10,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        }         
        else if (ensemble_size <= 12)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,12,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        }              
        else if (ensemble_size <= 14)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,14,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        }
        */
        /*
        else if (ensemble_size <= 16)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,16,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        }
        */
        /*
        else if (ensemble_size <= 18)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,18,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        }          
        else if (ensemble_size <= 20)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,20,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        }
        else if (ensemble_size <= 22)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,22,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        }
        else if (ensemble_size <= 24)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,24,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        }      
        else if (ensemble_size <= 26)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,26,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        } 
        else if (ensemble_size <= 28)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,28,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        }
        else if (ensemble_size <= 30)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,30,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        }
        */
        /*
        else if (ensemble_size <= 32)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int,double,32,Kokkos::Threads>> ensemble_scalar;
            int out = tstart<ensemble_scalar>(nx,ny,ensemble_size,_comm);
            return out;        
        } 
        */
    //}
    //else
    {
        typedef double scalar;
        int out = tstart<scalar>(nx,ny,ensemble_size,_comm);
        return out;
    }
}
