#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh
import mrstlnos.utilities as u
from fwk.testing import *

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"

    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos

    geo_name = '1_cube.geo'
    mesh_name = '1_cube.msh'
    partitioned_mesh_name = '1_cube.msh'


    import shlex, subprocess
    import os
    import numpy as np

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    gmsh.MeshLoader(geo_name,__file__).execute()

    if rank == 0:
        command_line = 'gmsh -3 ' + work_dir + '/' +  mesh_name + ' -o ' + work_dir + '/' +  partitioned_mesh_name +' -part ' + str(siz)
        print command_line
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log','w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
        retcode = p.wait()
        for i in range(1,siz):
            comm.send(1, dest=i, tag=11)
    else:
        data = comm.recv(source=0, tag=11)


    if siz == 1:
        msh = gmsh.MeshLoader(mesh_name,work_dir).execute()
    else:
        msh = gmsh.MeshLoader(mesh_name,work_dir).myrank_execute(rank)


    ensemble_size = 1

    pbl = m.Problem(msh,comm)

    E = 1. * np.ones(ensemble_size)
    nu = 0.2 * np.ones(ensemble_size)
    k = 0.114
    d = 0.

    hf = 1 * np.ones(ensemble_size)
    zero = np.zeros(ensemble_size)
    m.Medium(pbl, "Body 1", "test",E,nu,k,d)

    m.uDirichlet(pbl, "Surf 1 1", "Clamped",1,0.,1,0.,1,0.,1,10.,ensemble_size)
    m.Neumann(pbl, "Surf 1 2","heat fluxes",0,zero,0,zero,0,zero,1,hf,ensemble_size)

    dx = np.zeros(ensemble_size)
    dy = 2. * np.ones(ensemble_size)
    dz = np.zeros(ensemble_size)
    dT = 10. * np.ones(ensemble_size)

    solverList = Teuchos.ParameterList()
    solverList['Convergence Tolerance'] = 10**(-8)
    solverList['Maximum Iterations'] = 1000
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    solverList['Use preconditioner'] = True

    solverList['type']= "BlockGmres"
    solverList['Maximum active set iteration'] = 1

    solverList['Print Teuchos timers'] = True

    mueluParams = Teuchos.ParameterList()
    mueluParams['number of equations'] = 4
    mueluParams['max levels'] = 1
    mueluParams['smoother: type'] = "RELAXATION"

    mueluParamsSub2 = Teuchos.ParameterList()
    mueluParamsSub2['relaxation: type'] = "Symmetric Gauss-Seidel"
    mueluParamsSub2['relaxation: sweeps'] = 1
    mueluParamsSub2['relaxation: damping factor'] = 0.9

    mueluParams['verbosity'] = "low"

    mueluParams['smoother: params'] = mueluParamsSub2

    solverList['mueluParams'] = mueluParams
    solverList['Write txt files'] = True

    useEP = (ensemble_size>1)

    slv = m.IterativeSolver(pbl,args.k,solverList,4, useEP, ensemble_size)

    slv.start()

    tag_name = "Measure point 1"
    which_pos = 0
    which_dof = 3
    DOFperNode = 4
    name_sol = "x_mm.txt"

    pos_1, sol_1 = u.extract(msh, tag_name, DOFperNode, which_pos, which_dof, ensemble_size, name_sol)

    tag_name = "Measure point 2"

    pos_2, sol_2 = u.extract(msh, tag_name, DOFperNode, which_pos, which_dof, ensemble_size, name_sol)

    tests = CTests()
    measured_T_1 =  sol_1[0]
    measured_T_2 =  sol_2[0]
    tests.add(CTest('Temperature 1', measured_T_1, 97.7193,0.05))
    tests.add(CTest('Temperature 2', measured_T_2, 10.,0.05))
    tests.run()


if __name__ == "__main__":
    main()
