#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh
#import matplotlib.pyplot as plt

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"



    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos

    geo_name = '3_cubes.geo'
    mesh_name = '3_cubes.msh'
    partitioned_mesh_name = '3_cubes.msh'


    import shlex, subprocess
    import os
    import numpy as np

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    gmsh.MeshLoader(geo_name,__file__).execute()

    if rank == 0:
        command_line = 'gmsh -3 ' + work_dir + '/' +  mesh_name + ' -o ' + work_dir + '/' +  partitioned_mesh_name +' -part ' + str(siz)
        print command_line
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log','w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
        retcode = p.wait()
        for i in range(1,siz):
            comm.send(1, dest=i, tag=11)
    else:
        data = comm.recv(source=0, tag=11)


    if siz == 1:
        msh = gmsh.MeshLoader(mesh_name,work_dir).execute()
    #elif siz == 3:
    #    msh = gmsh.MeshLoader(mesh_name,file_dir).myrank_execute(rank)
    else:
        msh = gmsh.MeshLoader(mesh_name,work_dir).myrank_execute(rank)


    ensemble_size = 1

    pbl = m.Problem(msh,comm)

    #E = np.linspace(0.5,1.5,ensemble_size)
    E = 1. * np.ones(ensemble_size)
    nu = 0.2 * np.ones(ensemble_size)
    k = 0.114
    d = 0.
    m.Medium(pbl, "Body 1", "test",E,nu,k,d)
    m.Medium(pbl, "Body 2", "test",E,nu,k,d)
    m.Medium(pbl, "Body 3", "test",E,nu,k,d)


    m.uDirichlet(pbl, "Surf 1 1", "Clamped",1,0.,1,0.,1,0.,1,10.,ensemble_size)
    m.uDirichlet(pbl, "Surf 3 2", "Clamped",1,2.,1,2.,1,-2.,1,20.,ensemble_size)

    norm = tbox.Pt(0,0,-1)
    cont = m.Contact(pbl,"Surf 1 2","contact",norm)
    #cont.setInitialyOpen()
    #cont.setNoUpdate()
    #cont.setSticking()
    cont.setMaster(pbl,"Surf 2 1",norm)

    norm = tbox.Pt(0,0,-1)
    cont = m.Contact(pbl,"Surf 2 2","contact",norm)
    #cont.setInitialyOpen()
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl,"Surf 3 1",norm)

    dx = np.zeros(ensemble_size)
    dy = 2. * np.ones(ensemble_size)
    #dy = np.linspace(0.5,1.5,ensemble_size)
    dz = np.zeros(ensemble_size)
    dT = 10. * np.ones(ensemble_size)
    #m.uDirichlet(pbl, "Surf 2", "Moved",0,dx,1,dy,0,dz,1,dT)

    solverList = Teuchos.ParameterList()
    solverList['Convergence Tolerance'] = 10**(-8)
    solverList['Maximum Iterations'] = 1000
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    solverList["MueLu xml file name"] = file_dir + '/direct_solver.xml'
    solverList["convert MueLu xml file"] = True

    solverList['Use preconditioner'] = True

    solverList['type']= "BlockGmres"
    solverList['Maximum active set iteration'] = 5

    #solverList['Print Teuchos timers'] = True
    solverList['Write matrix and vectors'] = True
    solverList['Write txt files'] = True

    if ensemble_size>1:
        slv = m.IterativeSolver(pbl,args.k,solverList,3, True, ensemble_size)
    else:
        slv = m.IterativeSolver(pbl,args.k,solverList,3, False, ensemble_size)

    slv.start()

if __name__ == "__main__":
    main()
