/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wGaussLine2.h"
#include <stdexcept>
#include <cmath>

using namespace tbox;

GaussLine2::GaussLine2(int n) : Gauss()
{
    p.resize(n);
    w.resize(n);
    switch (n)
    {
    case 1:
        p[0] = Pt(0.0);
        w[0] = 2.;
        break;
    case 2:
        p[0] = Pt(-1.0 / sqrt(3.0));
        w[0] = 1.;
        p[1] = Pt(+1.0 / sqrt(3.0));
        w[1] = 1.;
        break;
    case 3:
        p[0] = Pt(-sqrt(3.0 / 5.0));
        w[0] = 5. / 9.;
        p[1] = Pt(0.0);
        w[1] = 8. / 9.;
        p[2] = Pt(+sqrt(3.0 / 5.0));
        w[2] = 5. / 9.;
        break;
    default:
        throw std::runtime_error("not implemented (GaussLine2)!");
    }
}
