/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WHEX8_H
#define WHEX8_H

#include "tbox.h"
#include "wElement.h"

namespace tbox
{

/**
 * @brief Linear hexaedron finite element
 * @authors Kim Liegeois, Romain Boman, Adrien Crovato
 */
class TBOX_API Hex8 : public Element
{
#ifndef SWIG
    friend class MemHex8;
    MemHex8 *pmem; ///< private memory of precalculated values
    static CacheHex8 &getCache();
    MemHex8 &getMem();
    virtual double buildJ(size_t k, gmm::dense_matrix<double> &J) const override;

    virtual void build(MATTYPE type, gmm::dense_matrix<double> &A, std::vector<double> const &u, Fct0 const &fct);
    virtual void buildK(gmm::dense_matrix<double> &K, std::vector<double> const &u, Fct0 const &fct) override;
    virtual void buildM(gmm::dense_matrix<double> &M) override;
    virtual void buildK_mechanic(gmm::dense_matrix<double> &K_e, gmm::dense_matrix<double> &H) override;
    virtual void buildS_thermomechanic(gmm::dense_matrix<double> &S_e, gmm::dense_matrix<double> &D) override;
    virtual void buildL_thermic(gmm::dense_matrix<double> &L_e, gmm::dense_matrix<double> &K_conductivity) override;

    virtual void strain_stress(gmm::dense_matrix<double> &Strain, gmm::dense_matrix<double> &Stress, gmm::dense_matrix<double> &H, std::vector<double> &u) override;

    virtual double computeV(std::vector<double> const &u, Fct0 const &fct) override;
    virtual double computeV() override;
#endif

public:
    Hex8(int n, Tag *_ptag, Tag *_etag, std::vector<Tag *> &_parts, std::vector<Node *> &nods);
    virtual ~Hex8();
    virtual ELTYPE type() const override { return ELTYPE::HEX8; }

#ifndef SWIG
    virtual Cache &getVCache() override;
    virtual Mem &getVMem() override;
#endif
};

} // namespace tbox

#endif //WHEX8_H
