/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wMshData.h"
#include "wNode.h"
#include "wElement.h"
#include "wTag.h"

using namespace tbox;

MshData::MshData() : wSharedObject(), verb(false), name("unnamed")
{
}

void MshData::write(std::ostream &out) const
{
    out << "MshData ['" << name << "']= \n";
    out << "\t" << nodes.size() << " nodes\n";
    out << "\t" << elems.size() << " elements\n";
    out << "\t" << ptags.size() << " physical tags\n";
    out << "\t" << etags.size() << " elementary tags\n";
    out << "\t" << parts.size() << " partitions\n";
}

MshData::~MshData()
{
    std::cout << "~MshData()\n";
    clear();
}

/*
 * @brief Clear the memory
 * @authors Romain Boman
 */
void MshData::clear()
{
    for (auto n : nodes)
        delete n;
    nodes.clear();

    for (auto e : elems)
        delete e;
    elems.clear();

    for (auto &p : ptags)
        delete p.second;
    ptags.clear();

    for (auto &p : etags)
        delete p.second;
    etags.clear();

    for (auto &p : parts)
        delete p.second;
    parts.clear();
}
