#include "wExample1.h"

#ifdef MYAPP_MPI
#include "Teuchos_DefaultMpiComm.hpp"
#else
#include "Teuchos_DefaultSerialComm.hpp"
#endif

#ifdef MYAPP_EPETRA
#include "Epetra_SerialDenseVector.h"
#endif

using namespace tlnos;

Example1::Example1()
{
}

int 
Example1::src_file(std::string const &infile, Teuchos::Comm<int> &comm) 
{
  Teuchos::ParameterList p;
  Teuchos::Ptr<Teuchos::ParameterList> pref(&p);
  Teuchos::updateParametersFromXmlFileAndBroadcast(infile, pref, comm);

  std::cout << "\nProcessor " << comm.getRank()
            << "  has param list : \n" << p << std::endl;

  return 0;
}

void 
Example1::write(std::ostream &out) const
{
    out << "tlnos::Example1:\n";
}

int 
Example1::execute(std::string const &infile) 
{
    /*
    int *argc = new int(1);// mem leak volontaire
    // ATTENTION: argv doit rester alloué pdt toute la durée de vie d'appli!
    char **argv = new char*[2];// mem leak volontaire
    argv[0] = new char[6]; strcpy(argv[0], "MyApp"); // mem leak volontaire
    argv[1] = new char[1]; strcpy(argv[1], ""); // mem leak volontaire
    */
    int status = 0;

    // Initialize MPI and timer
    //Teuchos::GlobalMPISession mpiSession(argc,&argv); // [RB] does not work anymore... (MPI already initialised?)
    Teuchos::GlobalMPISession mpiSession(); // [RB] from https://trilinos.org/pipermail/trilinos-users/2016-July/005597.html

    // Use of flag set in CMakeLists.txt by how Trilinos was configured
#ifdef MYAPP_MPI
    Teuchos::MpiComm<int> comm =
      Teuchos::MpiComm<int>(Teuchos::opaqueWrapper((MPI_Comm)MPI_COMM_WORLD));
#else
    Teuchos::SerialComm<int> comm = Teuchos::SerialComm<int>();
#endif

    try 
    {
        // Function from another file
        status = src_file(infile, comm);

        // Flag set in CMakeLists.txt that detects if Epetra was enabled in Trilinos
#ifdef MYAPP_EPETRA
        const int len=10;
        Epetra_SerialDenseVector vec(len);
        if (vec.Length() != len) status += 1000;
        std::cout << "\nEpetra called for vec of length " << len << std::endl;
#endif
    }

    catch (std::exception& e) 
    {
        std::cout << e.what() << std::endl;
        status = 10;
    }
    catch (std::string& s) 
    {
        std::cout << s << std::endl;
        status = 20;
    }
    catch (char *s) 
    {
        std::cout << s << std::endl;
        status = 30;
    }
    catch (...) {
        std::cout << "Caught unknown exception!" << std::endl;
        status = 40;
    }

    // Status=0 signals to ctest that the test passed.
    return status;
}


