/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WTIMEINTEGRATION_H
#define WTIMEINTEGRATION_H

#include "waves.h"
#include "wObject.h"
#include "wTimer.h"
#include "wElement.h"
#include <iostream>
#include <vector>
#include <memory>
#include "gmm_gmm.h"

namespace waves
{

enum class Wgmsh
{
    WG_NONE = 0, ///< do not write results on disk
    WG_ONE = 1,  ///< continuously overwrite the same result file (onelab)
    WG_ALL = 2   ///< write a separate file for each archived step
};

/**
 * @brief Time integration base class
 */

class WAVES_API TimeIntegration : public fwk::wObject
{
public:
    bool verbose;
    std::shared_ptr<Problem> pbl;
#ifndef SWIG
    DisplayHook *dhook;
#endif

    bool stopit;

    int nthreads; ///< nb of threads for the assembly
    Wgmsh wgmsh;  ///< how to write results on disk?

    double ttot;  ///< final time
    double dt;    ///< time step
    int savefreq; ///< archive/update_gui frequency

    fwk::Timer chrono1;

    std::vector<double> u1;
    std::vector<double> v1;

    TimeIntegration(std::shared_ptr<waves::Problem> _pbl);

    virtual char const *clsname() const { return "TimeIntegration"; }

    virtual void start(std::shared_ptr<tbox::MshExport> mshWriter);

    void dummyIC();

    void buildS(gmm::csr_matrix<double> &S2);

    void buildKM(gmm::csr_matrix<double> &K2, std::vector<double> &Md, std::vector<double> const &u);
    void buildKM_tbb_lambda(gmm::csr_matrix<double> &K2, std::vector<double> &Md, std::vector<double> const &u);

    void setGUI(DisplayHook &hook) { dhook = &hook; }

    void build(tbox::MATTYPE type, gmm::csr_matrix<double> &A2);

    void stop() { stopit = true; }

#ifndef SWIG
    virtual void write(std::ostream &out) const override;
#endif
};

} // namespace waves

#endif //WTIMEINTEGRATION_H
